
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#include <zlib.h> 
#include <errno.h> 

#include "dbGDS2Reader.h"
#include "dbArray.h"

#include "tlException.h"
#include "tlString.h"

namespace db
{

// ---------------------------------------------------------------
//  Stream implementation

Stream::Stream ()
  : m_pos (0), mp_bptr (0)
{ 
  m_bcap = 4096; // initial buffer capacity
  m_blen = 0;
  mp_buffer = new char [m_bcap];
}

const char * 
Stream::get (size_t n)
{
  if (m_blen < n) {

    //  to keep move activity low, allocate twice as much as required
    if (m_bcap < n * 2) {

      while (m_bcap < n) {
        m_bcap *= 2;
      }

      char *buffer = new char [m_bcap];
      memcpy (buffer, mp_bptr, m_blen);
      delete [] mp_buffer;
      mp_buffer = buffer;

    } else {
      memmove (mp_buffer, mp_bptr, m_blen);
    }

    m_blen += read (mp_buffer + m_blen, m_bcap - m_blen); 
    mp_bptr = mp_buffer;

  }

  if (m_blen >= n) {
    const char *r = mp_bptr;
    mp_bptr += n;
    m_blen -= n;
    m_pos += n;
    return r;
  } else {
    return 0;
  }

}

// ---------------------------------------------------------------
//  FileStream implementation

FileStream::FileStream ()
  : m_file (NULL)
{
  // .. nothing yet ..
}

FileStream::~FileStream ()
{
  if (m_file != NULL) {
    close ();
  }  
}

void 
FileStream::open (const std::string &path) throw (tl::Exception)
{
  tl_assert (m_file == NULL);
  reset_pos ();

  m_source = path;
  m_file = fopen (path.c_str (), "rb");
  if (m_file == NULL) {
    throw FileOpenErrorException (m_source, errno);
  }
}

void 
FileStream::close ()
{
  tl_assert (m_file != NULL);
  fclose (m_file);
  m_file = NULL;
}

size_t 
FileStream::read (char *b, size_t n) throw (tl::Exception)
{
  tl_assert (m_file != NULL);
  size_t ret = fread (b, 1, n, m_file);
  if (ret < n) {
    if (ferror (m_file)) {
      throw FileReadErrorException (m_source, pos (), ferror (m_file));
    }
  }

  return ret;
}

// ---------------------------------------------------------------
//  ZLibStream implementation

ZLibStream::ZLibStream ()
  : m_zs (NULL)
{
  // .. nothing yet ..
}

ZLibStream::~ZLibStream ()
{
  if (m_zs != NULL) {
    close ();
  }  
}

void 
ZLibStream::open (const std::string &path) throw (tl::Exception)
{
  tl_assert (m_zs == NULL);
  reset_pos ();

  m_source = path;
  m_zs = gzopen (path.c_str (), "rb");
  if (m_zs == NULL) {
    throw FileOpenErrorException (m_source, errno);
  }
}

void 
ZLibStream::close ()
{
  tl_assert (m_zs != NULL);
  gzclose (m_zs);
  m_zs = NULL;
}

size_t 
ZLibStream::read (char *b, size_t n) throw (tl::Exception)
{
  tl_assert (m_zs != NULL);
  int ret = gzread (m_zs, b, n);
  if (ret < 0) {
    int gz_err = 0;
    const char *em = gzerror (m_zs, &gz_err);
    if (gz_err == Z_ERRNO) {
      throw FileReadErrorException (m_source, pos (), errno);
    } else {
      throw ZLibReadErrorException (m_source, pos (), em);
    }
  }

  return ret;
}

// ---------------------------------------------------------------
//  FileStream implementation

FilePipe::FilePipe ()
  : m_file (NULL)
{
  // .. nothing yet ..
}

FilePipe::~FilePipe ()
{
  if (m_file != NULL) {
    close ();
  }  
}

void 
FilePipe::open (const std::string &path) throw (tl::Exception)
{
  tl_assert (m_file == NULL);
  reset_pos ();

  m_source = path;
  m_file = popen (path.c_str (), "r");
  if (m_file == NULL) {
    throw FilePOpenErrorException (m_source, errno);
  }
}

void 
FilePipe::close ()
{
  tl_assert (m_file != NULL);
  fclose (m_file);
  m_file = NULL;
}

size_t 
FilePipe::read (char *b, size_t n) throw (tl::Exception)
{
  tl_assert (m_file != NULL);
  size_t ret = fread (b, 1, n, m_file);
  if (ret < n) {
    if (ferror (m_file)) {
      throw FilePReadErrorException (m_source, pos (), ferror (m_file));
    }
  }

  return ret;
}

// ---------------------------------------------------------------
//  GDS2LayerMap

/// A helper class to join two datatype map members
struct LmapJoinOp1
{
  void operator() (unsigned int &a, unsigned int b)
  {
    //  TODO: could also do?: throw an exception ..
    a = b;
  }
};

/// A helper class to join two layer map members 
/// (this implementation basically merged the datatype maps)  
struct LmapJoinOp2
{
  void operator() (GDS2LayerMap::datatype_map &a, const GDS2LayerMap::datatype_map &b)
  {
    LmapJoinOp1 op1;
    a.add (b.begin (), b.end (), op1);
  }
};

/// Utility typedefs for the expression parser  
typedef std::pair<gds2_ld_type, gds2_ld_type> gds2_ld_interval;

/// Utility typedefs for the expression parser  
typedef std::vector<gds2_ld_interval> gds2_ld_interval_vector;


GDS2LayerMap::GDS2LayerMap ()
  : m_ld_map ()
{
  //  .. nothing yet ..
}

std::pair<bool, unsigned int> 
GDS2LayerMap::logical (const GDS2LDPair &p) const
{
  const datatype_map *dm = m_ld_map.mapped (p.layer);
  if (dm) {
    const unsigned int *l = dm->mapped (p.datatype);
    if (l) {
      return std::make_pair (true, *l);
    }
  }
  return std::make_pair (false, 0);
}

std::string
GDS2LayerMap::mapping_str (unsigned int ll) const
{
  std::string s;
  bool f1 = true;

  for (ld_map::const_iterator l = m_ld_map.begin (); l != m_ld_map.end (); ++l) {

    std::string t = tl::to_string (l->first.first);
    if (l->first.first < l->first.second - 1) {
      t += "-";
      t += tl::to_string (l->first.second - 1);
    }
    t += "/";

    bool f2 = true;
    for (datatype_map::const_iterator d = l->second.begin (); d != l->second.end (); ++d) {
      if (d->second == ll) {

        //  create a string representation
        if (!f2) {
          s += ",";
        } else {
          if (!f1) {
            s += ";";
          }
          f1 = false;
          s += t;
        }
        f2 = false;

        s += tl::to_string (d->first.first);
        if (d->first.first < d->first.second - 1) {
          s += "-";
          s += tl::to_string (d->first.second - 1);
        }

      }

    }
    
  }

  return s;
}

GDS2LDPair 
GDS2LayerMap::mapping (unsigned int ll) const
{
  for (ld_map::const_iterator l = m_ld_map.begin (); l != m_ld_map.end (); ++l) {
    for (datatype_map::const_iterator d = l->second.begin (); d != l->second.end (); ++d) {
      if (d->second == ll) {
        return GDS2LDPair (l->first.first, d->first.first);
      }
    }
  }
  return GDS2LDPair::invalid ();
}

void 
GDS2LayerMap::map (const GDS2LDPair &p, unsigned int l)
{
  insert (p, p, l);
}

void 
GDS2LayerMap::map (const GDS2LDPair &p1, const GDS2LDPair &p2, unsigned int l)
{
  insert (p1, p2, l);
}

/// Utility function for the expression parser:  
/// skip blanks in a expression
const char * 
skip_blanks (const char *&cp)
{
  while (*cp && std::isspace (*cp)) {
    ++cp;
  }
  return cp;
}

/// Utility function for the expression parser:  
/// Parse an interval
const char *
parse_interval (const char *&cp, gds2_ld_interval &p) throw (GDS2LayerSpecFormatException)
{
  gds2_ld_type n1 = 0, n2 = 0;

  skip_blanks (cp); 
  while (*cp && std::isdigit (*cp)) {
    n1 = n1 * 10 + gds2_ld_type (*cp - '0');
  }

  if (*skip_blanks (cp) == '-') {

    ++cp;

    skip_blanks (cp);
    while (*cp && std::isdigit (*cp)) {
      n2 = n2 * 10 + gds2_ld_type (*cp - '0');
    }

  } else {
    n2 = n1;
  }

  p.first = n1;
  p.second = n2;
  return cp;
}

/// Utility function for the expression parser:  
/// Parse an interval list
const char *
parse_intervals (const char *&cp, gds2_ld_interval_vector &v) throw (GDS2LayerSpecFormatException)
{
  --cp;
  do {
    ++cp;

    v.push_back (gds2_ld_interval (0, 0));
    parse_interval (cp, v.back ());

  } while (*skip_blanks (cp) == ',');

  return cp;
}

void 
GDS2LayerMap::map (const std::string &expr, unsigned int l) throw (GDS2LayerSpecFormatException)
{
  const char *cp = expr.c_str ();

  --cp;
  do {
    ++cp;

    gds2_ld_interval_vector vl, vd;

    parse_intervals (cp, vl);

    if (*skip_blanks (cp) == '/') {
      parse_intervals (cp, vd);
    } else {
      vd.push_back (gds2_ld_interval (0, 0));
    }

    datatype_map dm;
    for (gds2_ld_interval_vector::const_iterator di = vd.begin (); di != vd.end (); ++di) {
      LmapJoinOp1 op1;
      dm.add (di->first, di->second + 1, l, op1);
    }
    for (gds2_ld_interval_vector::const_iterator li = vl.begin (); li != vl.end (); ++li) {
      LmapJoinOp2 op2;
      m_ld_map.add (li->first, li->second + 1, dm, op2);
    }

  } while (*skip_blanks (cp) == ';');

  if (*cp != '\0') {
    throw GDS2LayerSpecFormatException (cp);
  }

}

void
GDS2LayerMap::insert (const GDS2LDPair &p1, const GDS2LDPair &p2, unsigned int l) 
{
  //  create a single-interval list for the datatype range
  GDS2LayerMap::datatype_map dt;
  LmapJoinOp1 op1;
  dt.add (p1.datatype, p2.datatype + 1, l, op1);

  //  add this to the layers using the special join operator that
  //  combines the datatype intervals
  LmapJoinOp2 op2;
  m_ld_map.add (p1.layer, p2.layer + 1, dt, op2);
}

// ---------------------------------------------------------------
//  GDS2Reader


GDS2Reader::GDS2Reader (Stream &s)
  : m_stream (s), 
    m_progress (tl::translate ("Reading GDS2 file"), 0, 10000)
{
  m_progress.set_format (tl::translate ("%.0f MB"));
  m_progress.set_res (1.0);
  m_progress.set_unit (1024 * 1024);
}

GDS2Reader::~GDS2Reader ()
{
  //  .. nothing yet ..
}

const GDS2LayerMap &
GDS2Reader::read (db::Layout &layout, const GDS2LayerMap &map, bool create) throw (tl::Exception)
{
  m_layer_map = map;
  layout.start_changes ();
  do_read (layout, create);
  layout.end_changes ();
  return m_layer_map;
}

const GDS2LayerMap &
GDS2Reader::read (db::Layout &layout) throw (tl::Exception)
{
  m_layer_map = GDS2LayerMap ();
  layout.start_changes ();
  do_read (layout, true /*create layers*/);
  layout.end_changes ();
  return m_layer_map;
}

short 
GDS2Reader::get_record () throw (tl::Exception)
{  
  unsigned char *b = (unsigned char *) m_stream.get (4);
  if (! b) {
    error (tl::translate ("Unexpected end-of-file"));
  }

  //  @@@ this can be done more efficiently
  //  @@@ officially, the record length must be less then 0x8000.
  m_recnum++;
  m_reclen = (size_t (b[0]) << 8) | size_t (b[1]);
  short rec_id = (short (b[2]) << 8) | short (b[3]);

  if (m_reclen < 4) {
    error (tl::translate ("Invalid record length (less than 4)"));
  }
  if (m_reclen >= 0x8000) {
    error (tl::translate ("Invalid record length (larger than 0x8000)"));
  }
  if (m_reclen % 2 == 1) {
    error (tl::translate ("Odd record length"));
  }

  m_reclen -= 4;

  if (m_reclen > 0) {
    mp_rec_buf = (unsigned char *) m_stream.get (m_reclen);
    if (! mp_rec_buf) {
      error (tl::translate ("Unexpected end-of-file"));
    }
  } else {
    mp_rec_buf = 0;
  }
   
  m_recptr = 0; 
  return rec_id;
}

inline int 
GDS2Reader::get_int () throw (tl::Exception)
{
  unsigned char *b = mp_rec_buf + m_recptr;
  m_recptr += 4;

  //  @@@ this can be done more efficiently ..
  return (int (b[0]) << 24) | (int (b[1]) << 16) | (int (b[2]) << 8) | int (b[3]);
}

inline short 
GDS2Reader::get_short () throw (tl::Exception)
{
  unsigned char *b = mp_rec_buf + m_recptr;
  m_recptr += 2;

  //  @@@ this can be done more efficiently ..
  return (short (b[0]) << 8) | short (b[1]);
}

inline double 
GDS2Reader::get_double () throw (tl::Exception)
{
  unsigned char *b = mp_rec_buf + m_recptr;
  m_recptr += 8;

  //  @@@ this can be done more efficiently ..
  double x = 0;
  for (int i = 1; i < 8; ++i) {
    x = x * 256.0 + double (b[i]);
  }

  if (b[0] & 0x80) {
    x = -x;
  }
  
  double e = double (b[0] & 0x7f) - (64.0 + 14);
  x *= pow (16.0, e);
  return x;
}

const char *
GDS2Reader::get_string () throw (tl::Exception)
{
  if (m_reclen == 0) {
    return "";
  }

  if (mp_rec_buf [m_reclen - 1] == 0) {
    //  we already have a terminating '\0': just return the string's location
    return (const char *) mp_rec_buf;
  } else {
    //  use the temporary buffer to create a zero-terminated string
    m_string_buf.assign ((const char *) mp_rec_buf, 0, m_reclen);
    return m_string_buf.c_str ();
  }
}

void
GDS2Reader::get_string (tl::string &s) const throw (tl::Exception)
{
  s.assign ((const char *) mp_rec_buf, 0, m_reclen);
}

void 
GDS2Reader::error (const std::string &msg) throw (tl::Exception)
{
  throw GDS2ReaderException (msg, m_stream.pos (), m_recnum, m_cellname.c_str ());
}

void 
GDS2Reader::warn (const std::string &msg) 
{
  // @@@ simple handling ..
  fprintf (stderr, tl::translate ("WARNING: %s (position=%ld, record number=%ld, cell=%s)\n").c_str (), msg.c_str (), m_stream.pos (), m_recnum, m_cellname.c_str ());
}

std::pair <bool, unsigned int> 
GDS2Reader::open_dl (db::Layout &layout, const GDS2LDPair &dl, bool create) throw (tl::Exception)
{
  std::pair<bool, unsigned int> ll = m_layer_map.logical (dl);
  if (ll.first || ! create) {
    return ll;
  } else {

    //  and create the layer
    db::LayerProperties lp;
    lp.layer = dl.layer;
    lp.datatype = dl.datatype;
    lp.name = tl::sprintf ("%d/%d", dl.layer, dl.datatype);
    unsigned int ll = layout.insert_layer (lp);
    m_layer_map.map (dl, ll);

    return std::make_pair (true, ll);

  }
}

const short sHEADER       = 0x0002;
const short sBGNLIB       = 0x0102;
const short sLIBNAME      = 0x0206;
const short sUNITS        = 0x0305;
const short sENDLIB       = 0x0400;
const short sBGNSTR       = 0x0502;
const short sSTRNAME      = 0x0606;
const short sENDSTR       = 0x0700;
const short sBOUNDARY     = 0x0800;
const short sPATH         = 0x0900;
const short sSREF         = 0x0a00;
const short sAREF         = 0x0b00;
const short sTEXT         = 0x0c00;
const short sLAYER        = 0x0d02;
const short sDATATYPE     = 0x0e02;
const short sWIDTH        = 0x0f03;
const short sXY           = 0x1003;
const short sENDEL        = 0x1100;
const short sSNAME        = 0x1206;
const short sCOLROW       = 0x1302;
const short sTEXTNODE     = 0x1400;
const short sNODE         = 0x1500;
const short sTEXTTYPE     = 0x1602;
const short sPRESENTATION = 0x1701;
const short sSTRING       = 0x1906;
const short sSTRANS       = 0x1a01;
const short sMAG          = 0x1b05;
const short sANGLE        = 0x1c05;
const short sREFLIBS      = 0x1f06;
const short sFONTS        = 0x2006;
const short sPATHTYPE     = 0x2102;
const short sGENERATIONS  = 0x2202;
const short sATTRTABLE    = 0x2306;
const short sSTYPTABLE    = 0x2406;
const short sSTRTYPE      = 0x2502;
const short sELFLAGS      = 0x2601;
const short sELKEY        = 0x2703;
const short sNODETYPE     = 0x2a02;
const short sPROPATTR     = 0x2b02;
const short sPROPVALUE    = 0x2c06;
const short sBOX          = 0x2d00;
const short sBOXTYPE      = 0x2e02;
const short sPLEX         = 0x2f03;
const short sBGNEXTN      = 0x3003;
const short sENDEXTN      = 0x3103;
const short sTAPENUM      = 0x3202;
const short sTAPECODE     = 0x3302;
const short sSTRCLASS     = 0x3401;
const short sRESERVED     = 0x3503;
const short sFORMAT       = 0x3602;
const short sMASK         = 0x3706;
const short sENDMASKS     = 0x3800;
const short sLIBDIRSIZE   = 0x3902;
const short sSRFNAME      = 0x3a06;

struct GDS2XY
{
  unsigned char x[4];
  unsigned char y[4];
};

struct GDS2XYToPoint
{
  db::Point operator() (const GDS2XY &p) 
  {
    //  @@@ this can be done more efficiently ..
    int x = (int (p.x[0]) << 24) | (int (p.x[1]) << 16) | (int (p.x[2]) << 8) | int (p.x[3]);
    int y = (int (p.y[0]) << 24) | (int (p.y[1]) << 16) | (int (p.y[2]) << 8) | int (p.y[3]);
    return db::Point (x, y);
  }
};

inline bool 
eq_x (const GDS2XY &a, const GDS2XY &b)
{
  return a.x[0] == b.x[0] && a.x[1] == b.x[1] && a.x[2] == b.x[2] && a.x[3] == b.x[3];
}

inline bool 
eq_y (const GDS2XY &a, const GDS2XY &b)
{
  return a.y[0] == b.y[0] && a.y[1] == b.y[1] && a.y[2] == b.y[2] && a.y[3] == b.y[3];
}

void 
GDS2Reader::do_read (db::Layout &layout, bool create_layers) throw (tl::Exception)
{
  m_recnum = 0;
  --m_recnum;
  m_reclen = 0;
  m_cellname = "";
  m_libname = "";

  //  read header
  if (get_record () != sHEADER) {
    error (tl::translate ("HEADER record expected"));
  }
  if (get_record () != sBGNLIB) {
    error (tl::translate ("BGNLIB record expected"));
  }

  //  read until 
  short rec_id = 0;
  do {
    rec_id = get_record ();
    if (rec_id == sLIBDIRSIZE ||
        rec_id == sSRFNAME ||
        rec_id == sREFLIBS ||
        rec_id == sFONTS ||
        rec_id == sATTRTABLE ||
        rec_id == sGENERATIONS ||
        rec_id == sFORMAT ||
        rec_id == sMASK ||
        rec_id == sENDMASKS) {
      //  OK and overread
    } else if (rec_id == sLIBNAME) {
      m_libname = get_string ();
    } else if (rec_id == sUNITS) {
      break;
    } else {
      error (tl::translate ("Invalid record or data type"));
    }
  } while (true);

  //  get units
  m_dbuu = get_double ();
  m_dbu = get_double ();
  layout.dbu (m_dbuu);

  //  this container has been found to grow quite a lot.
  //  using a list instead of a vector should make this more efficient.
  tl::vector<db::CellInstArray> instances;

  //  get cells
  while ((rec_id = get_record ()) == sBGNSTR) {

    m_progress.set (m_stream.pos ());

    //  erase current instance list 
    instances.erase (instances.begin (), instances.end ());

    if (get_record () != sSTRNAME) {
      error (tl::translate ("STRNAME record expected"));
    }

    get_string (m_cellname);
    std::pair<bool, db::Layout::cell_index_type> c = layout.cell_by_name (m_cellname.c_str ()); 
    if (c.first) {
      //  cell already there: just add shapes (cell might have been created through forward reference)
      m_cellindex = c.second;
    } else {
      m_cellindex = layout.add_cell (m_cellname.c_str ());
    }
    db::Cell &cell = layout.cell (m_cellindex);

    //  read cell content
    while ((rec_id = get_record ()) != sENDSTR) { 

      m_progress.set (m_stream.pos ());

      if (rec_id == sBOUNDARY) {

        GDS2LDPair ld; 

        do {
          rec_id = get_record ();
        } while (rec_id == sELFLAGS || rec_id == sPLEX);
        if (rec_id != sLAYER) {
          error (tl::translate ("LAYER record expected"));
        }
        ld.layer = get_short ();
        if (get_record () != sDATATYPE) {
          error (tl::translate ("DATATYPE record expected"));
        }
        ld.datatype = get_short ();

        if (get_record () != sXY) {
          error (tl::translate ("XY record expected"));
        }

        std::pair<bool, unsigned int> ll = open_dl (layout, ld, create_layers);
        if (ll.first) {

          GDS2XYToPoint pt_conv;

          //  create a box object if possible
          GDS2XY *xy = (GDS2XY *) mp_rec_buf;
          if ((m_reclen == 4 * sizeof (GDS2XY) || 
               (m_reclen == 5 * sizeof(GDS2XY) && pt_conv (xy[4]) == pt_conv (xy[0]))) && 
              ((eq_x (xy[0], xy[1]) && eq_x (xy[2], xy[3]) && 
                eq_y (xy[1], xy[2]) && eq_y (xy[0], xy[3])) ||
               (eq_x (xy[1], xy[2]) && eq_x (xy[0], xy[3]) && 
                eq_y (xy[0], xy[1]) && eq_y (xy[2], xy[3])))) {

            //  we can create a box object:
            db::Box box;
            while (xy < (GDS2XY *) (mp_rec_buf + m_reclen)) {
              box += pt_conv (*xy++);
            }

            cell.shapes (ll.second).insert (box);

          } else {

            //  convert the GDS2 record into the polygon.
            db::SimplePolygon poly;
            poly.assign_hull ((GDS2XY *) mp_rec_buf, (GDS2XY *) (mp_rec_buf + m_reclen), pt_conv);

            if (poly.hull ().size () < 3) {
              warn (tl::translate ("BOUNDARY with less than 3 points ignored"));
            } else {
              //  this will copy the polyon:
              cell.shapes (ll.second).insert (db::SimplePolygonRef (poly, layout.shape_repository ()));
            }

          }

        }

        while (get_record () != sENDEL) { 
          //  @@@ this is not pretty clean ..
          //  overread all unto ENDEL
        } 

      } else if (rec_id == sPATH) {

        GDS2LDPair ld; 

        do {
          rec_id = get_record ();
        } while (rec_id == sELFLAGS || rec_id == sPLEX);
        if (rec_id != sLAYER) {
          error (tl::translate ("LAYER record expected"));
        }
        ld.layer = get_short ();
        if (get_record () != sDATATYPE) {
          error (tl::translate ("DATATYPE record expected"));
        }
        ld.datatype = get_short ();
          
        rec_id = get_record ();

        short type = 0; 
        if (rec_id == sPATHTYPE) {
          type = get_short (); 
          rec_id = get_record ();
        }

        if (type != 0 && type != 1 && type != 2 && type != 4) {
          warn (tl::translate ("Unsupported PATHTYPE"));
          type = 0;
        }

        db::Coord w = 0;
        if (rec_id == sWIDTH) {
          w = get_int ();
          rec_id = get_record ();
        }

        db::Coord bgn_ext = 0;
        db::Coord end_ext = 0;

        if (rec_id == sBGNEXTN) {
          bgn_ext = get_int ();
          rec_id = get_record ();
        } else {
          if (type == 2 || type == 1) {
            bgn_ext = w / 2;
          } 
        }

        if (rec_id == sENDEXTN) {
          end_ext = get_int ();
          rec_id = get_record ();
        } else {
          if (type == 2 || type == 1) {
            end_ext = w / 2;
          } 
        }

        if (rec_id != sXY) {
          error (tl::translate ("XY record expected"));
        }

        std::pair<bool, unsigned int> ll = open_dl (layout, ld, create_layers);
        if (ll.first) {

          //  convert the GDS2 record into the polygon.
          GDS2XYToPoint pt_conv;

          //  this will copy the path:
          db::Path path;
          path.width (w);
          path.extensions (bgn_ext, end_ext);
          path.assign ((GDS2XY *) mp_rec_buf, (GDS2XY *) (mp_rec_buf + m_reclen), pt_conv);
          path.round (type == 1);

          if (path.points () < 2) {
            warn (tl::translate ("PATH with less than 2 points ignored"));
          } else {
            cell.shapes (ll.second).insert (db::PathRef (path, layout.shape_repository ()));
          }

        }

        while (get_record () != sENDEL) { 
          //  @@@ this is not pretty clean ..
          //  overread all unto ENDEL
        } 

      } else if (rec_id == sSREF || rec_id == sAREF) {

        bool array = (rec_id == sAREF);

        do {
          rec_id = get_record ();
        } while (rec_id == sELFLAGS || rec_id == sPLEX);
        if (rec_id != sSNAME) {
          error (tl::translate ("SNAME record expected"));
        }

        db::Layout::cell_index_type ci;

        {
          //  obtain cell or create new: "cn" is not valid beyond the 
          //  scope of the get_string() call
          const char *cn = get_string ();
          std::pair<bool, db::Layout::cell_index_type> c = layout.cell_by_name (cn); 
          if (c.first) {
            //  cell already there: just add shapes (cell might have been created through forward reference)
            ci = c.second;
          } else {
            ci = layout.add_cell (cn);
          }
        }

        bool mirror = false;
        int angle = 0;
        double mag = 1.0;
        bool is_mag = false;

        rec_id = get_record ();
        if (rec_id == sSTRANS) {
          short f = get_short ();
          if ((f & 0x8000) != 0) {
            mirror = true;
          }
          if ((f & (4 | 2)) != 0) {
            error (tl::translate ("Absolute transformations are supported"));
          }  
          rec_id = get_record ();
        }

        if (rec_id == sMAG) {
          mag = get_double ();
          if (fabs (mag - 1.0) > 1e-9) { 
            is_mag = true;
          }
          rec_id = get_record ();
        }

        if (rec_id == sANGLE) {
          double a = get_double () / 90.0;
          if (a < -4 || a > 4) {
            error (tl::translate ("Invalid rotation angle (less than -360 or larger than 360)"));
          }
          angle = int (a < 0 ? (a - 0.5) : (a + 0.5));
          if (fabs (double (angle) - a) > 1e-9) {
            error (tl::translate ("Invalid rotation angle (not a multiple of 90)"));
          }
          while (angle < 0) {
            angle += 4;
          }
          while (angle >= 4) {
            angle -= 4;
          }
          rec_id = get_record ();
        }

        if (array) {

          //  Array reference
          if (rec_id != sCOLROW) {
            error (tl::translate ("COLROW record expected"));
          }

          int cols = get_short ();
          int rows = get_short ();
          if (cols <= 0) {
            error (tl::translate ("Invalid COLROW record (negative or zero column count)"));
          }
          if (rows <= 0) {
            error (tl::translate ("Invalid COLROW record (negative or zero row count)"));
          }

          //  Array reference
          if (get_record () != sXY) {
            error (tl::translate ("XY record expected"));
          }

          //  Create the instance
          db::Coord x = get_int ();
          db::Coord y = get_int ();
          db::Coord xc = get_int () - x;
          db::Coord yc = get_int () - y;
          db::Coord xr = get_int () - x;
          db::Coord yr = get_int () - y;

          if (cols > 1) {
            if (xc % cols != 0 || yc % cols != 0) {
              warn (tl::translate ("Off-grid AREF column vector"));
            }
            xc /= cols;
            yc /= cols;
          }

          if (rows > 1) {
            if (xr % rows != 0 || yr % rows != 0) {
              warn (tl::translate ("Off-grid AREF row vector"));
            }
            xr /= rows;
            yr /= rows;
          }

          //  insert the cell array
          if (is_mag) {
            instances.push_back (db::CellInstArray (db::CellInst (ci), 
                                                    db::Trans (angle, mirror, db::Point (x, y)), mag, db::Point (xc, yc), db::Point (xr, yr), cols, rows));
          } else {
            instances.push_back (db::CellInstArray (db::CellInst (ci), 
                                                    db::Trans (angle, mirror, db::Point (x, y)), db::Point (xc, yc), db::Point (xr, yr), cols, rows));
          }

        } else {

          //  Single reference
          if (rec_id != sXY) {
            error (tl::translate ("XY record expected"));
          }

          //  Create the instance
          db::Coord x = get_int ();
          db::Coord y = get_int ();

          if (is_mag) {
            instances.push_back (db::CellInstArray (db::CellInst (ci), db::Trans (angle, mirror, db::Point (x, y)), mag));
          } else {
            instances.push_back (db::CellInstArray (db::CellInst (ci), db::Trans (angle, mirror, db::Point (x, y))));
          }

        }  

        while (get_record () != sENDEL) { 
          //  @@@ this is not pretty clean ..
          //  overread all unto ENDEL
        } 

      } else if (rec_id == sTEXT) {

        GDS2LDPair ld; 

        do {
          rec_id = get_record ();
        } while (rec_id == sELFLAGS || rec_id == sPLEX);
        if (rec_id != sLAYER) {
          error (tl::translate ("LAYER record expected"));
        }
        ld.layer = get_short ();
        if (get_record () != sTEXTTYPE) {
          error (tl::translate ("DATATYPE record expected"));
        }
        ld.datatype = get_short ();

        std::pair<bool, unsigned int> ll = open_dl (layout, ld, create_layers);
        if (ll.first) {

          rec_id = get_record ();
          if (rec_id == sPRESENTATION) {
            rec_id = get_record ();
          }
          if (rec_id == sPATHTYPE) {
            rec_id = get_record ();
          }
          if (rec_id == sWIDTH) {
            rec_id = get_record ();
          }

          bool mirror = false;
          int angle = 0;

          if (rec_id == sSTRANS) {

            short f = get_short ();
            if ((f & 0x8000) != 0) {
              mirror = true;
            }

            rec_id = get_record ();

            if (rec_id == sMAG) {
              rec_id = get_record ();
            }

            if (rec_id == sANGLE) {

              double a = get_double () / 90.0;
              if (a < -4 || a > 4) {
                error (tl::translate ("Invalid rotation angle (less than -360 or larger than 360)"));
              }
              angle = int (a < 0 ? (a - 0.5) : (a + 0.5));
              if (fabs (double (angle) - a) > 1e-9) {
                error (tl::translate ("Invalid rotation angle (not a multiple of 90)"));
              }
              while (angle < 0) {
                angle += 4;
              }
              while (angle >= 4) {
                angle -= 4;
              }

              rec_id = get_record ();
              
            }
          }

          if (rec_id != sXY) {
            error (tl::translate ("XY record expected"));
          }
          db::Coord x = get_int ();
          db::Coord y = get_int ();

          db::Trans t (angle, mirror, db::Point (x, y));

          if (get_record () != sSTRING) {
            error (tl::translate ("STRING record expected"));
          }

          //  Create the text
          db::Text text (get_string (), t);
          cell.shapes (ll.second).insert (db::TextRef (text, layout.shape_repository ()));
  
        }

        while (get_record () != sENDEL) { 
          //  @@@ this is not pretty clean ..
          //  overread all unto ENDEL
        } 

      } else if (rec_id == sBOX) {

        GDS2LDPair ld; 

        do {
          rec_id = get_record ();
        } while (rec_id == sELFLAGS || rec_id == sPLEX);
        if (rec_id != sLAYER) {
          error (tl::translate ("LAYER record expected"));
        }
        ld.layer = get_short ();
        if (get_record () != sBOXTYPE) {
          error (tl::translate ("DATATYPE record expected"));
        }
        ld.datatype = get_short ();

        std::pair<bool, unsigned int> ll = open_dl (layout, ld, create_layers);
        if (ll.first) {

          if (get_record () != sXY) {
            error (tl::translate ("XY record expected"));
          }

          GDS2XY *xy = (GDS2XY *) mp_rec_buf;
          GDS2XYToPoint pt_conv;
          db::Box box;
          while (xy < (GDS2XY *) (mp_rec_buf + m_reclen)) {
            box += pt_conv (*xy++);
          }

          if (! box.empty ()) {
            cell.shapes (ll.second).insert (box);
          }

        }

        while (get_record () != sENDEL) { 
          //  @@@ this is not pretty clean ..
          //  overread all unto ENDEL
        } 

      } else if (rec_id == sNODE) {

        //  NODE records are ignored.
        warn (tl::translate ("NODE records are ignored"));
        while (get_record () != sENDEL) { }

      } else {
        error (tl::translate ("Invalid record or data type"));
      }
    
    }

    //  insert all instances collected
    if (! instances.empty ()) {
      cell.insert (instances.begin (), instances.end ());
    }

    m_cellname = "";

  }

  //  check, if the last record is a ENDLIB
  if (rec_id != sENDLIB) {
    error (tl::translate ("ENDLIB record expected"));
  }

}

}

