
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/




#ifndef HDR_dbGDS2Reader
#define HDR_dbGDS2Reader

#include "tlException.h"
#include "tlInternational.h"
#include "tlIntervalMap.h"
#include "tlProgress.h"
#include "tlString.h"

#include "dbLayout.h"

#include <stdio.h>
#include <zlib.h>


namespace db
{

/**
 *  @brief A stream abstraction class
 *
 *  The basic objective of this interface is to provide
 *  the capability to read a block of n bytes into a buffer.
 */

class Stream
{
public:
  /**
   *  @brief Default constructor
   */
  Stream ();

  /**
   *  @brief Default (empty) destructor
   */
  virtual ~Stream () { }

  /** 
   *  @brief This is the outer read method to call
   *  
   *  This implementation obtains data through the 
   *  protected read call and buffers the data accordingly so
   *  a contigous memory block can be returned
   *
   *  @return 0 if not enough data can be obtained
   */
  const char *get (size_t n);

  /**
   *  @brief Obtain the current file position
   */
  size_t pos () const 
  {
    return m_pos;
  }
    
protected:
  /** 
   *  @brief Read a block of n bytes
   *
   *  Read the requested number of bytes or less.
   *  May throw an exception if a read error occures.
   *  
   *  @param b The buffer where to write to
   *  @param n The number of bytes to read (or less)
   *
   *  @return The number of bytes read. Should report 0 on EOF
   */
  virtual size_t read (char *b, size_t n) throw (tl::Exception) = 0;

  void reset_pos ()
  {
    m_pos = 0;
  }

private:
  size_t m_pos;
  char *mp_buffer;
  size_t m_bcap;
  size_t m_blen;
  char *mp_bptr;
};

class FileOpenErrorException
  : public tl::Exception
{
public:
  FileOpenErrorException (const std::string &f, int en)
    : tl::Exception (tl::translate ("Unable to open file: %s (errno=%d)").c_str (), f.c_str (), en)
  { }
};

class FileReadErrorException
  : public tl::Exception
{
public:
  FileReadErrorException (const std::string &f, size_t p, int en)
    : tl::Exception (tl::translate ("Read error on file: %s (position=%d,errno=%d)").c_str (), f.c_str (), long (p), en)
  { }
};

class ZLibReadErrorException
  : public tl::Exception
{
public:
  ZLibReadErrorException (const std::string &f, size_t p, const char *em)
    : tl::Exception (tl::translate ("Read error on file in decompression library: %s (position=%d,message=%s)").c_str (), f.c_str (), long (p), em)
  { }
};

/**
 *  @brief A zlib file adapter class
 *
 *  Implements the stream adapter for a zlib stream
 */
class ZLibStream
  : public Stream
{
public:
  ZLibStream ();
  virtual ~ZLibStream ();

  /**
   *  @brief Open a file with the given path
   *
   *  Opening a file is a prerequisite for reading from the 
   *  object. open() will throw a FileOpenErrorException if 
   *  an error occurs.
   *
   *  @param path The (relative) path of the file to open
   */
  void open (const std::string &path) throw (tl::Exception);

  /**
   *  @brief Close the file
   *
   *  The destructor will automatically close the file.
   */
  void close ();

protected:
  /**
   *  @brief Read from a file 
   *
   *  Implements the basic read method. 
   *  Will throw a FileReadErrorException if an error occurs.
   */
  virtual size_t read (char *b, size_t n) throw (tl::Exception);

private:
  std::string m_source;
  gzFile m_zs;
};

/**
 *  @brief A simple file adapter class
 *
 *  Implements the stream adapter for a simple file
 */
class FileStream
  : public Stream
{
public:
  FileStream ();
  virtual ~FileStream ();

  /**
   *  @brief Open a file with the given path
   *
   *  Opening a file is a prerequisite for reading from the 
   *  object. open() will throw a FileOpenErrorException if 
   *  an error occurs.
   *
   *  @param path The (relative) path of the file to open
   */
  void open (const std::string &path) throw (tl::Exception);

  /**
   *  @brief Close the file
   *
   *  The destructor will automatically close the file.
   */
  void close ();

protected:
  /**
   *  @brief Read from a file 
   *
   *  Implements the basic read method. 
   *  Will throw a FileReadErrorException if an error occurs.
   */
  virtual size_t read (char *b, size_t n) throw (tl::Exception);

private:
  std::string m_source;
  FILE *m_file;
};

class FilePOpenErrorException
  : public tl::Exception
{
public:
  FilePOpenErrorException (const std::string &f, int en)
    : tl::Exception (tl::translate ("Unable to get input from command through pipe: %s (errno=%d)").c_str (), f.c_str (), en)
  { }
};

class FilePReadErrorException
  : public tl::Exception
{
public:
  FilePReadErrorException (const std::string &f, size_t p, int en)
    : tl::Exception (tl::translate ("Read error on pipe from command: %s (position=%d,errno=%d)").c_str (), f.c_str (), long (p), en)
  { }
};

/**
 *  @brief A simple file adapter class
 *
 *  Implements the pipe adapter for a simple file
 */
class FilePipe
  : public Stream
{
public:
  FilePipe ();
  virtual ~FilePipe ();

  /**
   *  @brief Open a stream by connecting with the stdout of a given command
   *
   *  Opening a pipe is a prerequisite for reading from the 
   *  object. open() will throw a FilePOpenErrorException if 
   *  an error occurs - commonly if the command cannot be executed.
   *  This implementation is based on popen ().
   *
   *  @param cmd The command to execute
   */
  void open (const std::string &path) throw (tl::Exception);

  /**
   *  @brief Close the pipe
   *
   *  The destructor will automatically close the pipe.
   */
  void close ();

protected:
  /**
   *  @brief Read from the pipe 
   *
   *  Implements the basic read method. 
   *  Will throw a FilePReadErrorException if an error occurs.
   */
  virtual size_t read (char *b, size_t n) throw (tl::Exception);

private:
  FILE *m_file;
  std::string m_source;
};

/**
 *  @brief Generic base class of GDS2 reader exceptions
 */
class GDS2ReaderException
  : public tl::Exception 
{
public:
  GDS2ReaderException (const std::string &msg, size_t p, size_t n, const std::string &cell)
    : tl::Exception (tl::translate ("%s (position=%ld, record number=%ld, cell=%s)").c_str (), msg.c_str (), p, n, cell.c_str ())
  { }
};

/**
 *  @brief Layer spec string format error exception
 */
class GDS2LayerSpecFormatException
  : public tl::Exception 
{
public:
  GDS2LayerSpecFormatException (const char *s)
    : tl::Exception (tl::translate ("Not a valid GDS2 layer map expression: '..%s' (use '/' to separated layer and datatype, ',' to list numbers for layer or datatype, '-' to create ranges and ';' to concatenate multiple subexpressions)").c_str (), s)
  { }
};

/**
 *  @brief The basic GDS2 layer/datatype type
 */
typedef int gds2_ld_type;

/**
 *  @brief A struct for a GDS2 layer/datatype pair)
 */
struct GDS2LDPair
{
  GDS2LDPair ()
    : layer (0), datatype (0)
  { }

  GDS2LDPair (gds2_ld_type l, gds2_ld_type d)
    : layer (l), datatype (d)
  { }

  static GDS2LDPair invalid () 
  {
    return GDS2LDPair (-1, -1);
  }

  bool operator< (const GDS2LDPair &d) const
  {
    return layer < d.layer || (layer == d.layer && datatype < d.datatype);
  }

  bool operator== (const GDS2LDPair &d) const
  {
    return layer == d.layer && datatype == d.datatype;
  }

  bool operator!= (const GDS2LDPair &d) const
  {
    return ! operator== (d);
  }

  gds2_ld_type layer;
  gds2_ld_type datatype;
};

/**
 *  @brief A GDS2 layer map (GDS2 to logical layers)
 *
 *  A logical layer may consist of several GDS2 layers.
 *  this class allows to map a set of GDS2 
 *  layers to logical layers with a m:n relationship.
 */
class GDS2LayerMap
{
public:
  typedef tl::interval_map<gds2_ld_type, unsigned int> datatype_map;
  typedef tl::interval_map<gds2_ld_type, datatype_map> ld_map;

  /**
   *  @brief The constructor for an empty map
   */
  GDS2LayerMap ();

  /** 
   *  @brief Query a layer mapping
   *
   *  @return A pair telling if the layer is mapped (first=true) and
   *  the logical layer mapped (second) if this is the case.
   */
  std::pair<bool, unsigned int> logical (const GDS2LDPair &p) const;

  /** 
   *  @brief String description for the mapping of a logical layer
   *
   *  @return A string describing the mapping (an empty string if
   *  no mapping is present, a map expression (see below) if there
   *  is any mapping.
   */
  std::string mapping_str (unsigned int l) const;

  /** 
   *  @brief GDS2LDPair describing one mapping of a logical layer
   *
   *  In general, there are more than one GDS2LDPair pairs mapped
   *  to one logical layer. This method will return a single one of
   *  them. It will return the least layer and datatype that matches.
   *  It will return GDS2LDPair::invalid if the layer is not mapped.
   *
   *  @return A GDS2LDPair that is represented by the logical layer
   */
  GDS2LDPair mapping (unsigned int l) const;

  /**
   *  @brief Map a GDS2 ldpair to a logical layer
   */
  void map (const GDS2LDPair &p, unsigned int l);

  /**
   *  @brief Map a range of GDS2 ldpair's to a logical layer
   *
   *  The range is given by two pairs p1,p2. The layers
   *  mapped are [p1.l,p2.l], the datatypes mapped are [p1.d,p2.d].
   */
  void map (const GDS2LDPair &p1, const GDS2LDPair &p2, unsigned int l);

  /** 
   *  @brief Map a range given by a string expression to a logical layer
   * 
   *  The string expression is constructed using the syntax: 
   *  "list[/list][;..]" for layer/datatype pairs. "list" is a 
   *  sequence of numbers, separated by comma values or a range 
   *  separated by a hyphen. Examples are: "1/2", "1-5/0", "1,2,5/0",
   *  "1/5;5/6".
   *  This method will throw a GDS2LayerSpecFormatException if
   *  something is wrong with the format string
   */
  void map (const std::string &expr, unsigned int l) throw (GDS2LayerSpecFormatException);

private:
  ld_map m_ld_map;

  void insert (const GDS2LDPair &p1, const GDS2LDPair &p2, unsigned int);
};

/**
 *  @brief The GDS2 format stream reader
 */
class GDS2Reader
{
public: 
  /**
   *  @brief Construct a stream reader object
   *
   *  @param s The stream object from which to read stream data from
   */
  GDS2Reader (Stream &s);

  /**  
   *  @brief Destructor
   */
  ~GDS2Reader ();

  /** 
   *  @brief The basic read method 
   *
   *  This method will read the stream data and translate this to
   *  insert calls into the layout object. This will not do much
   *  on the layout object beside inserting the objects.
   *  It can be passed a layer map which tells which
   *  GDS2 layer(s) to read on which logical layers.
   *  In addition, a flag can be passed that tells wheter to create 
   *  new layers. The returned map will contain all layers, the passed
   *  ones and the newly created ones.
   *
   *  @param layout The layout object to write to
   *  @param map The GDS2LayerMap object
   *  @param create true, if new layers should be created
   *  @return The GDS2LayerMap object that tells where which layer was loaded
   */
  const GDS2LayerMap &read (db::Layout &layout, const GDS2LayerMap &map, bool create = false) throw (tl::Exception);

  /** 
   *  @brief The basic read method (without mapping)
   *
   *  This method will read the stream data and translate this to
   *  insert calls into the layout object. This will not do much
   *  on the layout object beside inserting the objects.
   *  This version will read all input layers and return a map
   *  which tells which GDS2 layer has been read into which logical
   *  layer.
   *
   *  @param layout The layout object to write to
   *  @return The GDS2LayerMap object
   */
  const GDS2LayerMap &read (db::Layout &layout) throw (tl::Exception);

  /**
   *  @brief Accessor method to the database unit
   */
  double dbu () const { return m_dbu; }

  /**
   *  @brief Accessor method to the database unit in user units
   */
  double dbuu () const { return m_dbuu; }

  /**
   *  @brief Accessor method to the library name
   */
  const std::string &libname () const { return m_libname; }

private:
  Stream &m_stream;
  GDS2LayerMap m_layer_map;
  size_t m_recnum;
  size_t m_reclen;
  size_t m_recptr;
  unsigned char *mp_rec_buf;
  tl::string m_string_buf;
  tl::string m_cellname;
  db::Layout::cell_index_type m_cellindex;
  std::string m_libname;
  double m_dbu, m_dbuu;
  tl::Progress m_progress;

  void do_read (db::Layout &layout, bool create_layers) throw (tl::Exception);

  void error (const std::string &txt) throw (tl::Exception);
  void warn (const std::string &txt);

  const char *get_string () throw (tl::Exception);
  void get_string (tl::string &s) const throw (tl::Exception);
  int get_int () throw (tl::Exception);
  short get_short () throw (tl::Exception);
  double get_double () throw (tl::Exception);
  short get_record () throw (tl::Exception);
  std::pair <bool, unsigned int> open_dl (db::Layout &layout, const GDS2LDPair &dl, bool create) throw (tl::Exception);
};

}

#endif

