
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "edtInstPropertiesPage.h"
#include "edtUtils.h"
#include "layLayoutView.h"
#include "tlString.h"

#include <string>

namespace edt
{


// -------------------------------------------------------------------------
//  InstPropertiesPage implementation

InstPropertiesPage::InstPropertiesPage (edt::InstService *service, QWidget *parent)
  : lay::PropertiesPage (parent, service), mp_service (service)
{
  m_pos = service->selection ().begin ();
  m_index = 0;
  mp_service->clear_highlights ();

  setupUi (this);
  connect (inst_pb, SIGNAL (clicked ()), this, SLOT (show_inst ()));

  for (unsigned int r = 0; r < 8; ++r) {
    orient_cmb->addItem (db::fixpoint_trans (r).to_string ().c_str ());
  }
}

InstPropertiesPage::~InstPropertiesPage ()
{
  mp_service->restore_highlights ();
}

void 
InstPropertiesPage::back ()
{
  m_pos = mp_service->selection ().end ();
  m_index = mp_service->selection ().size ();
}

void 
InstPropertiesPage::front ()
{
  m_pos = mp_service->selection ().begin ();
  m_index = 0;
}

bool 
InstPropertiesPage::at_begin () const
{
  return (m_pos == mp_service->selection ().begin ());
}

bool 
InstPropertiesPage::at_end () const
{
  return (m_pos == mp_service->selection ().end ());
}

void 
InstPropertiesPage::operator-- ()
{
  --m_pos;
  --m_index;
}

void 
InstPropertiesPage::operator++ ()
{
  ++m_pos;
  ++m_index;
}

void 
InstPropertiesPage::leave ()
{
  mp_service->clear_highlights ();
}

void 
InstPropertiesPage::update ()
{
  mp_service->highlight (m_index);

  const lay::CellView &cv = mp_service->view ()->cellview (m_pos->cv_index);
  double dbu = cv->layout ().dbu ();

  cell_name_le->setText (cv->layout ().cell_name (m_pos->cell_inst->object ().cell_index ()));

  db::Point rowv, columnv;
  unsigned long rows, columns;

  if (m_pos->cell_inst->is_regular_array (rowv, columnv, rows, columns)) {

    array_grp->setChecked (true);
    rows_le->setText (tl::to_string (rows).c_str ());
    columns_le->setText (tl::to_string (columns).c_str ());
    row_x_le->setText (tl::micron_to_string (rowv.x () * dbu).c_str ());
    row_y_le->setText (tl::micron_to_string (rowv.y () * dbu).c_str ());
    column_x_le->setText (tl::micron_to_string (columnv.x () * dbu).c_str ());
    column_y_le->setText (tl::micron_to_string (columnv.y () * dbu).c_str ());

    db::Trans t (m_pos->cell_inst->global_trans ());
    pos_x_le->setText (tl::micron_to_string (t.disp ().x () * dbu).c_str ());
    pos_y_le->setText (tl::micron_to_string (t.disp ().y () * dbu).c_str ());
    orient_cmb->setCurrentIndex (t.rot ());
    mag_le->setText (tl::to_string (m_pos->cell_inst->mag ()).c_str ());

    //  from the actual position vs. base position try to compute the array indices
    db::Trans t2 (*m_pos->array_inst);

    unsigned long row, column;
    if (rowv.x () != 0) {
      row = (t2.disp ().x () - t.disp ().x ()) / rowv.x (); 
    } else if (rowv.y () != 0) {
      row = (t2.disp ().y () - t.disp ().y ()) / rowv.y (); 
    } else {
      row = 0;
    }
    if (columnv.x () != 0) {
      column = (t2.disp ().x () - t.disp ().x ()) / columnv.x (); 
    } else if (columnv.y () != 0) {
      column = (t2.disp ().y () - t.disp ().y ()) / columnv.y (); 
    } else {
      column = 0;
    }
    inst_lbl->setText (tl::sprintf (tl::translate ("This is instance [%d,%d] of array with").c_str (), row, column).c_str ());

  } else {

    array_grp->setChecked (false);
    rows_le->setText ("");
    columns_le->setText ("");
    row_x_le->setText ("");
    row_y_le->setText ("");
    column_x_le->setText ("");
    column_y_le->setText ("");

    db::Trans t (*m_pos->array_inst);
    pos_x_le->setText (tl::micron_to_string (t.disp ().x () * dbu).c_str ());
    pos_y_le->setText (tl::micron_to_string (t.disp ().y () * dbu).c_str ());
    orient_cmb->setCurrentIndex (t.rot ());
    mag_le->setText (tl::to_string (m_pos->cell_inst->mag ()).c_str ());
    inst_lbl->setText ("");

  }

}

void
InstPropertiesPage::show_inst ()
{
  InstantiationForm inst_form (this);
  inst_form.show (mp_service->view (), m_pos->cv_index, m_pos->path);
}

bool 
InstPropertiesPage::readonly ()
{

  return true; // @@@

}

void 
InstPropertiesPage::apply ()
{

  //  @@@

}

}

