
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "edtPropertiesPages.h"
#include "edtUtils.h"
#include "layLayoutView.h"

#include <string>

namespace edt
{

// -------------------------------------------------------------------------
//  ShapePropertiesPage implementation

ShapePropertiesPage::ShapePropertiesPage (edt::Service *service, QWidget *parent)
  : lay::PropertiesPage (parent, service), mp_service (service)
{
  m_pos = service->selection ().begin ();
  m_index = 0;
  mp_service->clear_highlights ();
}

ShapePropertiesPage::~ShapePropertiesPage ()
{
  mp_service->restore_highlights ();
}

void 
ShapePropertiesPage::back ()
{
  m_pos = mp_service->selection ().end ();
  m_index = mp_service->selection ().size ();
}

void 
ShapePropertiesPage::front ()
{
  m_pos = mp_service->selection ().begin ();
  m_index = 0;
}

bool 
ShapePropertiesPage::at_begin () const
{
  return (m_pos == mp_service->selection ().begin ());
}

bool 
ShapePropertiesPage::at_end () const
{
  return (m_pos == mp_service->selection ().end ());
}

void 
ShapePropertiesPage::operator-- ()
{
  --m_pos;
  --m_index;
}

void 
ShapePropertiesPage::operator++ ()
{
  ++m_pos;
  ++m_index;
}

void 
ShapePropertiesPage::leave ()
{
  mp_service->clear_highlights ();
}

void 
ShapePropertiesPage::update ()
{
  mp_service->highlight (m_index);

  const lay::CellView &cv = mp_service->view ()->cellview (m_pos->cv_index);
  double dbu = cv->layout ().dbu ();

  do_update (m_pos->shape, dbu);
}

void
ShapePropertiesPage::show_inst ()
{
  InstantiationForm inst_form (this);
  inst_form.show (mp_service->view (), m_pos->cv_index, m_pos->path);
}

// -------------------------------------------------------------------------
//  PolygonPropertiesPage implementation

PolygonPropertiesPage::PolygonPropertiesPage (edt::Service *service, QWidget *parent)
  : ShapePropertiesPage (service, parent)
{
  setupUi (this);

  connect (inst_pb, SIGNAL (clicked ()), this, SLOT (show_inst ()));
}

void 
PolygonPropertiesPage::do_update (const db::Shape &shape, double dbu)
{
  db::Polygon poly;
  shape.polygon (poly);

  std::string ptlist;
  ptlist.reserve (4096);

  bool first = true;
  for (db::Polygon::polygon_contour_iterator pt = poly.begin_hull (); pt != poly.end_hull (); ++pt) {
    if (! first) {
      ptlist += "\n";
    } else {
      first = false;
    }
    ptlist += tl::micron_to_string ((*pt).x () * dbu) + "\t" + tl::micron_to_string ((*pt).y () * dbu);
  }
  pointListEdit->setText (ptlist.c_str ());
}

bool 
PolygonPropertiesPage::readonly ()
{

  return true; // @@@

}

void 
PolygonPropertiesPage::apply ()
{

  //  @@@

}

// -------------------------------------------------------------------------
//  BoxPropertiesPage implementation

BoxPropertiesPage::BoxPropertiesPage (edt::Service *service, QWidget *parent)
  : ShapePropertiesPage (service, parent)
{
  setupUi (this);

  connect (inst_pb, SIGNAL (clicked ()), this, SLOT (show_inst ()));
}

void 
BoxPropertiesPage::do_update (const db::Shape &shape, double dbu)
{
  db::Box box;
  shape.box (box);

  x1_le->setText (tl::micron_to_string (box.left () * dbu).c_str ());
  y1_le->setText (tl::micron_to_string (box.bottom () * dbu).c_str ());
  x2_le->setText (tl::micron_to_string (box.right () * dbu).c_str ());
  y2_le->setText (tl::micron_to_string (box.top () * dbu).c_str ());
  w_le->setText (tl::micron_to_string (box.width () * dbu).c_str ());
  h_le->setText (tl::micron_to_string (box.height () * dbu).c_str ());
}

bool 
BoxPropertiesPage::readonly ()
{

  return true; // @@@

}

void 
BoxPropertiesPage::apply ()
{

  //  @@@

}

// -------------------------------------------------------------------------
//  TextPropertiesPage implementation

TextPropertiesPage::TextPropertiesPage (edt::Service *service, QWidget *parent)
  : ShapePropertiesPage (service, parent)
{
  setupUi (this);

  connect (inst_pb, SIGNAL (clicked ()), this, SLOT (show_inst ()));
}

void 
TextPropertiesPage::do_update (const db::Shape &shape, double dbu)
{
  db::Text text;
  shape.text (text);

  text_le->setText (text.string ().c_str ());
  x_le->setText (tl::micron_to_string (text.trans ().disp ().x () * dbu).c_str ());
  y_le->setText (tl::micron_to_string (text.trans ().disp ().y () * dbu).c_str ());
}

bool 
TextPropertiesPage::readonly ()
{

  return true; // @@@

}

void 
TextPropertiesPage::apply ()
{

  //  @@@

}

// -------------------------------------------------------------------------
//  PathPropertiesPage implementation

PathPropertiesPage::PathPropertiesPage (edt::Service *service, QWidget *parent)
  : ShapePropertiesPage (service, parent)
{
  setupUi (this);

  connect (inst_pb, SIGNAL (clicked ()), this, SLOT (show_inst ()));
}

void 
PathPropertiesPage::do_update (const db::Shape &shape, double dbu)
{
  db::Path path;
  shape.path (path);

  std::string ptlist;
  ptlist.reserve (4096);

  bool first = true;
  for (db::Path::iterator pt = path.begin (); pt != path.end (); ++pt) {
    if (! first) {
      ptlist += "\n";
    } else {
      first = false;
    }
    ptlist += tl::micron_to_string ((*pt).x () * dbu) + "\t" + tl::micron_to_string ((*pt).y () * dbu);
  }
  ptlist_le->setText (ptlist.c_str ());

  width_le->setText (tl::micron_to_string (path.width ()).c_str ());
  start_ext_le->setText (tl::micron_to_string (path.extensions ().first).c_str ());
  end_ext_le->setText (tl::micron_to_string (path.extensions ().second).c_str ());
}

bool 
PathPropertiesPage::readonly ()
{

  return true; // @@@

}

void 
PathPropertiesPage::apply ()
{

  //  @@@

}

}

