
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "edtUtils.h"
#include "layCellView.h"
#include "layLayoutView.h"

#include <dbLayout.h>

namespace edt {

// ----------------------------------------------------------------------
//  edt::Finder implementation

Finder::Finder (bool point_mode)
  : m_min_level (0), m_max_level (0),
    mp_layout (0), m_point_mode (point_mode)
{
  m_distance = std::numeric_limits<double>::max ();
}

Finder::~Finder ()
{
  //  .. nothing yet ..
}

bool 
Finder::closer (double d)
{
  if (d <= m_distance) {
    m_distance = d;
    return true;
  } else {
    return false;
  }
}

void 
Finder::start (lay::LayoutView *view, unsigned int cv_index, const db::Layout &layout, const db::Box &region, const db::Cell &cell, int min_level, int max_level, int layer)
{
  m_layer = (unsigned int) layer;
  m_region = region; 
  mp_layout = &layout;
  mp_view = view;
  m_cv_index = cv_index;
  m_min_level = min_level;
  m_max_level = max_level;

  if (layer >= 0) {
    m_box_convert = db::box_convert <db::CellInst> (layout, (unsigned int) layer);
    m_cell_box_convert = db::box_convert <db::Cell> ((unsigned int) layer);
  } else {
    m_box_convert = db::box_convert <db::CellInst> (layout);
    m_cell_box_convert = db::box_convert <db::Cell> ();
  }

  m_path.erase (m_path.begin (), m_path.end ());

  db::MagTrans t;
  do_find (cell, 0, t);
}

void
Finder::test_edge (const db::DEdge &edg, double &distance, bool &match)
{
  //  we hit the region with the edge end points - take the closest vertex
  if (m_region.contains (edg.p1 ()) || m_region.contains (edg.p2 ())) {
    double d = std::min (edg.p1 ().distance (m_region.center ()), 
                         edg.p2 ().distance (m_region.center ()));
    if (! match || d < distance) {
      distance = d;
    }
    match = true;
  }
  
  //  if the edge cuts through the active region: test the
  //  edge as a whole
  if (edg.clipped (m_region).first) {
    double d = edg.distance_abs (m_region.center ());
    if (! match || d < distance) {
      distance = d;
    }
    match = true;
  }
}

void
Finder::do_find (const db::Cell &cell, int level, const db::MagTrans &t)
{
  if (level >= m_max_level || ! (t * m_cell_box_convert (cell)).touches (m_region) || mp_view->is_cell_hidden (cell.cell_index (), m_cv_index)) {
    return;
  }

  db::Box touch_box (t.inverted () * m_region); 

  if (level >= m_min_level) {
    visit_cell (cell, touch_box, t, level);
  }

  db::Cell::touching_iterator inst = cell.begin_touching (touch_box); 
  while (! inst.at_end ()) {

    const db::CellInstArray &cell_inst = *inst;
    for (db::CellInstArray::iterator p = cell_inst.begin_touching (touch_box, m_box_convert); ! p.at_end (); ++p) {

      m_path.push_back (edt::InstElement (cell_inst, p));

      do_find (mp_layout->cell (cell_inst.object ().cell_index ()), 
               level + 1,
               t * db::MagTrans (*p, cell_inst.mag ()));

      m_path.pop_back ();

    }

    ++inst;

  }

}

// ----------------------------------------------------------------------
//  edt::InstantiationForm implementation

InstantiationForm::InstantiationForm (QWidget *parent)
  : QDialog (parent)
{
  Ui::InstantiationForm::setupUi (this);
}

void 
InstantiationForm::show (lay::LayoutView *view, unsigned int cv_index, const tl::vector<edt::InstElement> &path)
{
  const lay::CellView &cv = view->cellview (cv_index);
  double dbu = cv->layout ().dbu ();

  layout_le->setText (cv->name ().c_str ());

  std::string text;
  text.reserve (4096);

  text += cv->layout ().cell_name (cv.cell_index ());

  for (tl::vector<edt::InstElement>::const_iterator p = path.begin (); p != path.end (); ++p) {

    text += "\n";

    //  build the instance information from the path

    db::Trans trans = *(p->array_inst);
    std::string line;
    line += cv->layout ().cell_name (p->cell_inst->object ().cell_index ());
    line += "\tat ";
    line += tl::micron_to_string (trans.disp ().x () * dbu);
    line += ",";
    line += tl::micron_to_string (trans.disp ().y () * dbu);

    line += " ";
    line += trans.fp_trans ().to_string ();

    double f = 1.0;
    if (p->cell_inst->is_mag ()) {
      f = p->cell_inst->mag ();
    }
    if (fabs (f - 1.0) > 1e-6) {
      line += tl::sprintf (" *%g", f);
    }

    text += line;
  }

  text_edit->setText (text.c_str ());

  exec ();
}

}

