
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <inttypes.h>

#include <string>
#include <algorithm>

#include "EditStipplesFormImpl.h"
#include "SelectStippleFormImpl.h"
#include "layLayerControlPanel.h"
#include "layColorPalette.h"
#include "layViewOp.h"
#include "layLayoutCanvas.h"
#include "tlInternational.h"

#include <QtGui/QVBoxLayout>
#include <QtGui/QHBoxLayout>
#include <QtGui/QImage>
#include <QtGui/QInputDialog>
#include <QtGui/QColorDialog>
#include <QtGui/QLineEdit>
#include <QtGui/QCheckBox>
#include <QtGui/QBitmap>
#include <QtGui/QPixmap>
#include <QtGui/QTreeWidget>
#include <QtGui/QTreeWidgetItem>
#include <QtGui/QHeaderView>
#include <QtGui/QMenu>
#include <QtGui/QToolButton>

namespace lay
{

// --------------------------------------------------------------------
//  LCPRemitter implementation

LCPRemitter::LCPRemitter (int index, QObject *parent, const char *name)
  : QObject (parent, name),
    m_index (index)
{
  // .. 
}

void 
LCPRemitter::the_slot ()
{
  emit the_signal (m_index);
}

// --------------------------------------------------------------------
//  LCPActiveLabel implementation

LCPActiveLabel::LCPActiveLabel (int index, QWidget *parent)
  : LabelWithBackground (parent), m_index (index), m_grabbed (false)
{
  setFrameStyle (QFrame::Panel | QFrame::Raised);
  setBackgroundRole (QPalette::Window);
}

void 
LCPActiveLabel::mousePressEvent (QMouseEvent *e)
{
  if (! m_grabbed && e->button () == Qt::LeftButton) {
    setFrameShadow (QFrame::Sunken);
    m_grabbed = true;
    grabMouse ();
  }
}

void 
LCPActiveLabel::mouseReleaseEvent (QMouseEvent *e)
{
  if (m_grabbed) {
   
    setFrameShadow (QFrame::Raised);
    releaseMouse ();
    m_grabbed = false;

    if (e->button () == Qt::LeftButton && rect ().contains (e->pos ())) {
      emit clicked (m_index);
    }

  }
}

// --------------------------------------------------------------------
//  LCPDitherPalette implementation

LCPDitherPalette::LCPDitherPalette (QWidget *parent, const lay::DitherPattern &pattern, const char *name)
  : QFrame (parent, name), m_pattern (pattern)
{
  int n = 0;
  QVBoxLayout *l = new QVBoxLayout (this);
  l->setMargin (0);
  l->setSpacing (0);

  for (unsigned int i = 0; i < 4; ++i) {

    QFrame *f = new QFrame (this);
    l->addWidget (f);
    QHBoxLayout *ll = new QHBoxLayout (f);
    ll->setMargin (0);
    ll->setSpacing (0);

    for (unsigned int j = 0; j < 4; ++j) {
      
      LCPActiveLabel *b = new LCPActiveLabel (n, f);

      b->setMinimumSize (28, 28);
      b->setAlignment (Qt::AlignCenter);
      b->setLineWidth (1);
      b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));

      color_t color0 = b->paletteBackgroundColor ().rgb ();
      color_t color1 = b->paletteForegroundColor ().rgb ();

      const unsigned int h = 24;
      const unsigned int w = 24;

      QImage image (w, h, 32);
      const uint32_t *dp = m_pattern.pattern (n);
      for (unsigned int l = 0; l < h; ++l, ++dp) {
        uint32_t m = *dp;
        if (l == 0 || l == h - 1) { 
          m |= ((1 << w) - 1);
        } else {
          m |= ((1 << (w - 1)) | 1);
        }
        color_t *pt = (color_t *) image.scanLine (h - 1 - l);
        for (unsigned int b = 0; b < w; ++b) {
          *pt++ = (m & 1) ? color1 : color0;
          m >>= 1;
        }
      }
      b->setPixmap (QPixmap (image));

      connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));

      ll->addWidget (b);

      ++n;

    }

  }

  {
    QFrame *f = new QFrame (this);
    l->addWidget (f);
    QHBoxLayout *ll = new QHBoxLayout (f);
    ll->setMargin (0);
    ll->setSpacing (0);

    //  More pattern 
    LCPActiveLabel *b;
    b = new LCPActiveLabel (-2, f);
    b->setFrameStyle (QFrame::Panel | QFrame::Raised);
    b->setLineWidth (1);
    b->setText (tl::translate ("More ..").c_str ());
    b->setBackgroundMode (Qt::PaletteButton);
    connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
    ll->addWidget (b);

    //  Edit pattern 
    b = new LCPActiveLabel (-1, f);
    b->setFrameStyle (QFrame::Panel | QFrame::Raised);
    b->setLineWidth (1);
    b->setText (tl::translate ("Edit ..").c_str ());
    b->setBackgroundMode (Qt::PaletteButton);
    connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
    ll->addWidget (b);
  }

}

void 
LCPDitherPalette::set_dither_pattern (const lay::DitherPattern &pattern)
{
  m_pattern = pattern;
}

void 
LCPDitherPalette::button_clicked (int index)
{
  if (index == -1) {
    
    //  edit pattern
    EditStipplesFormImpl stipples_form (0, m_pattern);
    if (stipples_form.exec () && stipples_form.pattern () != m_pattern) {
      set_dither_pattern (stipples_form.pattern ());
      emit pattern_changed (m_pattern);
    }

  } else if (index == -2) {
    
    //  edit pattern
    SelectStippleFormImpl stipples_form (0, m_pattern);
    if (stipples_form.exec () && stipples_form.selected () >= 0) {
      emit dither_selected ((unsigned int) stipples_form.selected ());
    }

  } else {
    emit dither_selected ((unsigned int) index);
  }
}

// --------------------------------------------------------------------
//  LCPVisibilityPalette implementation

LCPVisibilityPalette::LCPVisibilityPalette (QWidget *parent, const char *name)
  : QFrame (parent, name)
{
  QVBoxLayout *l = new QVBoxLayout (this);
  l->setMargin (0);
  l->setSpacing (0);

  QFrame *f;
  LCPActiveLabel *b;
  QHBoxLayout *ll;

  f = new QFrame (this);
  l->addWidget (f);
  ll = new QHBoxLayout (f);
  ll->setMargin (0);
  ll->setSpacing (0);

  b = new LCPActiveLabel (0, f);
  b->setMinimumSize (50, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("Show").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);

  b = new LCPActiveLabel (1, f);
  b->setMinimumSize (50, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("Hide").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);

  f = new QFrame (this);
  l->addWidget (f);
  ll = new QHBoxLayout (f);
  ll->setMargin (0);
  ll->setSpacing (0);

  b = new LCPActiveLabel (2, f);
  b->setMinimumSize (50, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("Transp.").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);

  b = new LCPActiveLabel (3, f);
  b->setMinimumSize (50, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("Opaque").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);
}

void 
LCPVisibilityPalette::button_clicked (int index)
{
  if (index == 0) {
    emit visibility_change (true);
  } else if (index == 1) {
    emit visibility_change (false);
  } else if (index == 2) {
    emit transparency_change (true);
  } else if (index == 3) {
    emit transparency_change (false);
  }
}

// --------------------------------------------------------------------
//  LCPAnimationPalette implementation

LCPAnimationPalette::LCPAnimationPalette (QWidget *parent, const char *name)
  : QFrame (parent, name)
{
  QVBoxLayout *l = new QVBoxLayout (this);
  l->setMargin (0);
  l->setSpacing (0);

  QFrame *f;
  LCPActiveLabel *b;
  QHBoxLayout *ll;

  f = new QFrame (this);
  l->addWidget (f);
  ll = new QHBoxLayout (f);
  ll->setMargin (0);
  ll->setSpacing (0);

  b = new LCPActiveLabel (0, f);
  b->setMinimumSize (50, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("Static").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);

  b = new LCPActiveLabel (1, f);
  b->setMinimumSize (50, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("Scroll").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);

  f = new QFrame (this);
  l->addWidget (f);
  ll = new QHBoxLayout (f);
  ll->setMargin (0);
  ll->setSpacing (0);

  b = new LCPActiveLabel (2, f);
  b->setMinimumSize (50, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("Blink").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);

  b = new LCPActiveLabel (3, f);
  b->setMinimumSize (50, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("/Blink").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);
}

void 
LCPAnimationPalette::button_clicked (int index)
{
  emit animation_selected (index);
}

// --------------------------------------------------------------------
//  LCPStylePalette implementation

LCPStylePalette::LCPStylePalette (QWidget *parent, const char *name)
  : QFrame (parent, name)
{
  QVBoxLayout *l = new QVBoxLayout (this);
  l->setMargin (0);
  l->setSpacing (0);

  QFrame *f;
  LCPActiveLabel *b;
  QHBoxLayout *ll;

  f = new QFrame (this);
  l->addWidget (f);
  ll = new QHBoxLayout (f);
  ll->setMargin (0);
  ll->setSpacing (0);

  b = new LCPActiveLabel (0, f);
  b->setMinimumSize (25, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("0px").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);

  b = new LCPActiveLabel (1, f);
  b->setMinimumSize (25, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("1px").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);

  b = new LCPActiveLabel (2, f);
  b->setMinimumSize (25, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText ("2px");
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);

  b = new LCPActiveLabel (3, f);
  b->setMinimumSize (25, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("3px").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);

  f = new QFrame (this);
  l->addWidget (f);
  ll = new QHBoxLayout (f);
  ll->setMargin (0);
  ll->setSpacing (0);

  b = new LCPActiveLabel (100, f);
  b->setMinimumSize (50, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("Simple").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);

  b = new LCPActiveLabel (101, f);
  b->setMinimumSize (50, 16);
  b->setAlignment (Qt::AlignCenter);
  b->setLineWidth (1);
  b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
  b->setText (tl::translate ("Marked").c_str ());
  connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
  ll->addWidget (b);
}

void 
LCPStylePalette::button_clicked (int index)
{
  if (index >= 0 && index < 16) {
    emit width_selected (index);
  } else if (index == 100) {
    emit marked_selected (false);
  } else if (index == 101) {
    emit marked_selected (true);
  }
}

// --------------------------------------------------------------------
//  LCPColorPalette implementation

LCPColorPalette::LCPColorPalette (QWidget *parent, const char *name)
  : QFrame (parent, name)
{
  QVBoxLayout *l = new QVBoxLayout (this);
  l->setMargin (0);
  l->setSpacing (0);
  for (unsigned int i = 0; i < 6; ++i) {

    QFrame *f = new QFrame (this);
    l->addWidget (f);
    QHBoxLayout *ll = new QHBoxLayout (f);
    ll->setMargin (0);
    ll->setSpacing (0);

    for (unsigned int j = 0; j < 7; ++j) {
      
      unsigned int n = j * 6 + i;

      LCPActiveLabel *b = new LCPActiveLabel (n, f);
      b->setMinimumSize (16, 16);
      b->setLineWidth (1);
      b->setSizePolicy (QSizePolicy (QSizePolicy::Ignored, QSizePolicy::Ignored, 0, 0));
      b->setText ("");
      QPalette pl;
      pl.setColor (QColorGroup::Window, colorByIndex (n));
      b->setPalette (pl);
      connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));

      ll->addWidget (b);

    }

  }

  {
    //  More colors ..
    LCPActiveLabel *b = new LCPActiveLabel (-1, this);
    b->setFrameStyle (QFrame::Panel | QFrame::Raised);
    b->setLineWidth (1);
    b->setText (tl::translate ("More Colors ..").c_str ());
    b->setBackgroundMode (Qt::PaletteButton);
    connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
    l->addWidget (b);
  }

  {
    //  darker and brighter colors
    QFrame *f = new QFrame (this);
    l->addWidget (f);
    QHBoxLayout *ll = new QHBoxLayout (f);
    ll->setMargin (0);
    ll->setSpacing (0);

    LCPActiveLabel *b;

    b = new LCPActiveLabel (-10, f);
    b->setFrameStyle (QFrame::Panel | QFrame::Raised);
    b->setLineWidth (1);
    b->setText (tl::translate ("Dark").c_str ());
    b->setBackgroundMode (Qt::PaletteButton);
    connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
    ll->addWidget (b);

    b = new LCPActiveLabel (-11, f);
    b->setFrameStyle (QFrame::Panel | QFrame::Raised);
    b->setLineWidth (1);
    b->setText (tl::translate ("Bright").c_str ());
    b->setBackgroundMode (Qt::PaletteButton);
    connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
    ll->addWidget (b);

    b = new LCPActiveLabel (-12, f);
    b->setFrameStyle (QFrame::Panel | QFrame::Raised);
    b->setLineWidth (1);
    b->setText (tl::translate ("Reset").c_str ());
    b->setBackgroundMode (Qt::PaletteButton);
    connect (b, SIGNAL (clicked (int)), this, SLOT (button_clicked (int)));
    ll->addWidget (b);

  }

}

QColor 
LCPColorPalette::colorByIndex (int index)
{
  return QColor (lay::ColorPalette::color_by_index (index));
}

void 
LCPColorPalette::button_clicked (int index)
{
  if (index >= 0) {
    emit color_selected (colorByIndex (index));
  } else if (index == -1) {
    QColor c = QColorDialog::getColor ();
    if (c.isValid ()) {
      emit color_selected (c);
    }
  } else if (index == -10) {
    emit color_brightness_selected (-16); // darker
  } else if (index == -11) {
    emit color_brightness_selected (16); // brighter
  } else if (index == -12) {
    emit color_brightness_selected (0); // brighter
  }
}

// --------------------------------------------------------------------
//  A specialization of the TreeWidget that allows us to control
//  sizeHint which otherwise is pretty large (around 100 pixel per column)

class LCPTreeWidget : public QTreeWidget
{
public:
  LCPTreeWidget (QWidget *parent)
    : QTreeWidget (parent) 
  {
    //  .. nothing yet ..
  }

  virtual QSize sizeHint () const
  {
    return QSize (0, 0);
  }
};

// --------------------------------------------------------------------
//  LayerControlPanel implementation

LayerControlPanel::LayerControlPanel (lay::LayoutView *view, db::Manager *manager, QWidget *parent, const char *name)
  : QFrame (parent, name), db::Object (manager),
    mp_view (view), m_enable_cb (true)
{
  setSizePolicy (QSizePolicy (QSizePolicy::Minimum, QSizePolicy::Preferred, 0, 0));

  mp_layer_frame = new QFrame (this);
  QVBoxLayout *l = new QVBoxLayout (mp_layer_frame);
  l->setMargin (0);
  l->setSpacing (0);

  mp_layer_list = new LCPTreeWidget (mp_layer_frame);
  l->addWidget (mp_layer_list);
  connect (mp_layer_list, SIGNAL(itemDoubleClicked (QTreeWidgetItem *, int)), this, SLOT (double_clicked (QTreeWidgetItem *, int)));
  mp_layer_list->setContextMenuPolicy (Qt::CustomContextMenu);
  connect (mp_layer_list, SIGNAL(customContextMenuRequested (const QPoint &)), this, SLOT (context_menu (const QPoint &)));
  mp_layer_list->setColumnCount (2);
  mp_layer_list->setSortingEnabled (false);
  mp_layer_list->header ()->hide ();
  mp_layer_list->setSelectionMode (QTreeWidget::ExtendedSelection);
  mp_layer_list->setRootIsDecorated (false);
  mp_layer_list->setIconSize (QSize (32, 16));
  //  Custom resize mode makes the columns as narrow as possible
  mp_layer_list->header ()->setResizeMode (QHeaderView::Custom);

  QFrame *button_frame = new QFrame (mp_layer_frame);
  l->addWidget (button_frame);
  QHBoxLayout *ll = new QHBoxLayout (button_frame);
  ll->setMargin (0);
  ll->setSpacing (0);

  QToolButton *b;

  b = new QToolButton (button_frame);
  b->setIcon (QIcon (":downdown.png"));
  connect (b, SIGNAL (clicked ()), this, SLOT (downdown_clicked ()));
  ll->addWidget (b);

  b = new QToolButton (button_frame);
  b->setIcon (QIcon (":down.png"));
  connect (b, SIGNAL (clicked ()), this, SLOT (down_clicked ()));
  ll->addWidget (b);

  b = new QToolButton (button_frame);
  b->setIcon (QIcon (":up.png"));
  connect (b, SIGNAL (clicked ()), this, SLOT (up_clicked ()));
  ll->addWidget (b);

  b = new QToolButton (button_frame);
  b->setIcon (QIcon (":upup.png"));
  connect (b, SIGNAL (clicked ()), this, SLOT (upup_clicked ()));
  ll->addWidget (b);
  ll->addStretch (0);

  LCPVisibilityPalette *vp = new LCPVisibilityPalette (this);
  add_panel (vp, tl::translate ("Visibility").c_str ());
  connect (vp, SIGNAL(visibility_change (bool)), this, SLOT (visibility_changed (bool)));
  connect (vp, SIGNAL(transparency_change (bool)), this, SLOT (transparency_changed (bool)));

  LCPStylePalette *sp = new LCPStylePalette (this);
  add_panel (sp, tl::translate ("Style").c_str ());
  connect (sp, SIGNAL(width_selected (int)), this, SLOT (width_changed (int)));
  connect (sp, SIGNAL(marked_selected (bool)), this, SLOT (marked_changed (bool)));

  LCPAnimationPalette *ap = new LCPAnimationPalette (this);
  add_panel (ap, tl::translate ("Animation").c_str ());
  connect (ap, SIGNAL(animation_selected (int)), this, SLOT (animation_changed (int)));

  LCPDitherPalette *dp = new LCPDitherPalette (this, mp_view->dither_pattern ());
  mp_dither_palette = dp;
  add_panel (dp, tl::translate ("Stipple").c_str ());
  connect (dp, SIGNAL(dither_selected (unsigned int)), this, SLOT (dither_changed (unsigned int)));
  connect (dp, SIGNAL(pattern_changed (const lay::DitherPattern &)), this, SLOT (dither_pattern_changed (const lay::DitherPattern &)));

  LCPColorPalette *pf = new LCPColorPalette (this);
  add_panel (pf, tl::translate ("Frame color").c_str ());
  connect (pf, SIGNAL(color_selected (QColor)), this, SLOT (frame_color_changed (QColor)));
  connect (pf, SIGNAL(color_brightness_selected (int)), this, SLOT (frame_color_brightness (int)));

  LCPColorPalette *p = new LCPColorPalette (this);
  add_panel (p, tl::translate ("Color").c_str ());
  connect (p, SIGNAL(color_selected (QColor)), this, SLOT (fill_color_changed (QColor)));
  connect (p, SIGNAL(color_brightness_selected (int)), this, SLOT (fill_color_brightness (int)));

  m_needs_update = true;
  m_phase = 0;
  update_content (m_phase);

  mp_view->add_layer_list_observer (*this);
}

LayerControlPanel::~LayerControlPanel ()
{
  mp_view->remove_layer_list_observer (*this);
}

void 
LayerControlPanel::props_to_display (QTreeWidgetItem *item, const lay::LayerProperties &props)
{
  bool animate_visible = true;
  unsigned int di_off = 0;
  if (props.display.animation) {
    if (props.display.animation == 1) {
      // scrolling 
      di_off = ((m_phase % 32) << 16);
    } else if (props.display.animation == 2) {
      // blinking
      animate_visible = ((m_phase & 1) == 0);
    } else {
      // inversly blinking
      animate_visible = ((m_phase & 1) != 0);
    }
  }

  unsigned int w = 32;
  unsigned int h = 16;
    
  if (props.display.visible && animate_visible) {

    lay::Bitmap fill (w, h);
    for (unsigned int i = 1; i < h - 1; ++i) {
      fill.scanline (i) [0] = 0xffffffff;
    }

    int p0 = props.display.width / 2;
    int p1 = (props.display.width - 1) / 2;
    if (p0 < 0) {
      p0 = 0;
    } else if (p0 > 7) {
      p0 = 7;
    }
    if (p1 < 0) {
      p1 = 0;
    } else if (p1 > 7) {
      p1 = 7;
    }

    lay::Bitmap frame (w, h);
    frame.scanline (p0) [0] = 0xffffffff;
    for (unsigned int i = p0 + 1; i < h - 1; ++i) {
      frame.scanline (i) [0] = (0x80000000 >> p0) | (1 << p1);
    }
    frame.scanline (15 - p1) [0] = 0xffffffff;

    lay::Bitmap text (w, h);
    lay::Bitmap vertex (w, h);
    vertex.scanline (7) [0] = 0x00010000;

    lay::ViewOp::Mode mode = lay::ViewOp::Copy;

    QImage image (w, h, 32);
    image.fill (item->treeWidget ()->palette ().color (QColorGroup::Base).rgb ());

    //  create fill 
    mp_view->canvas ()->bitmap_to_image (lay::ViewOp (props.display.eff_fill_color (), mode, props.display.dither_pattern | di_off), fill, &image, w, h);
    //  create frame 
    mp_view->canvas ()->bitmap_to_image (lay::ViewOp (props.display.eff_frame_color (), mode, 0, lay::ViewOp::Rect, props.display.width), frame, &image, w, h);
    //  create text 
    mp_view->canvas ()->bitmap_to_image (lay::ViewOp (props.display.eff_frame_color (), mode, 0), text, &image, w, h);
    //  create vertex 
    mp_view->canvas ()->bitmap_to_image (lay::ViewOp (props.display.eff_frame_color (), mode, 0, lay::ViewOp::Cross, props.display.marked ? 9/*mark size*/ : 0), vertex, &image, w, h);

    item->setIcon (0, QIcon (QPixmap (image)));

  } else {
    item->setIcon (0, QIcon ());
  }

  item->setText (1, props.display_string (mp_view).c_str ());

  QPalette pl (mp_layer_list->palette ());
  QColor c1 = pl.color (QColorGroup::Text);
  QColor cb = pl.color (QColorGroup::Base);
  QColor c0 = QColor ((c1.red () + cb.red ()) / 2, (c1.green () + cb.green ()) / 2, (c1.blue () + cb.blue ()) / 2);

  //  set the color to one with less contrast if there is nothing on this layer here
  bool empty = props.bbox (mp_view).empty ();
  QFont f = item->font (1);
  f.setBold (! empty);
  item->setFont (1, f);
  item->setTextColor (1, empty ? c0 : c1);
}

void 
LayerControlPanel::add_panel (QWidget *panel_widget, const char *text)
{
  panel_widget->hide ();

  QFrame *f = new FrameWithBackground (this);
  QHBoxLayout *l = new QHBoxLayout (f, 3, 0);
  l->setMargin (0);
  l->setSpacing (0);

  f->setFrameStyle (QFrame::Panel | QFrame::Raised);
  f->setLineWidth (1);
  f->setBackgroundRole (QPalette::Highlight);

  QCheckBox *b = new QCheckBox (f);
  l->addWidget (b);

  b->setBackgroundRole (QPalette::Highlight);
  QPalette pl (b->palette ());
  pl.setColor (QColorGroup::Foreground, pl.color (QPalette::Active, QColorGroup::HighlightedText));
  b->setPalette (pl);

  b->setText (text);

  LCPRemitter *e = new LCPRemitter (int (m_tool_panels.size ()), this);
  connect (b, SIGNAL(clicked ()), e, SLOT (the_slot ()));
  connect (e, SIGNAL(the_signal (int)), this, SLOT (panel_button_clicked (int)));

  m_tool_panels.push_back (std::make_pair (f, panel_widget));
}

QSize
LayerControlPanel::sizeHint () const
{
  int w = 0;
  for (std::vector <std::pair <QWidget *, QWidget *> >::const_iterator i = m_tool_panels.begin (); i != m_tool_panels.end (); ++i) {
    w = std::max (std::max (i->first->sizeHint ().width (), i->second->sizeHint ().width ()), w);
  }
  w = std::max (mp_layer_list->sizeHint ().width (), w);
  return QSize (w, 0);
}

void 
LayerControlPanel::resizeEvent (QResizeEvent *re)
{
  rearrange (re->size ().width (), re->size ().height ());
}

void 
LayerControlPanel::resize (int w, int h)
{
  QFrame::resize (w, h);
  rearrange (w, h);
}

void 
LayerControlPanel::setGeometry (int x, int y, int w, int h)
{
  QFrame::setGeometry (x, y, w, h);
  rearrange (w, h);
}

void 
LayerControlPanel::rearrange (int w, int h)
{
  for (std::vector <std::pair <QWidget *, QWidget *> >::iterator i = m_tool_panels.begin (); i != m_tool_panels.end (); ++i) {

    int hh;

    if (i->second->isVisible ()) {
      hh = i->second->sizeHint ().height ();
      h -= hh;
      i->second->setGeometry (0, h, w, hh);
    }

    hh = i->first->sizeHint ().height () - 6;
    h -= hh;
    i->first->setGeometry (0, h, w, hh);

  }

  if (h < 0) {
    mp_layer_frame->hide ();
  } else {
    mp_layer_frame->setGeometry (0, 0, w, h);
    mp_layer_frame->show ();
  }
}

template <class Op>
void 
LayerControlPanel::foreach_selected (const Op &op)
{
  m_enable_cb = false;

  QTreeWidgetItemIterator item (mp_layer_list);
  for (unsigned int i = 0; i < mp_view->layers () && *item; ++i) {

    if (mp_layer_list->isItemSelected (*item)) {

      lay::LayerProperties props = mp_view->properties (i);
      op (props);
      mp_view->set_properties (i, props);

      props_to_display (*item, props);

    }

    ++item;
  }

  m_enable_cb = true;
}

struct SetColor
{
  /** 
   *  @brief set (some) colors of the properties
   *
   *  @param flags Bitmask that defines which colors to change: #0=fill,#1=frame,#2=vertex,#3=text
   *  @param c The color to apply
   */
  SetColor (QColor c, unsigned int flags)
    : m_color (c.rgb ()), m_flags (flags)
  {
    // .. nothing yet ..
  }

  void operator() (lay::LayerProperties &props) const
  {
    if (m_flags & 2) {
      props.display.fill_color = m_color;
      props.display.fill_brightness = 0;
    } 
    if (m_flags & 1) {
      props.display.frame_color = m_color;
      props.display.frame_brightness = 0;
    }
  }

private:
  color_t m_color;
  unsigned int m_flags;
};

void 
LayerControlPanel::fill_color_changed (QColor c)
{
  manager ()->transaction (tl::translate ("Change fill color")); 
  SetColor op (c, 3 /*fill,frame and vertex*/);
  foreach_selected (op);
  manager ()->commit ();

  emit color_changed ();
}

void 
LayerControlPanel::frame_color_changed (QColor c)
{
  manager ()->transaction (tl::translate ("Change frame color")); 
  SetColor op (c, 1 /*frame and vertex*/);
  foreach_selected (op);
  manager ()->commit ();

  emit color_changed ();
}

struct SetBrightness
{
  /** 
   *  @brief set (some) colors of the properties
   *
   *  @param flags Bitmask that defines which colors to change: #0=fill,#1=frame,#2=vertex,#3=text
   *  @param c The color to apply
   */
  SetBrightness (int delta, unsigned int flags)
    : m_delta (delta), m_flags (flags)
  {
    // .. nothing yet ..
  }

  void operator() (lay::LayerProperties &props) const
  {
    if (m_flags & 2) {
      if (m_delta == 0) {
        props.display.fill_brightness = 0;
      } else {
        props.display.fill_brightness += m_delta;
      }
    } 
    if (m_flags & 1) {
      if (m_delta == 0) {
        props.display.frame_brightness = 0;
      } else {
        props.display.frame_brightness += m_delta;
      }
    }
  }

private:
  int m_delta;
  unsigned int m_flags;
};

void 
LayerControlPanel::fill_color_brightness (int delta)
{
  manager ()->transaction (tl::translate ("Change fill color brightness")); 
  SetBrightness op (delta, 3 /*fill,frame and vertex*/);
  foreach_selected (op);
  manager ()->commit ();

  emit color_changed ();
}

void 
LayerControlPanel::frame_color_brightness (int delta)
{
  manager ()->transaction (tl::translate ("Change frame color brightness")); 
  SetBrightness op (delta, 1 /*frame and vertex*/);
  foreach_selected (op);
  manager ()->commit ();

  emit color_changed ();
}

struct SetDither
{
  SetDither (unsigned int di)
    : m_di (di)
  { }

  void operator() (lay::LayerProperties &props) const
  {
    props.display.dither_pattern = m_di;
  }

private:
  unsigned int m_di;
};

void
LayerControlPanel::dither_pattern_changed (const lay::DitherPattern &pattern)
{
  manager ()->transaction (tl::translate ("Edit stipple pattern")); 
  m_enable_cb = false;
  mp_view->set_dither_pattern (pattern);
  refresh_display ();
  m_enable_cb = true;
  manager ()->commit ();
}

void
LayerControlPanel::dither_changed (unsigned int di)
{
  manager ()->transaction (tl::translate ("Set stipple pattern")); 
  SetDither op (di);
  foreach_selected (op);
  manager ()->commit ();
  emit stipple_changed ();
}

struct SetVisible
{
  SetVisible (bool v)
    : m_visible (v)
  { }

  void operator() (lay::LayerProperties &props) const
  {
    props.display.visible = m_visible;
  }

private:
  bool m_visible;
};

void 
LayerControlPanel::visibility_changed (bool visible)
{
  if (visible) {
    manager ()->transaction (tl::translate ("Show layer")); 
  } else {
    manager ()->transaction (tl::translate ("Hide layer")); 
  }

  SetVisible op (visible);
  foreach_selected (op);
  manager ()->commit ();

  emit visibility_changed ();
}

struct SetTransparency
{
  SetTransparency (bool t)
    : m_transparent (t)
  { }

  void operator() (lay::LayerProperties &props) const
  {
    props.display.transparent = m_transparent;
  }

private:
  bool m_transparent;
};

void 
LayerControlPanel::transparency_changed (bool transparent)
{
  manager ()->transaction (tl::translate ("Change transparency")); 
  SetTransparency op (transparent);
  foreach_selected (op);
  manager ()->commit ();

  emit transparency_changed ();
}

struct SetAnimation
{
  SetAnimation (int mode)
    : m_mode (mode)
  { }

  void operator() (lay::LayerProperties &props) const
  {
    props.display.animation = m_mode;
  }

private:
  int m_mode;
};

void 
LayerControlPanel::animation_changed (int mode)
{
  manager ()->transaction (tl::translate ("Change animation mode")); 
  SetAnimation op (mode);
  foreach_selected (op);
  manager ()->commit ();

  emit animation_changed ();
}

struct SetWidth
{
  SetWidth (int w)
    : m_width (w)
  { }

  void operator() (lay::LayerProperties &props) const
  {
    props.display.width = m_width;
  }

private:
  int m_width;
};

void 
LayerControlPanel::width_changed (int width)
{
  manager ()->transaction (tl::translate ("Change line width")); 
  SetWidth op (width);
  foreach_selected (op);
  manager ()->commit ();

  emit width_changed ();
}

struct SetMarked
{
  SetMarked (bool m)
    : m_marked (m)
  { }

  void operator() (lay::LayerProperties &props) const
  {
    props.display.marked = m_marked;
  }

private:
  bool m_marked;
};

void 
LayerControlPanel::marked_changed (bool marked)
{
  manager ()->transaction (tl::translate ("Change marked vertices")); 
  SetMarked op (marked);
  foreach_selected (op);
  manager ()->commit ();

  emit marked_changed ();
}

void 
LayerControlPanel::cm_rename ()
{
  int index = -1;

  QTreeWidgetItemIterator item (mp_layer_list);
  for (int i = 0; i < int (mp_view->layers ()) && *item; ++i) {
    if (*item == mp_layer_list->currentItem ()) {
      index = i;
      break;
    }
    ++item;
  }

  if (index >= 0) {

    lay::LayerProperties props = mp_view->properties (index);

    bool ok = false;
    QString n = QInputDialog::getText (tl::translate ("Rename layer").c_str (),
                                       tl::translate ("Enter new name of layer").c_str (),
                                       QLineEdit::Normal, 
                                       props.display_string (mp_view).c_str (),
                                       &ok);
   
    if (ok) {

      props.display.name = n.ascii ();

      manager ()->transaction (tl::translate ("Rename layer")); 

      mp_view->set_properties (index, props);
      props_to_display (*item, props);

      manager ()->commit ();

    }

  }

}

void 
LayerControlPanel::cm_show ()
{
  manager ()->transaction (tl::translate ("Show layer")); 

  SetVisible op (true);
  foreach_selected (op);

  manager ()->commit ();

  emit visibility_changed ();
}

void 
LayerControlPanel::cm_hide ()
{
  manager ()->transaction (tl::translate ("Hide layer")); 

  SetVisible op (false);
  foreach_selected (op);

  manager ()->commit ();

  emit visibility_changed ();
}

void 
LayerControlPanel::cm_select_all ()
{
  mp_layer_list->selectAll ();
}

void
LayerControlPanel::up_clicked ()
{
  manager ()->transaction (tl::translate ("Move up")); 

  std::vector<lay::LayerProperties> new_props;
  new_props.reserve (mp_view->layers ());

  std::set<unsigned int> selection;

  int src = -1;
  bool changed = false;

  QTreeWidgetItemIterator item (mp_layer_list);
  for (unsigned int i = 0; i < mp_view->layers () && *item; ++i) {

    if (mp_layer_list->isItemSelected (*item)) {
      selection.insert (new_props.size ());
      new_props.push_back (mp_view->properties (i));
      if (src >= 0) {
        changed = true;
      }
    } else {
      if (src >= 0) {
        new_props.push_back (mp_view->properties ((unsigned int) src));
      }
      src = i;
    }

    ++item;
  }

  if (src >= 0) {
    new_props.push_back (mp_view->properties ((unsigned int) src));
  }

  if (changed) {

    m_enable_cb = false;
    for (unsigned int i = 0; i < new_props.size (); ++i) {
      mp_view->set_properties (i, new_props [i]);
    }
    refresh_display (selection);
    m_enable_cb = true;

    emit order_changed ();

  }

  manager ()->commit ();
}

void
LayerControlPanel::upup_clicked ()
{
  manager ()->transaction (tl::translate ("Move fully up")); 

  std::vector<lay::LayerProperties> new_props;
  new_props.reserve (mp_view->layers ());

  std::set<unsigned int> selection;

  bool changed = false;

  QTreeWidgetItemIterator item (mp_layer_list);
  for (unsigned int i = 0; i < mp_view->layers () && *item; ++i) {
    if (mp_layer_list->isItemSelected (*item)) {
      selection.insert (new_props.size ());
      new_props.push_back (mp_view->properties (i));
      changed = true;
    }
    ++item;
  }

  item = QTreeWidgetItemIterator (mp_layer_list);
  for (unsigned int i = 0; i < mp_view->layers () && *item; ++i) {
    if (! mp_layer_list->isItemSelected (*item)) {
      new_props.push_back (mp_view->properties (i));
    }
    ++item;
  }

  if (changed) {

    m_enable_cb = false;
    for (unsigned int i = 0; i < new_props.size (); ++i) {
      mp_view->set_properties (i, new_props [i]);
    }
    refresh_display (selection);
    m_enable_cb = true;

    emit order_changed ();

  }

  manager ()->commit ();
}

void
LayerControlPanel::down_clicked ()
{
  manager ()->transaction (tl::translate ("Move down")); 

  std::vector<lay::LayerProperties> new_props;
  new_props.reserve (mp_view->layers ());

  std::set<unsigned int> selection;

  int src = -1;
  bool changed = false;

  QTreeWidgetItemIterator it (mp_layer_list);
  QTreeWidgetItemIterator item (it);
  while (*it) {
    item = it;
    ++it;
  }
  for (unsigned int i = mp_view->layers (); i > 0 && *item; --i) {

    if (mp_layer_list->isItemSelected (*item)) {
      selection.insert (mp_view->layers () - 1 - new_props.size ());
      new_props.push_back (mp_view->properties (i - 1));
      if (src >= 0) {
        changed = true;
      }
    } else {
      if (src >= 0) {
        new_props.push_back (mp_view->properties ((unsigned int) src));
      }
      src = i - 1;
    }

    --item;
  }

  if (src >= 0) {
    new_props.push_back (mp_view->properties ((unsigned int) src));
  }

  if (changed) {

    m_enable_cb = false;
    for (unsigned int i = 0; i < new_props.size (); ++i) {
      mp_view->set_properties (i, new_props [new_props.size () - 1 - i]);
    }
    refresh_display (selection);
    m_enable_cb = true;

    emit order_changed ();

  }

  manager ()->commit ();
}

void
LayerControlPanel::downdown_clicked ()
{
  manager ()->transaction (tl::translate ("Move fully up")); 

  std::vector<lay::LayerProperties> new_props;
  new_props.reserve (mp_view->layers ());

  std::set<unsigned int> selection;

  bool changed = false;

  QTreeWidgetItemIterator item (mp_layer_list);
  for (unsigned int i = 0; i < mp_view->layers () && *item; ++i) {
    if (! mp_layer_list->isItemSelected (*item)) {
      new_props.push_back (mp_view->properties (i));
      changed = true;
    }
    ++item;
  }

  item = QTreeWidgetItemIterator (mp_layer_list);
  for (unsigned int i = 0; i < mp_view->layers () && *item; ++i) {
    if (mp_layer_list->isItemSelected (*item)) {
      selection.insert (new_props.size ());
      new_props.push_back (mp_view->properties (i));
    }
    ++item;
  }

  if (changed) {

    m_enable_cb = false;
    for (unsigned int i = 0; i < new_props.size (); ++i) {
      mp_view->set_properties (i, new_props [i]);
    }
    refresh_display (selection);
    m_enable_cb = true;

    emit order_changed ();

  }

  manager ()->commit ();
}

void
LayerControlPanel::cm_sort_by_name ()
{
  manager ()->transaction (tl::translate ("Sort layers")); 
  sort_layers (ByName);
  manager ()->commit ();

  emit order_changed ();
}

void
LayerControlPanel::cm_sort_by_ild ()
{
  manager ()->transaction (tl::translate ("Sort layers")); 
  sort_layers (ByIndexLayerDatatype);
  manager ()->commit ();

  emit order_changed ();
}

void
LayerControlPanel::cm_sort_by_idl ()
{
  manager ()->transaction (tl::translate ("Sort layers")); 
  sort_layers (ByIndexDatatypeLayer);
  manager ()->commit ();

  emit order_changed ();
}

void
LayerControlPanel::cm_sort_by_ldi ()
{
  manager ()->transaction (tl::translate ("Sort layers")); 
  sort_layers (ByLayerDatatypeIndex);
  manager ()->commit ();

  emit order_changed ();
}

void
LayerControlPanel::cm_sort_by_dli ()
{
  manager ()->transaction (tl::translate ("Sort layers")); 
  sort_layers (ByDatatypeLayerIndex);
  manager ()->commit ();

  emit order_changed ();
}

struct LDSortingProps
  : std::vector <int>
{
  LDSortingProps (int l1, int l2, int l3)
  {
    reserve (3);
    push_back (l1);
    push_back (l2);
    push_back (l3);
  }
};

void
LayerControlPanel::sort_layers (SortOrder order)
{
  std::vector<lay::LayerProperties> new_props;
  new_props.reserve (mp_view->layers ());

  if (order == ByName) {

    //  sort by name and create a list of new properties in new_props:

    std::vector< std::pair<std::string, unsigned int> > data;
    data.reserve (mp_view->layers ());
    for (unsigned int i = 0; i < mp_view->layers (); ++i) {
      const lay::LayerProperties &props = mp_view->properties (i);
      data.push_back (std::make_pair (props.display_string (mp_view), i));
    }

    std::sort (data.begin (), data.end ());

    for (std::vector< std::pair<std::string, unsigned int> >::const_iterator p = data.begin (); p != data.end (); ++p) {
      new_props.push_back (mp_view->properties (p->second));
    }

  } else {

    //  sort by layer/datatype or datatype/layer and create a list of new properties in new_props:

    std::vector< std::pair<LDSortingProps, unsigned int> > data;
    data.reserve (mp_view->layers ());
    for (unsigned int i = 0; i < mp_view->layers (); ++i) {
      const lay::LayerProperties &props = mp_view->properties (i);
      const db::LayerProperties &db_props = mp_view->cellview (props.cellview_index)->layout ().get_properties (props.layer_index);
      if (order == ByIndexLayerDatatype) {
        data.push_back (std::make_pair (LDSortingProps (int (props.cellview_index), db_props.layer, db_props.datatype), i));
      } else if (order == ByIndexDatatypeLayer) {
        data.push_back (std::make_pair (LDSortingProps (int (props.cellview_index), db_props.datatype, db_props.layer), i));
      } else if (order == ByLayerDatatypeIndex) {
        data.push_back (std::make_pair (LDSortingProps (db_props.layer, db_props.datatype, int (props.cellview_index)), i));
      } else if (order == ByDatatypeLayerIndex) {
        data.push_back (std::make_pair (LDSortingProps (db_props.datatype, db_props.layer, int (props.cellview_index)), i));
      }
    }

    std::sort (data.begin (), data.end ());

    for (std::vector< std::pair< LDSortingProps, unsigned int> >::const_iterator p = data.begin (); p != data.end (); ++p) {
      new_props.push_back (mp_view->properties (p->second));
    }

  }

  m_enable_cb = false;
  for (unsigned int i = 0; i < mp_view->layers (); ++i) {
    mp_view->set_properties (i, new_props [i]);
  }
  refresh_display ();
  m_enable_cb = true;
}

void
LayerControlPanel::refresh_display ()
{
  QTreeWidgetItemIterator item (mp_layer_list);
  for (unsigned int i = 0; i < mp_view->layers () && *item; ++i) {
    props_to_display (*item, mp_view->properties (i));
    ++item;
  }
}

struct LayerSelectionChangedOp
  : public db::Op
{
  LayerSelectionChangedOp (unsigned int i, bool s1, bool s2)
    : db::Op (), item (i), sel1 (s1), sel2 (s2)
  { }
  unsigned int item;
  bool sel1, sel2;
};

struct LayerSelectionBeginEndOp
  : public db::Op
{
  LayerSelectionBeginEndOp (bool e)
    : db::Op (), end (e)
  { }
  bool end;
};

void
LayerControlPanel::refresh_display (const std::set<unsigned int> &selected)
{
  if (manager ()->transacting ()) {
    manager ()->queue (this, new LayerSelectionBeginEndOp (false));
  }

  QTreeWidgetItemIterator item (mp_layer_list);
  for (unsigned int i = 0; i < mp_view->layers () && *item; ++i) {

    props_to_display (*item, mp_view->properties (i));
    bool sel1 = mp_layer_list->isItemSelected (*item);
    bool sel2 = (selected.find (i) != selected.end ());
    if (sel1 || sel2) {
      if (manager ()->transacting ()) {
        manager ()->queue (this, new LayerSelectionChangedOp (i, sel1, sel2));
      }
      if (sel1 != sel2) {
        mp_layer_list->setItemSelected (*item, sel2);
      }
    }

    ++item;
    
  }

  if (manager ()->transacting ()) {
    manager ()->queue (this, new LayerSelectionBeginEndOp (true));
  }
}

void
LayerControlPanel::context_menu (const QPoint &p)
{
  QMenu menu (this);
  menu.addAction (tl::translate ("Select All").c_str (), this, SLOT (cm_select_all ()));
  menu.addSeparator ();
  menu.addAction (tl::translate ("Hide").c_str (), this, SLOT (cm_hide ()));
  menu.addAction (tl::translate ("Show").c_str (), this, SLOT (cm_show ()));
  menu.addAction (tl::translate ("Rename").c_str (), this, SLOT (cm_rename ()));
  menu.addSeparator ();
  QMenu *sort_menu = menu.addMenu (tl::translate ("Sort By").c_str ());
  sort_menu->addAction (tl::translate ("Index, Layer And Datatype").c_str (), this, SLOT (cm_sort_by_ild ()));
  sort_menu->addAction (tl::translate ("Index, Datatype And Layer").c_str (), this, SLOT (cm_sort_by_idl ()));
  sort_menu->addAction (tl::translate ("Layer, Datatype And Index").c_str (), this, SLOT (cm_sort_by_ldi ()));
  sort_menu->addAction (tl::translate ("Datatype, Layer And Index").c_str (), this, SLOT (cm_sort_by_dli ()));
  sort_menu->addAction (tl::translate ("Name").c_str (), this, SLOT (cm_sort_by_name ()));
  menu.exec (mp_layer_list->mapToGlobal (p));
}

void 
LayerControlPanel::double_clicked (QTreeWidgetItem *clicked_item, int)
{
  if (!clicked_item) {
    return;
  }

  QTreeWidgetItemIterator item (mp_layer_list);
  unsigned int i;
  for (i = 0; i < mp_view->layers () && *item && *item != clicked_item; ++i, ++item) {
    ;
  }

  if (*item == clicked_item && i < mp_view->layers ()) {

    lay::LayerProperties props = mp_view->properties (i);
    props.display.visible = ! props.display.visible;

    if (props.display.visible) {
      manager ()->transaction (tl::translate ("Show layer")); 
    } else {
      manager ()->transaction (tl::translate ("Hide layer")); 
    }

    m_enable_cb = false;
    mp_view->set_properties (i, props);
    props_to_display (*item, props);
    m_enable_cb = true;

    manager ()->commit ();

    emit visibility_changed ();
  }
}

void 
LayerControlPanel::panel_button_clicked (int index)
{
  if (index < 0 || index >= int (m_tool_panels.size ())) {
    return;
  }

  if (m_tool_panels [index].second->isVisible ()) {
    m_tool_panels [index].second->hide ();
  } else {
    m_tool_panels [index].second->show ();
  }

  rearrange (width (), height ());
}

void
LayerControlPanel::set_background_color (QColor c)
{
  QPalette pl (mp_layer_list->palette ());
  pl.setColor (QColorGroup::Base, c);
  mp_layer_list->setPalette (pl);

  //  recreate the pixmaps
  m_needs_update = true;
  update_content (m_phase);
}

void
LayerControlPanel::set_text_color (QColor c)
{
  QPalette pl (mp_layer_list->palette ());
  pl.setColor (QColorGroup::Text, c);
  mp_layer_list->setPalette (pl);
}

void
LayerControlPanel::signal_observed ()
{
  if (m_enable_cb) {
    m_needs_update = true;
  }
}

void
LayerControlPanel::update_content (unsigned int phase, bool force)
{
  if (force || m_needs_update || phase != m_phase) {

    mp_dither_palette->set_dither_pattern (mp_view->dither_pattern ());

    m_phase = phase;

    QTreeWidgetItemIterator iter (mp_layer_list);
    //  For Qt 4.1.x < 4.1.4 the iterator of an empty QTreeWidget is undefined.
    QTreeWidgetItem *item = (mp_layer_list->topLevelItemCount () == 0) ? 0 : *iter;
    QTreeWidgetItem *last_item = 0;
    for (unsigned int i = 0; i < mp_view->layers (); ++i) {
      QTreeWidgetItem *next_item = 0;
      if (! item) {
        item = new QTreeWidgetItem (mp_layer_list, last_item);
      } else {
        next_item = *++iter;
      }
      props_to_display (item, mp_view->properties (i));
      last_item = item;
      item = next_item;
    }
    while (item) {
      QTreeWidgetItem *next_item = *++iter;
      delete item;
      item = next_item;
    }

    m_needs_update = false;

  }
}

int 
LayerControlPanel::current_layer () const
{
  QTreeWidgetItemIterator item (mp_layer_list);
  for (int i = 0; i < int (mp_view->layers ()) && *item; ++i) {
    if (*item == mp_layer_list->currentItem ()) {
      return i;
    }
    ++item;
  }

  return -1;
}

std::vector <unsigned int>
LayerControlPanel::selected_layers () const
{
  std::vector <unsigned int> llist;

  QTreeWidgetItemIterator item (mp_layer_list);
  for (unsigned int i = 0; i < mp_view->layers () && *item; ++i) {
    if (mp_layer_list->isItemSelected (*item)) {
      llist.push_back (i);
    }
    ++item;
  }

  return llist;
}

void 
LayerControlPanel::undo (db::Op *op) throw ()
{
  LayerSelectionChangedOp *scop = dynamic_cast <LayerSelectionChangedOp *> (op);
  if (scop) {
    mp_layer_list->setItemSelected (mp_layer_list->topLevelItem (scop->item), scop->sel1);
    return;
  }
  LayerSelectionBeginEndOp *beop = dynamic_cast <LayerSelectionBeginEndOp *> (op);
  if (beop->end) {
    mp_layer_list->selectionModel ()->reset ();
    return;
  }
}

void 
LayerControlPanel::redo (db::Op *op) throw ()
{
  LayerSelectionChangedOp *scop = dynamic_cast <LayerSelectionChangedOp *> (op);
  if (scop) {
    mp_layer_list->setItemSelected (mp_layer_list->topLevelItem (scop->item), scop->sel2);
    return;
  }
  LayerSelectionBeginEndOp *beop = dynamic_cast <LayerSelectionBeginEndOp *> (op);
  if (!beop->end) {
    mp_layer_list->selectionModel ()->reset ();
    return;
  }
}

} // namespace lay 
