
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <iostream>
#include <fstream>
#include <vector>

#include "QtCore/QTimer"
#include "QtGui/QSpinBox"
#include "QtGui/QPainter"
#include "QtGui/QPaintEvent"
#include "QtGui/QComboBox"
#include "QtGui/QDialog"
#include "QtGui/QSplitter"
#include "QtGui/QImageWriter"

#include "tlInternational.h"
#include "tlTimer.h"
#include "tlString.h"
#include "tlLog.h"
#include "layViewOp.h"
#include "layViewObject.h"
#include "layLayoutViewConfigPages.h"
#include "layMove.h"
#include "layZoomBox.h"
#include "layMouseTracker.h"
#include "layEditable.h"
#include "laySelector.h"
#include "layColorPalette.h"
#include "layLayoutCanvas.h"
#include "layLayerControlPanel.h"
#include "layHierarchyControlPanel.h"
#include "layMain.h"
#include "layBrowser.h"
#include "dbLayout.h"
#include "dbManager.h"
#include "dbGDS2Reader.h"
#include "tlXMLParser.h"
#include "BookmarkManagementFormImpl.h"

namespace lay
{

// -------------------------------------------------------------
//  LayerProperties implementation

LayerDisplayProperties::LayerDisplayProperties ()
  : frame_color (0), 
    fill_color (0),
    frame_brightness (0),
    fill_brightness (0),
    dither_pattern (0), 
    visible (true),
    transparent (false), 
    width (1),
    marked (false),
    animation (0)
{
  // .. nothing yet ..
}

bool 
LayerDisplayProperties::operator== (const LayerDisplayProperties &d) 
{
  //  do not consider the layer index - this is not really a property
  return frame_color == d.frame_color && 
         fill_color == d.fill_color && 
         frame_brightness == d.frame_brightness &&
         fill_brightness == d.fill_brightness &&
         dither_pattern == d.dither_pattern && 
         visible == d.visible && 
         transparent == d.transparent && 
         width == d.width &&
         marked == d.marked &&
         animation == d.animation &&
         name == d.name;
}

bool 
LayerDisplayProperties::operator!= (const LayerDisplayProperties &d)
{
  return ! operator== (d);
}

static 
color_t brighter (color_t in, int b)
{
  QColor c (in);
  int v, s, h;
  c.getHsv (&h, &s, &v);

  const int vmax = 255;
  const int smax = 255;

  if (v + b > vmax) {
    s -= v + b - vmax;
    v = vmax;
    if (s < 0) {
      s = 0;
    }
  } else if (v + b < 0) {
    s -= v + b;
    v = 0;
    if (s > smax) {
      s = smax;
    }
  } else {
    v += b;
  }

  return QColor::fromHsv (h, s, v).rgb ();
}

color_t 
LayerDisplayProperties::eff_frame_color () const
{
  return brighter (frame_color, frame_brightness);
}

color_t 
LayerDisplayProperties::eff_fill_color () const
{
  return brighter (fill_color, fill_brightness);
}



LayerProperties::LayerProperties ()
  : layer_index (0),
    cellview_index (-1)
{ 
  // .. nothing yet ..
}

db::DBox 
LayerProperties::bbox (const lay::LayoutView *view) const
{
  lay::CellView cv = view->cellview (cellview_index);
  
  if (! cv.is_valid ()) {
    return db::DBox ();
  } else {
    return db::Mag (cv->layout ().dbu ()) * cv.cell ()->bbox (layer_index);
  }
}

std::string
LayerProperties::display_string (const lay::LayoutView *view) const
{
  if (! display.name.empty ()) {
    return display.name;
  } else {
    return source_expr (view);
  }
}

std::string
LayerProperties::source_expr (const lay::LayoutView *view) const
{
  std::string t;
  const lay::CellView &cv = view->cellview (cellview_index);
  if (view->cellviews () > 1) {
    t = cv->layout ().get_properties (layer_index).name + "@" + tl::to_string (cellview_index + 1);
  } else {
    t = cv->layout ().get_properties (layer_index).name;
  }
  return t;
}
    
bool 
LayerProperties::operator== (const LayerProperties &d) 
{
  return display == d.display &&
         cellview_index == d.cellview_index && 
         layer_index == d.layer_index;
}

bool 
LayerProperties::operator!= (const LayerProperties &d)
{
  return ! operator== (d);
}

// -------------------------------------------------------------

struct OpHideShowCell 
  : public db::Op
{
  OpHideShowCell (lay::CellView::cell_index_type ci, int cv_index, bool show)
    : m_cell_index (ci), m_cellview_index (cv_index), m_show (show)
  { }

  lay::CellView::cell_index_type m_cell_index;
  int m_cellview_index;
  bool m_show;
};

struct OpSetDitherPattern
  : public db::Op 
{
  OpSetDitherPattern (const lay::DitherPattern &o, const lay::DitherPattern &n)
    : db::Op (), m_old (o), m_new (n)
  { 
    //  nothing yet.
  }

  lay::DitherPattern m_old, m_new;
};

struct OpLayerList
  : public db::Op
{
  enum Mode { Set, Delete, Insert };

  OpLayerList (unsigned int i, Mode m)
    : m_index (i), m_mode (m)
  { }

  unsigned int m_index;
  lay::LayerProperties m_old, m_new;
  Mode m_mode;
};

struct OpSetLayerProps 
  : public OpLayerList 
{
  OpSetLayerProps (unsigned int i, const lay::LayerProperties &o, const lay::LayerProperties &n)
    : OpLayerList (i, Set)
  { 
    m_old = o; 
    m_new = n;
  }
};

struct OpInsertLayerProps 
  : public OpLayerList
{
  OpInsertLayerProps (unsigned int i, const lay::LayerProperties &n)
    : OpLayerList (i, Insert)
  { 
    m_new = n;
  }
};

struct OpDeleteLayerProps 
  : public OpLayerList
{
  OpDeleteLayerProps (unsigned int i, const lay::LayerProperties &n)
    : OpLayerList (i, Delete)
  { 
    m_old = n;
  }
};

// -------------------------------------------------------------
//  DisplayState implementation

DisplayState::DisplayState ()
  : m_box (), m_cellnames ()
{
  //  .. nothing yet ..
}

DisplayState::DisplayState (const db::DBox &b, const std::vector< std::vector<std::string> > &cns)
  : m_box (b), m_cellnames (cns)
{
  //  .. nothing yet ..
}

DisplayState::DisplayState (const db::DBox &b, const std::vector <lay::CellView> &cvs)
  : m_box (b)
{
  m_cellnames.reserve (cvs.size ());

  //  save the cell names
  for (unsigned int i = 0; i < cvs.size (); ++i) {
    m_cellnames.push_back (std::vector <std::string> ());
    for (CellView::cell_path_type::const_iterator cp = cvs[i].path ().begin (); cp != cvs[i].path ().end (); ++cp) {
      m_cellnames.back ().push_back (cvs[i]->layout ().cell_name (*cp));
    }
  }
}

lay::CellView 
DisplayState::cellview (unsigned int index, lay::LayoutHandle *layout_h) const
{
  //  try to restore the cell path
  lay::CellView::cell_path_type cell_path;

  //  check, if we can reconstruct the path from the names
  bool valid_path = true;
  if (index < m_cellnames.size ()) {
    for (std::vector <std::string>::const_iterator cn = m_cellnames [index].begin (); valid_path && cn != m_cellnames [index].end (); ++cn) {
      std::pair<bool, db::CellGraph::cell_index_type> pci = layout_h->layout ().cell_by_name (cn->c_str ());
      if (pci.first) {
        cell_path.push_back (pci.second);
      } else {
        tl::warn << "Cellname cannot be reconstructed: " << *cn;
        valid_path = false;
      }
    }
  }

  //  create the cellview to return
  lay::CellView cv;
  cv.set (layout_h);

  //  if available, take the path that was reconstructed
  if (valid_path) {
    cv.set_path (cell_path);
    return cv;
  }

  //  otherwise, try to locate the cell by the last component's name 
  if (m_cellnames.size () > index && ! m_cellnames [index].empty ()) {
    std::pair<bool, db::CellGraph::cell_index_type> pci = layout_h->layout ().cell_by_name (m_cellnames [index].back ().c_str ());
    if (pci.first) {
      cv.set_cell (pci.second);
      return cv;
    }
  }

  //  return an empty cellview instead.
  return cv;
}

// -------------------------------------------------------------
//  BookmarkList implementation

/**
 *  @brief A helper class for XML parser: a different representation of the display state structure
 */
struct ExtBookmarkStruct
{
  std::string name;
  double xleft; 
  double ybottom;
  double xright;
  double ytop;
  std::vector< std::vector<std::string> > cellnames;
};

typedef std::vector <ExtBookmarkStruct> ext_bookmark_vector;

static tl::XMLStruct <ext_bookmark_vector>
bookmarks_structure ()
{
  typedef std::vector<std::string> string_v;
  typedef std::vector<string_v> string_vv;

  return tl::XMLStruct <ext_bookmark_vector> ("bookmarks", 
    tl::make_element<ExtBookmarkStruct, ext_bookmark_vector::const_iterator, ext_bookmark_vector> (&ext_bookmark_vector::begin, &ext_bookmark_vector::end, &ext_bookmark_vector::push_back, "bookmark", 
      tl::make_member (&ExtBookmarkStruct::name, "name") + 
      tl::make_member (&ExtBookmarkStruct::xleft, "x-left") + 
      tl::make_member (&ExtBookmarkStruct::xright, "x-right") + 
      tl::make_member (&ExtBookmarkStruct::ybottom, "y-bottom") + 
      tl::make_member (&ExtBookmarkStruct::ytop, "y-top") + 
      tl::make_element (&ExtBookmarkStruct::cellnames, "cellpaths", 
        tl::make_element (&string_vv::begin, &string_vv::end, &string_vv::push_back, "cellpath",
          tl::make_member (&string_v::begin, &string_v::end, &string_v::push_back, "cellname")
        )
      )
    )
  );
}

void 
BookmarkList::load (const std::string &fn) throw (tl::Exception)
{
  ext_bookmark_vector bookmarks;

  tl::XMLFileSource in (fn);
  bookmarks_structure ().parse (in, bookmarks); 

  m_list.clear ();
  m_list.reserve (bookmarks.size ());

  for (ext_bookmark_vector::const_iterator b = bookmarks.begin (); b != bookmarks.end (); ++b) {
    m_list.push_back (std::make_pair (b->name, DisplayState (db::DBox (b->xleft, b->ybottom, b->xright, b->ytop), b->cellnames)));
  }

  if (tl::verbosity () >= 10) {
    tl::info << "Loaded bookmarks from " << fn;
  }
}

void 
BookmarkList::save (const std::string &fn) const throw (tl::Exception)
{
  //  convert the internal bookmark list into the external representation ..

  ext_bookmark_vector bookmarks;

  bookmarks.reserve (m_list.size ());

  for (bookmark_list_type::const_iterator b = m_list.begin (); b != m_list.end (); ++b) {

    bookmarks.push_back (ExtBookmarkStruct ());

    bookmarks.back ().name = b->first;

    bookmarks.back ().xleft   = b->second.box ().left ();
    bookmarks.back ().xright  = b->second.box ().right ();
    bookmarks.back ().ytop    = b->second.box ().top ();
    bookmarks.back ().ybottom = b->second.box ().bottom ();

    bookmarks.back ().cellnames = b->second.cellnames ();

  }

  //  .. and save it:

  std::ofstream out (fn.c_str ());
  if (! out.good ()) {
    throw tl::Exception (tl::translate ("Unable to write bookmarks to file: %s").c_str (), fn.c_str ());
  }
  bookmarks_structure ().write (out, bookmarks); 

  if (tl::verbosity () >= 10) {
    tl::info << "Saved bookmarks to " << fn;
  }

}

// -------------------------------------------------------------

const int timer_interval = 500;

LayoutView::LayoutView (lay::MainWindow *main_window, QWidget *parent, const char *name)
  : QFrame (parent, name), 
    lay::Editables (),
    lay::Plugin (main_window),
    m_manager (),
    mp_status_bar (main_window),
    m_from_level (0), m_to_level (0), 
    m_box_text_transform (true),
    m_box_font (0),
    m_min_size_for_label (16),
    m_cell_box_visible (true),
    m_text_visible (true),
    m_text_lazy_rendering (true),
    m_apply_text_trans (true),
    m_default_text_size (0.1),
    m_text_font (0),
    m_fit_new_cell (true),
    m_activated (true),
    m_animated (false),
    m_phase (0),
    m_display_state_ptr (0),
    m_mode ((unsigned int)-1), // nothing selected yet.
    mp_tracker (0),
    mp_zoom_service (0),
    mp_selection_service (0),
    mp_move_service (0)
{
  //  attach to our own manager:
  manager (&m_manager);

  QVBoxLayout *vbl = new QVBoxLayout (this);
  vbl->setMargin (0);
  vbl->setSpacing (0);

  QSplitter *splitter = new QSplitter (this);
  vbl->addWidget (splitter);

  QFrame *left_frame = new QFrame (splitter);
  QVBoxLayout *left_frame_ly = new QVBoxLayout (left_frame);
  left_frame_ly->setMargin (0);
  left_frame_ly->setSpacing (0);

  mp_hierarchy_panel = new lay::HierarchyControlPanel (this, left_frame, "hcp");
  left_frame_ly->addWidget (mp_hierarchy_panel, 1 /*stretch*/);

  QFrame *levels_frame = new QFrame (left_frame);
  left_frame_ly->addWidget (levels_frame);
  QHBoxLayout *levels_frame_ly = new QHBoxLayout (levels_frame);
  levels_frame_ly->setMargin (1);
  QLabel *level_l1 = new QLabel ((" " + tl::translate ("Levels")).c_str (), levels_frame);
  levels_frame_ly->addWidget (level_l1);
  mp_min_hier_spbx = new QSpinBox (levels_frame);
  levels_frame_ly->addWidget (mp_min_hier_spbx);
  QLabel *level_l2 = new QLabel ("..", levels_frame);
  levels_frame_ly->addWidget (level_l2);
  mp_max_hier_spbx = new QSpinBox (levels_frame);
  levels_frame_ly->addWidget (mp_max_hier_spbx);

  mp_min_hier_spbx->setMaximum (0);
  mp_min_hier_spbx->setValue (0);
  mp_max_hier_spbx->setMaximum (999);
  mp_max_hier_spbx->setValue (0);

  connect (mp_min_hier_spbx, SIGNAL (valueChanged (int)), this, SLOT (min_hier_changed (int)));
  connect (mp_max_hier_spbx, SIGNAL (valueChanged (int)), this, SLOT (max_hier_changed (int)));

  mp_canvas = new lay::LayoutCanvas (splitter, this);
  QSizePolicy sz (mp_canvas->sizePolicy ());
  sz.setHorStretch (1);
  mp_canvas->setSizePolicy (sz);

  //  occupy services and editables:
  //  these services get deleted by the canvas destructor automatically:
  mp_tracker = new lay::MouseTracker (mp_canvas);
  mp_zoom_service = new lay::ZoomService (mp_canvas, this);
  mp_selection_service = new lay::SelectionService (mp_canvas, this);
  mp_move_service = new lay::MoveService (mp_canvas, this);

  mp_control_panel = new lay::LayerControlPanel (this, manager (), splitter, "lcp");

  connect (mp_control_panel, SIGNAL (order_changed ()), this, SLOT (prop_changed ()));
  connect (mp_control_panel, SIGNAL (marked_changed ()), this, SLOT (prop_changed ()));
  connect (mp_control_panel, SIGNAL (width_changed ()), this, SLOT (prop_changed ()));
  connect (mp_control_panel, SIGNAL (animation_changed ()), this, SLOT (prop_changed ()));
  connect (mp_control_panel, SIGNAL (visibility_changed ()), this, SLOT (visibility_changed ()));
  connect (mp_control_panel, SIGNAL (transparency_changed ()), this, SLOT (prop_changed ()));
  connect (mp_control_panel, SIGNAL (stipple_changed ()), this, SLOT (prop_changed ()));
  connect (mp_control_panel, SIGNAL (color_changed ()), this, SLOT (prop_changed ()));
  
  connect (mp_hierarchy_panel, SIGNAL (cell_selected (cell_path_type, int)), this, SLOT (select_cell_dispatch (cell_path_type, int)));

  mp_timer = new QTimer (this);
  connect (mp_timer, SIGNAL (timeout ()), this, SLOT (timer ()));
  mp_timer->start (timer_interval);

  //  create the plugins
  for (tl::Registrar<lay::PluginDeclaration>::iterator cls = tl::Registrar<lay::PluginDeclaration>::begin (); cls != tl::Registrar<lay::PluginDeclaration>::end (); ++cls) {
    mp_plugins.push_back (cls->create_plugin (&m_manager, main_window, this));
  }

  mode (default_mode ());

  config_setup ();
}

LayoutView::~LayoutView ()
{
  //  delete all plugins
  for (std::vector<lay::Plugin *>::iterator p = mp_plugins.begin (); p != mp_plugins.end (); ++p) {
    if (*p) {
      delete *p;
    }
  }
  mp_plugins.clear ();

  //  detach from the manager, so we can safely delete the manager 
  manager (0);

  stop ();

  //  because LayoutView and LayoutCanvas both control lifetimes of
  //  ruler objects for example, it is safer to explictly delete the
  //  LayoutCanvas object here:
  delete mp_canvas;
  mp_canvas = 0;

  delete mp_control_panel;
  mp_control_panel = 0;

  delete mp_hierarchy_panel;
  mp_hierarchy_panel = 0;
}

bool 
LayoutView::configure (const std::string &name, const std::string &value)
{
  if (name == cfg_background_color) {

    QColor color;
    ColorConverter ().from_string (value, color);
    background_color (color);
    return true;

  } else if (name == cfg_min_inst_label_size) {

    int n;
    tl::from_string (value, n);
    min_inst_label_size (n);
    return true;

  } else if (name == cfg_cell_box_text_font) {

    int n;
    tl::from_string (value, n);
    cell_box_text_font (n);
    return true;

  } else if (name == cfg_cell_box_text_transform) {

    bool flag;
    tl::from_string (value, flag);
    cell_box_text_transform (flag);
    return true;

  } else if (name == cfg_cell_box_visible) {

    bool flag;
    tl::from_string (value, flag);
    cell_box_visible (flag);
    return true;

  } else if (name == cfg_cell_box_color) {

    QColor color;
    ColorConverter ().from_string (value, color);
    cell_box_color (color);
    return true;

  } else if (name == cfg_text_color) {

    QColor color;
    ColorConverter ().from_string (value, color);
    text_color (color);
    return true;

  } else if (name == cfg_text_visible) {

    bool flag;
    tl::from_string (value, flag);
    text_visible (flag);
    return true;

  } else if (name == cfg_text_lazy_rendering) {

    bool flag;
    tl::from_string (value, flag);
    text_lazy_rendering (flag);
    return true;

  } else if (name == cfg_apply_text_trans) {

    bool flag;
    tl::from_string (value, flag);
    apply_text_trans (flag);
    return true;

  } else if (name == cfg_default_text_size) {

    double sz;
    tl::from_string (value, sz);
    default_text_size (sz);
    return true;

  } else if (name == cfg_text_font) {

    int n;
    tl::from_string (value, n);
    text_font (n);
    return true;

  } else if (name == cfg_fit_new_cell) {

    bool flag;
    tl::from_string (value, flag);
    fit_new_cell (flag);
    return true;

  } else {
    return false;
  }
}

void 
LayoutView::enable_edits (bool enable)
{
  //  enable or disable these services:
  mp_selection_service->mouse_enable (enable);
  mp_move_service->mouse_enable (enable);

  //  enable or disable the services that implement "lay::ViewService"
  for (std::vector<lay::Plugin *>::iterator p = mp_plugins.begin (); p != mp_plugins.end (); ++p) {
    if (*p) {
      lay::ViewService *svc = (*p)->view_service_interface ();
      if (svc) {
        svc->mouse_enable (enable);
      }
    }
  }
}

unsigned int 
LayoutView::layers ()
{
  return m_layers.size ();
}

int 
LayoutView::current_layer () const
{
  return mp_control_panel->current_layer ();
}

std::vector <unsigned int> 
LayoutView::selected_layers () const
{
  return mp_control_panel->selected_layers ();
}

lay::LayerProperties 
LayoutView::properties (unsigned int n)
{
  return m_layers [n];
}

const lay::DitherPattern &
LayoutView::dither_pattern () const
{
  return mp_canvas->dither_pattern ();
}

void
LayoutView::set_dither_pattern (const lay::DitherPattern &pattern)
{
  if (canvas ()->dither_pattern () != pattern) {
    if (transacting ()) {
      manager ()->queue (this, new OpSetDitherPattern (canvas ()->dither_pattern (), pattern));
    }
    canvas ()->set_dither_pattern (pattern); 
    m_layer_list_observed.signal_observers ();
  }
}

void 
LayoutView::set_properties (unsigned int n, const LayerProperties &props)
{
  if (m_layers [n] != props) {

    if (transacting ()) {
      manager ()->queue (this, new OpSetLayerProps (n, m_layers [n], props));
    }

    //  if the source specification changed, a redraw is required
    bool do_redraw = false;
    if (props.layer_index != m_layers [n].layer_index ||
        props.cellview_index != m_layers [n].cellview_index) {
      do_redraw = true;
    }

    m_layers [n] = props;
    m_layer_list_observed.signal_observers ();

    if (do_redraw) {
      redraw ();
    }

  }
}

unsigned int 
LayoutView::insert_layer (unsigned int before, const LayerProperties &props)
{
  LayerProperties p (props);

  if (transacting ()) {
    manager ()->queue (this, new OpInsertLayerProps (before, p));
  }
  m_layers.insert (m_layers.begin () + before, p);

  //  signal to the observers that something has changed
  m_layer_list_observed.signal_observers ();

  return p.layer_index;
}

void 
LayoutView::delete_layer (unsigned int n)
{
  if (transacting ()) {
    manager ()->queue (this, new OpDeleteLayerProps (n, m_layers [n]));
  }

  //  delete the layer
  m_layers.erase (m_layers.begin () + n);

  //  signal to the observers that something has changed
  m_layer_list_observed.signal_observers ();
}

void 
LayoutView::reserve_layers (unsigned int n)
{
  m_layers.reserve (n);
}

void
LayoutView::redo (db::Op *op) throw ()
{
  tl_assert (! transacting ());

  OpLayerList *lop = dynamic_cast <OpLayerList *> (op);
  if (lop) {

    if (lop->m_mode == OpLayerList::Set) {
      set_properties (lop->m_index, lop->m_new);
    } else if (lop->m_mode == OpLayerList::Insert) {
      unsigned int index = insert_layer (lop->m_index, lop->m_new);
      //  TODO: we assert here that the layer gets the same
      //  index than before. A better strategy would be to
      //  move the layer away if it was occupied in between.
      tl_assert (index == lop->m_new.layer_index);
    } else if (lop->m_mode == OpLayerList::Delete) {
      delete_layer (lop->m_index);
    }

    return;

  } 

  OpSetDitherPattern *stpop = dynamic_cast <OpSetDitherPattern *> (op);
  if (stpop) {
    set_dither_pattern (stpop->m_new);
    return;
  }

  OpHideShowCell *hscop = dynamic_cast <OpHideShowCell *> (op);
  if (hscop) {
    
    if (hscop->m_show) {
      show_cell (hscop->m_cell_index, hscop->m_cellview_index);
    } else {
      hide_cell (hscop->m_cell_index, hscop->m_cellview_index);
    }

    return;

  }

  db::Object::redo (op);
}

void 
LayoutView::undo (db::Op *op) throw ()
{
  tl_assert (! transacting ());

  OpLayerList *lop = dynamic_cast <OpLayerList *> (op);
  if (lop) {

    if (lop->m_mode == OpLayerList::Set) {
      set_properties (lop->m_index, lop->m_old);
    } else if (lop->m_mode == OpLayerList::Insert) {
      delete_layer (lop->m_index);
    } else if (lop->m_mode == OpLayerList::Delete) {
      unsigned int index = insert_layer (lop->m_index, lop->m_old);
      //  TODO: we assert here that the layer gets the same
      //  index than before. A better strategy would be to
      //  move the layer away if it was occupied in between.
      tl_assert (index == lop->m_new.layer_index);
    }

    return;

  } 

  OpHideShowCell *hscop = dynamic_cast <OpHideShowCell *> (op);
  if (hscop) {
    
    if (hscop->m_show) {
      hide_cell (hscop->m_cell_index, hscop->m_cellview_index);
    } else {
      show_cell (hscop->m_cell_index, hscop->m_cellview_index);
    }

    return;

  }

  OpSetDitherPattern *stpop = dynamic_cast <OpSetDitherPattern *> (op);
  if (stpop) {
    set_dither_pattern (stpop->m_old);
    return;
  }

  db::Object::undo (op);
}

void 
LayoutView::add_layer_list_observer (tl::Observer &observer)
{
  m_layer_list_observed.add_observer (observer);
}

void 
LayoutView::remove_layer_list_observer (tl::Observer &observer)
{
  m_layer_list_observed.remove_observer (observer);
}

void 
LayoutView::add_cell_visibility_observer (tl::Observer &observer)
{
  m_cell_visibility_observed.add_observer (observer);
}

void 
LayoutView::remove_cell_visibility_observer (tl::Observer &observer)
{
  m_cell_visibility_observed.remove_observer (observer);
}

void
LayoutView::erase_cellview (unsigned int index)
{
  //  stop the redraw thread since we are about to change the layout object.
  stop ();

  //  no undo available - clear all transactions
  if (manager ()) {
    manager ()->clear ();
  }

  //  from the layer list remove all the relevant entries & erase the cellview
  layer_list_type new_layers;
  new_layers.reserve (m_layers.size ());
  for (layer_list_type::const_iterator l = m_layers.begin (); l != m_layers.end (); ++l) {
    if (l->cellview_index < int (index)) {
      new_layers.push_back (*l);
    } else if (l->cellview_index > int (index)) {
      new_layers.push_back (*l);
      new_layers.back ().cellview_index -= 1;
    }
  }
  m_layers = new_layers;
  m_cellviews.erase (m_cellviews.begin () + index);

  m_layer_list_observed.signal_observers ();

  //  clear the history
  m_display_states.clear ();
  m_display_state_ptr = 0;

  redraw ();
  update_content ();
}

void
LayoutView::clear_cellviews ()
{
  //  stop the redraw thread since we are about to change the layout object.
  stop ();

  //  no undo available - clear all transactions
  if (manager ()) {
    manager ()->clear ();
  }

  //  clear the layer list and cellviews 
  m_layers.clear ();
  m_cellviews.clear ();

  redraw ();

  //  clear the history, store path and zoom box
  m_display_states.clear ();
  m_display_state_ptr = 0;
}

void 
LayoutView::create_deco_layer (unsigned int cvindex)
{
  if (m_cellviews [cvindex].deco_layer () < 0) {
    m_cellviews [cvindex].set_deco_layer (m_cellviews [cvindex]->layout ().insert_special_layer ());
  }
}

void
LayoutView::set_layout (const lay::CellView &cv, unsigned int cvindex) 
{
  //  stop the redraw thread since we are about to change the layout object.
  stop ();

  //  no undo available - clear all transactions
  if (manager ()) {
    manager ()->clear ();
  }

  //  signal the change of layer properties to the observer
  m_layer_list_observed.signal_observers ();

  //  create a new cellview if required
  while (m_cellviews.size () <= cvindex) {
    m_cellviews.push_back (lay::CellView ());
  }

  //  set the handle reference and clear all cell related stuff 
  m_cellviews [cvindex] = cv;

  redraw ();

  //  clear the history, store path and zoom box
  m_display_states.clear ();
  m_display_state_ptr = 0;

  //  this makes the hierarchy panel update it's cell view list:
  mp_hierarchy_panel->update_content ();
}

struct UIntColorConverter 
  : private ColorConverter
{
  std::string to_string (const color_t &c)
  {
    return ColorConverter::to_string (QColor (c));
  }

  void from_string (const std::string &s, color_t &c)
  {
    QColor qc;
    ColorConverter::from_string (s, qc);
    c = qc.rgb ();
  }
};

/**
 *  @brief A helper class for XML parser: extends the display properties by the source spec.
 */
struct ExtLayerDisplayProperties
  : public LayerDisplayProperties
{
  std::string src_expr;
};

struct ExtLayerDisplayPropertiesSet
{
  typedef std::vector <ExtLayerDisplayProperties> ext_prop_vector;
  typedef ext_prop_vector::const_iterator iterator;

  ext_prop_vector properties;
  lay::DitherPattern pattern;

  iterator begin () const 
  {
    return properties.begin ();
  }

  iterator end () const
  {
    return properties.end ();
  }

  void reserve (size_t n)
  {
    properties.reserve (n);
  }

  void push_back (const ExtLayerDisplayProperties &p) 
  {
    properties.push_back (p);
  }

  ExtLayerDisplayProperties &back () 
  {
    return properties.back ();
  }

  const ExtLayerDisplayProperties &back () const
  {
    return properties.back ();
  }

  lay::DitherPattern::iterator begin_custom_dither_pattern () const
  {
    return pattern.begin_custom ();
  }

  lay::DitherPattern::iterator end_custom_dither_pattern () const
  {
    return pattern.end ();
  }

  void push_custom_dither_pattern (const lay::DitherPatternInfo &info)
  {
    return pattern.replace_pattern (pattern.count (), info);
  }
};

static tl::XMLStruct <ExtLayerDisplayPropertiesSet>
layer_props_structure ()
{
  return tl::XMLStruct <ExtLayerDisplayPropertiesSet> ("layer-properties", 
    tl::make_element<ExtLayerDisplayProperties, ExtLayerDisplayPropertiesSet::iterator, ExtLayerDisplayPropertiesSet> (&ExtLayerDisplayPropertiesSet::begin, &ExtLayerDisplayPropertiesSet::end, &ExtLayerDisplayPropertiesSet::push_back, "properties", 
      tl::make_member<lay::color_t, ExtLayerDisplayProperties> (&ExtLayerDisplayProperties::frame_color,    "frame-color", UIntColorConverter ()) + 
      tl::make_member<lay::color_t, ExtLayerDisplayProperties> (&ExtLayerDisplayProperties::fill_color,     "fill-color",  UIntColorConverter ()) + 
      tl::make_member<int, ExtLayerDisplayProperties>          (&ExtLayerDisplayProperties::frame_brightness,  "frame-brightness") + 
      tl::make_member<int, ExtLayerDisplayProperties>          (&ExtLayerDisplayProperties::fill_brightness,   "fill-brightness") + 
      tl::make_member<unsigned int, ExtLayerDisplayProperties> (&ExtLayerDisplayProperties::dither_pattern, "dither-pattern") + 
      tl::make_member<bool, ExtLayerDisplayProperties>         (&ExtLayerDisplayProperties::visible,        "visible") + 
      tl::make_member<bool, ExtLayerDisplayProperties>         (&ExtLayerDisplayProperties::transparent,    "transparent") + 
      tl::make_member<int, ExtLayerDisplayProperties>          (&ExtLayerDisplayProperties::width,          "width") + 
      tl::make_member<bool, ExtLayerDisplayProperties>         (&ExtLayerDisplayProperties::marked,         "marked") + 
      tl::make_member<int, ExtLayerDisplayProperties>          (&ExtLayerDisplayProperties::animation,      "animation") + 
      tl::make_member<std::string, ExtLayerDisplayProperties>  (&ExtLayerDisplayProperties::name,           "name") +
      tl::make_member<std::string, ExtLayerDisplayProperties>  (&ExtLayerDisplayProperties::src_expr,       "source")
    ) +
    tl::make_element (&ExtLayerDisplayPropertiesSet::begin_custom_dither_pattern, &ExtLayerDisplayPropertiesSet::end_custom_dither_pattern, &ExtLayerDisplayPropertiesSet::push_custom_dither_pattern, "custom-dither-pattern", 
      tl::make_element (&lay::DitherPatternInfo::to_strings, &lay::DitherPatternInfo::from_strings, "pattern",
        tl::make_member (&std::vector<std::string>::begin, &std::vector<std::string>::end, &std::vector<std::string>::push_back, "line")
      ) +
      tl::make_member (&lay::DitherPatternInfo::order_index, &lay::DitherPatternInfo::set_order_index, "order") +
      tl::make_member (&lay::DitherPatternInfo::name, &lay::DitherPatternInfo::set_name, "name") 
    )
  );
}

void 
LayoutView::load_layer_props (const std::string &fn) throw (tl::Exception)
{
  //  make a copy so we can explicity set the source expression 
  ExtLayerDisplayPropertiesSet props;

  tl::XMLFileSource in (fn);
  layer_props_structure ().parse (in, props); 

  //  no correlate the read properties with the ones we currently have

  manager ()->transaction (tl::translate ("Load layer properties")); 

  //  build a name to index map and make all layers invisible for first.
  std::map <std::string, unsigned int> lmap; 
  unsigned int i = 0;
  for (layer_list_type::iterator p = m_layers.begin (); p != m_layers.end (); ++p, ++i) {
    lmap.insert (std::make_pair (p->source_expr (this), i));
    LayerProperties lp (*p);
    lp.display.visible = false;
    set_properties (i, lp);
  }

  //  lookup the properties loaded and set their properties
  for (ExtLayerDisplayPropertiesSet::iterator p = props.begin (); p != props.end (); ++p) {
    std::map <std::string, unsigned int>::const_iterator l = lmap.find (p->src_expr); 
    if (l != lmap.end ()) {
      LayerProperties lp (properties (l->second));
      lp.display = *p;
      set_properties (l->second, lp);
    }
  }

  set_dither_pattern (props.pattern);

  manager ()->commit ();

  update_content ();

  if (tl::verbosity () >= 10) {
    tl::info << "Loaded layer properties from " << fn;
  }
}

void 
LayoutView::save_layer_props (const std::string &fn) throw (tl::Exception)
{
  //  make a copy so we can explicity set the source expression 
  ExtLayerDisplayPropertiesSet props;
  props.reserve (m_layers.size ());
  for (layer_list_type::iterator p = m_layers.begin (); p != m_layers.end (); ++p) {
    props.push_back (ExtLayerDisplayProperties ());
    (LayerDisplayProperties &) props.back () = p->display;
    props.back ().src_expr = p->source_expr (this);
  }

  props.pattern = canvas ()->dither_pattern ();

  std::ofstream out (fn.c_str ());
  if (! out.good ()) {
    throw tl::Exception (tl::translate ("Unable to write layer properties to file: %s").c_str (), fn.c_str ());
  }
  layer_props_structure ().write (out, props); 

  if (tl::verbosity () >= 10) {
    tl::info << "Saved layer properties to " << fn;
  }
}

void 
LayoutView::save_screenshot (const std::string &fn) throw (tl::Exception)
{
  QImageWriter writer (fn.c_str (), "PNG");

  //  Unfortunately the PNG writer does not allow to write long strings.
  //  We separate the description into a set of keys:

  for (unsigned int i = 0; i < cellviews (); ++i) {
    if (cellview (i).is_valid ()) {
      std::string name = cellview (i)->layout ().cell_name (cellview (i).cell_index ());
      writer.setText (("Cell" + tl::to_string (int (i) + 1)).c_str (), name.c_str ());
    }
  }

  db::DBox b (box ());
  std::string desc;
  desc += tl::micron_to_string (b.left ()) + "," + tl::micron_to_string (b.bottom ());
  desc += "/";
  desc += tl::micron_to_string (b.right ()) + "," + tl::micron_to_string (b.top ());
  writer.setText ("Rect", desc.c_str ());
  
  if (! writer.write (canvas ()->screenshot ())) {
    throw tl::Exception (tl::translate ("Unable to write screenshot to file: %s (%s)").c_str (), fn.c_str (), writer.errorString ().ascii ());
  }

  if (tl::verbosity () >= 10) {
    tl::info << "Saved screen shot to " << fn;
  }
}

static std::string 
filename_for_caption (const std::string &fn)
{
  const char *cp = fn.c_str ();
  const char *cpp = cp + fn.size ();
  while (cpp > cp && cpp [-1] != '\\' && cpp [-1] != '/') {
    --cpp;
  }
  return cpp;
}

void
LayoutView::reload_layout (unsigned int cv_index) throw (tl::Exception)
{
  stop ();
  clear_selection (); 

  //  save the current view state
  lay::DisplayState state;
  save_view (state);

  //  remember the zoom window
  db::DMagTrans zoom_target (canvas ()->trans ());

  //  this is the cellview at the given index
  const CellView &cvorg = cellview (cv_index);

  //  obtain the original filename  
  std::string filename = cvorg->filename ();

  //  recreate hidden cells by doing a name referencing
  std::vector <std::string> hidden_cells;
  if (m_hidden_cells.size () > cv_index) {
    hidden_cells.reserve (m_hidden_cells [cv_index].size ());
    for (std::set <cell_index_type>::const_iterator ci = m_hidden_cells [cv_index].begin (); ci != m_hidden_cells [cv_index].end (); ++ci) {
      hidden_cells.push_back (std::string (cvorg->layout ().cell_name (*ci)));
    }
  }

  std::vector <std::pair <unsigned int, db::LayerProperties> > org_layers;
  db::GDS2LayerMap lmap;
  db::GDS2LayerMap new_lmap;

  for (unsigned int i = 0; i < cvorg->layout ().layers (); ++i) {
    if (cvorg->layout ().is_valid_layer (i)) {
      const db::LayerProperties &p = cvorg->layout ().get_properties (i);
      if (p.layer >= 0 && p.datatype >= 0) {
        lmap.map (db::GDS2LDPair (p.layer, p.datatype), i); 
      }
      org_layers.push_back (std::make_pair (i, p));
    }
  }

  //  reset the layout: create a dummy handle and install this in between
  //  this will clear the original layout if not further referenced.
  //  Since the dummy layout will act as a placeholder if something goes wrong
  //  when reading the file, it must have the layers created as well
  lay::CellView cv_empty;
  cv_empty.set (new lay::LayoutHandle (new db::Layout (manager ()), filename));
  for (std::vector <std::pair <unsigned int, db::LayerProperties> >::const_iterator ol = org_layers.begin (); ol != org_layers.end (); ++ol) {
    cv_empty->layout ().insert_layer (ol->first, ol->second);
  }
  set_layout (cv_empty, cv_index);
  cellview (cv_index)->rename (filename_for_caption (filename));

  //  create a new handle
  lay::CellView cv;
  cv.set (new lay::LayoutHandle (new db::Layout (manager ()), filename));

  try {

    //  re-create the layers required
    for (std::vector <std::pair <unsigned int, db::LayerProperties> >::const_iterator ol = org_layers.begin (); ol != org_layers.end (); ++ol) {
      cv->layout ().insert_layer (ol->first, ol->second);
    }
    
    {
      if (tl::verbosity () >= 10) {
        tl::info << tl::translate ("Loading GDS2 file: ") << filename;
      }
      tl::SelfTimer timer (tl::verbosity () >= 11, tl::translate ("Loading"));

      db::ZLibStream stream;
      db::GDS2Reader reader (stream);
      stream.open (filename);

      new_lmap = reader.read (cv->layout (), lmap, true);
    }

    //  sort the layout explicitly here. Otherwise it would be done
    //  implicitly at some other time. This may throw an exception
    //  if the operation was cancelled.
    {
      tl::SelfTimer timer (tl::verbosity () >= 11, tl::translate ("Sorting"));
      cv->layout ().update ();
    }

    //  print the memory statistics now.
    if (tl::verbosity () >= 31) {
      db::MemStatistics m;
      cv->layout ().collect_mem_stat (m);
      m.print ();
    }

    //  this is required to release every reference to the cv_empty layout
    cv_empty = lay::CellView ();

    //  install the new layout
    set_layout (cv, cv_index);
    cellview (cv_index)->rename (filename_for_caption (filename));

  } catch (...) {
    update_content ();
    throw;
  }

  //  recreate the hidden cell indices from the names
  if (m_hidden_cells.size () > cv_index) {
    m_hidden_cells [cv_index].clear ();
    for (std::vector <std::string>::const_iterator cn = hidden_cells.begin (); cn != hidden_cells.end (); ++cn) {
      std::pair<bool, cell_index_type> cid = cv->layout ().cell_by_name (cn->c_str ());
      if (cid.first) {
        m_hidden_cells [cv_index].insert (cid.second);
      }
    }
  }

  //  create the initial layer properties for new layers
  std::vector <std::pair <db::GDS2LDPair, unsigned int> > new_layers;

  for (unsigned int i = 0; i < cv->layout ().layers (); ++i) {
    db::GDS2LDPair ld = new_lmap.mapping (i);
    db::GDS2LDPair ld_old = lmap.mapping (i);
    if (ld != db::GDS2LDPair::invalid () && ld_old == db::GDS2LDPair::invalid ()) {
      new_layers.push_back (std::make_pair (ld, i));
    }
  }

  std::sort (new_layers.begin (), new_layers.end ());

  //  do a basic recoloring ..
  unsigned int dp = 0;
  for (std::vector <std::pair <db::GDS2LDPair, unsigned int> >::const_iterator l = new_layers.begin (); l != new_layers.end (); ++l) {
  
    lay::LayerProperties p;
    p.layer_index = l->second;
    p.cellview_index = cv_index;

    lay::color_t c = lay::ColorPalette::luminous_color_by_index (l->first.layer);
    p.display.dither_pattern = ((dp++ % 2) == 0 ? 5 : 9);
    p.display.fill_color = c;
    p.display.frame_color = c;
    p.display.fill_brightness = 0;
    p.display.frame_brightness = 0;

    insert_layer (layers (), p);

  }

  goto_view (state);
  
}

unsigned int 
LayoutView::load_layout (const std::string &filename, bool add_cellview) throw (tl::Exception)
{
  stop ();
  
  //  create a new layout handle 
  lay::CellView cv;
  cv.set (new lay::LayoutHandle (new db::Layout (manager ()), filename));
  cv->rename (filename_for_caption (filename));

  unsigned int cv_index;
  db::GDS2LayerMap lmap;

  try {

    //  load the file
    {
      if (tl::verbosity () >= 10) {
        tl::info << tl::translate ("Loading GDS2 file: ") << filename;
      }
      tl::SelfTimer timer (tl::verbosity () >= 11, tl::translate ("Loading"));

      db::ZLibStream stream;
      db::GDS2Reader reader (stream);
      stream.open (filename);
      lmap = reader.read (cv->layout ());
    }

    //  sort the layout explicitly here. Otherwise it would be done
    //  implicitly at some other time. This may throw an exception
    //  if the operation was cancelled.
    {
      tl::SelfTimer timer (tl::verbosity () >= 11, tl::translate ("Sorting"));
      cv->layout ().update ();
    }

    //  print the memory statistics now.
    if (tl::verbosity () >= 31) {
      db::MemStatistics m;
      cv->layout ().collect_mem_stat (m);
      m.print ();
    }

    //  clear the cellviews if required
    if (! add_cellview) {
      clear_cellviews ();
    }

    //  set the new layout as the layout for the last cellview
    cv_index = cellviews ();
    set_layout (cv, cv_index);

  } catch (...) {

    update_content ();
    throw;

  }

  //  select the first top cell
  db::CellGraph::top_down_const_iterator top = cv->layout ().begin_top_down ();
  if (top != cv->layout ().end_top_down ()) {
    std::vector <db::CellGraph::cell_index_type> p;
    p.push_back (*top);
    select_cell (p, cv_index);
  } else {
    //  even if there is no cell, select the cellview item
    //  to support applications with an active cellview (that is however invalid)
    mp_hierarchy_panel->select_active (cv_index);
  }
  
  //  create the initial layer properties
  std::vector <std::pair <db::GDS2LDPair, unsigned int> > new_layers;

  for (unsigned int i = 0; i < cv->layout ().layers (); ++i) {
    db::GDS2LDPair ld = lmap.mapping (i);
    if (ld != db::GDS2LDPair::invalid ()) {
      new_layers.push_back (std::make_pair (ld, i));
    }
  }

  std::sort (new_layers.begin (), new_layers.end ());

  //  do a basic recoloring ..
  
  unsigned int dp = 0;
  for (std::vector <std::pair <db::GDS2LDPair, unsigned int> >::const_iterator l = new_layers.begin (); l != new_layers.end (); ++l) {
  
    lay::LayerProperties p;
    p.layer_index = l->second;
    p.cellview_index = cellviews () - 1;

    lay::color_t c = lay::ColorPalette::luminous_color_by_index (l->first.layer);
    p.display.dither_pattern = ((dp++ % 2) == 0 ? 5 : 9);
    p.display.fill_color = c;
    p.display.frame_color = c;
    p.display.fill_brightness = 0;
    p.display.frame_brightness = 0;

    insert_layer (layers (), p);

  }

  //  do a fit and update layer lists etc.
  zoom_fit ();
  update_content ();

  return cv_index;
}

void
LayoutView::store_state ()
{
  //  erase all states after the current position
  if (m_display_state_ptr + 1 < m_display_states.size ()) {
    m_display_states.erase (m_display_states.begin () + m_display_state_ptr + 1, m_display_states.end ());
  }

  //  save the state
  DisplayState state (box (), m_cellviews);
  m_display_states.push_back (state);

  m_display_state_ptr = m_display_states.size () - 1;
}

db::DBox 
LayoutView::box () const
{
  db::DBox b = mp_canvas->trans ().inverted () * db::DBox (db::DPoint (0.0, 0.0), db::DPoint (mp_canvas->width (), mp_canvas->height ()));
  return b;
}

void
LayoutView::timer ()
{
  if (m_animated) {
    set_view_ops ();
    mp_control_panel->update_content (m_phase);
    if (m_animated) {
      ++m_phase;
    }
  }  
}

void
LayoutView::update_content ()
{
  if (m_activated) {
    mp_control_panel->update_content (m_phase, true /*force update to reflect empty layers on cell change*/);
    mp_hierarchy_panel->update_content ();
    set_view_ops ();
  }
}

void
LayoutView::zoom_fit ()
{
  //  compute the bounding box over all layers
  //  this will trigger the update procedures of the layout objects if not done yet ..

  db::DBox bbox;

  for (const_layer_iterator l = m_layers.begin (); l != m_layers.end (); ++l) {
    bbox += l->bbox (this);
  }

  if (bbox.empty ()) {
    bbox = db::DBox (0, 0, 100.0, 100.0); // default box
  } else {
    bbox = db::DBox (bbox.left () - 0.025 * bbox.width (), bbox.bottom () - 0.025 * bbox.height (),
                     bbox.right () + 0.025 * bbox.width (), bbox.top () + 0.025 * bbox.height ());
  }

  zoom_box (bbox);
}

void
LayoutView::zoom_box (const db::DBox &bbox)
{
  mp_canvas->zoom_box (bbox);
  store_state ();
  redraw ();
}

void
LayoutView::pan_left ()
{
  shift_window (1.0, -0.15, 0.0);
}

void
LayoutView::pan_center (const db::DPoint &p)
{
  mp_canvas->pan_center (p);
  store_state ();
  redraw ();
}

void
LayoutView::pan_right ()
{
  shift_window (1.0, 0.15, 0.0);
}

void
LayoutView::pan_up ()
{
  shift_window (1.0, 0.0, 0.15);
}

void
LayoutView::pan_down ()
{
  shift_window (1.0, 0.0, -0.15);
}

void
LayoutView::zoom_in ()
{
  shift_window (0.7, 0.0, 0.0);
}

void
LayoutView::zoom_out ()
{
  shift_window (1.5, 0.0, 0.0);
}

void
LayoutView::shift_window (double f, double dx, double dy)
{
  dx = mp_canvas->width () * (dx + (1 - f) * 0.5);
  dy = mp_canvas->height () * (dy + (1 - f) * 0.5);

  db::DMagTrans trans = mp_canvas->trans ();
  db::DMagTrans tnew (db::DTrans (trans.t1.rot (), trans.t1.disp () * (1.0 / f) + db::DPoint (-dx / f, -dy / f)), db::DMag (trans.t2.factor () / f));

  zoom_to (tnew);
}

void
LayoutView::goto_window (const db::DPoint &p, double s)
{
  if (s > 1e-6) {
    db::DBox b (p.x () - s * 0.5, p.y () - s * 0.5, p.x () + s * 0.5, p.y () + s * 0.5);
    zoom_box (b);
  } else {
    db::DBox b (box ());
    b.move (p - b.center ());
    zoom_box (b);
  }
}

void
LayoutView::zoom_to (const db::DMagTrans &tnew)
{
  if (tnew != mp_canvas->trans ()) {
    mp_canvas->zoom_to (tnew);
    store_state ();
    redraw ();
  }
}

void 
LayoutView::redraw_layer (unsigned int index)
{
  do_redraw (index);
}

void
LayoutView::redraw_cell_boxes ()
{
  do_redraw (-1);
}

void
LayoutView::redraw_deco_layer ()
{
  do_redraw (-2);
}

void
LayoutView::redraw (bool clear)
{
  std::set<int> layers;
  canvas ()->redraw (clear, layers);
}

void
LayoutView::cancel ()
{
  //  cancel all drag and pending edit operations such as move operations.
  mp_canvas->drag_cancel ();
  edit_cancel ();
  clear_selection ();
}

void
LayoutView::manage_bookmarks ()
{
  BookmarkManagementFormImpl dialog (this, "bookmark_form", bookmarks ());
  if (dialog.exec ()) {
    bookmarks (dialog.bookmarks ());
  }
}

void 
LayoutView::bookmarks (const BookmarkList &b)
{
  m_bookmarks = b;
}

void
LayoutView::bookmark_view (const std::string &name)
{
  DisplayState state (box (), m_cellviews);
  m_bookmarks.add (name, state);
}

void
LayoutView::goto_view (const DisplayState &state)
{
  mp_canvas->zoom_box (state.box ());

  std::vector <lay::CellView> cellviews;
  cellviews.reserve (m_cellviews.size ());
  for (unsigned int i = 0; i < m_cellviews.size (); ++i) {
    cellviews.push_back (state.cellview (i, m_cellviews [i].operator-> ()));
  }

  select_cellviews (cellviews);

  update_content ();
}

void
LayoutView::save_view (DisplayState &state)
{
  state = DisplayState (box (), m_cellviews);
}

void
LayoutView::do_redraw (int layer)
{
  std::set<int> layers;
  layers.insert (layer);

  canvas ()->redraw (false, layers);
}

void 
LayoutView::visibility_changed ()
{
  set_view_ops ();
  redraw (false /*no clear*/); // draw exposed layers if required
}

void 
LayoutView::prop_changed ()
{
  set_view_ops ();
}

void
LayoutView::set_view_ops ()
{
  std::vector <lay::ViewOp> view_ops;
  view_ops.reserve (layers () * 4 + 2);

  //  cell boxes
  if (m_cell_box_visible) {
    lay::color_t box_color;
    if (! m_box_color.isValid ()) {
      box_color = foreground_color ().rgb ();
    } else {
      box_color = m_box_color.rgb ();
    }
    view_ops.push_back (lay::ViewOp (box_color, lay::ViewOp::Copy, 0));
  } else {
    view_ops.push_back (lay::ViewOp (0, lay::ViewOp::Or, 0));
  }

  bool animated = false;

  for (const_layer_iterator l = m_layers.begin (); l != m_layers.end (); ++l) {

    bool animate_visible = true;
    unsigned int di_off = 0;

    if (l->display.animation) {

      animated = true;
      if (! m_animated) {
        m_animated = true;
        m_phase = 0;
      }

      if (l->display.animation == 1) {
        // scrolling 
        di_off = ((m_phase % 32) << 16);
      } else if (l->display.animation == 2) {
        // blinking
        animate_visible = ((m_phase & 1) == 0);
      } else {
        // inversly blinking
        animate_visible = ((m_phase & 1) != 0);
      }

    }

    if (l->display.visible && animate_visible) {

      lay::ViewOp::Mode mode = lay::ViewOp::Copy;
      if (l->display.transparent) {
        if (background_color ().green () > 128) {
          mode = lay::ViewOp::And;
        } else {
          mode = lay::ViewOp::Or;
        }
      }

      //  fill
      view_ops.push_back (lay::ViewOp (l->display.eff_fill_color (), mode, l->display.dither_pattern | di_off)); // fill
      //  frame
      view_ops.push_back (lay::ViewOp (l->display.eff_frame_color (), mode, 0, lay::ViewOp::Rect, l->display.width)); // frame
      //  text
      if (m_text_visible) {
        if (m_text_color.isValid ()) {
          view_ops.push_back (lay::ViewOp (m_text_color.rgb (), mode, 0));
        } else {
          view_ops.push_back (lay::ViewOp (l->display.eff_frame_color (), mode, 0));
        }
      } else {
        view_ops.push_back (lay::ViewOp (0, lay::ViewOp::Or, 0));
      }
      // vertex
      view_ops.push_back (lay::ViewOp (l->display.eff_frame_color (), mode, 0, lay::ViewOp::Cross, l->display.marked ? 9/*mark size*/ : 0)); // vertex

    } else {
      for (unsigned int i = 0; i < 4; ++i) {
        view_ops.push_back (lay::ViewOp (0, lay::ViewOp::Or, 0));
      }
    }

  }

  if (! animated) {
    m_animated = false;
    m_phase = 0;
  }

  mp_canvas->set_view_ops (view_ops);
}

void 
LayoutView::cell_box_color (QColor c)
{
  if (c != m_box_color) {
    m_box_color = c;
    set_view_ops ();
  }
}

void 
LayoutView::cell_box_text_transform (bool xform)
{
  if (xform != m_box_text_transform) {
    m_box_text_transform = xform;
    redraw ();
  } 
}

void 
LayoutView::cell_box_text_font (unsigned int f)
{
  if (f != m_box_font) {
    m_box_font = f;
    redraw ();
  } 
}

void 
LayoutView::hier_levels (std::pair<unsigned int, unsigned int> l)
{
  if (l != hier_levels ()) {

    mp_min_hier_spbx->setValue (l.first);
    mp_min_hier_spbx->setMaximum (l.second);
    mp_max_hier_spbx->setValue (l.second);
    mp_max_hier_spbx->setMinimum (l.first);

    m_from_level = l.first;
    m_to_level = l.second;

    //  notify the plugins as well
    for (lay::ViewObjectCanvas::service_iterator svc = canvas ()->begin_services (); svc != canvas ()->end_services (); ++svc) {
      svc->hier_levels (l);
    }

    redraw ();

  } 
}

std::pair<unsigned int, unsigned int> 
LayoutView::hier_levels () const
{
  return std::make_pair (m_from_level, m_to_level);
}

void 
LayoutView::min_hier_changed (int i)
{
  mp_max_hier_spbx->setMinimum (i);
  hier_levels (std::make_pair (i, m_to_level));
}

void 
LayoutView::max_hier_changed (int i)
{
  mp_min_hier_spbx->setMaximum (i);
  hier_levels (std::make_pair (m_from_level, i));
}

void 
LayoutView::max_hier ()
{
  //  determine the maximum level of hierarchies
  unsigned int max_level = 0;
  bool ml_set = false;
  for (std::vector <CellView>::const_iterator cv = m_cellviews.begin (); cv != m_cellviews.end (); ++cv) {
    if (cv->is_valid ()) {
      unsigned int nl = cv->cell ()->hierarchy_levels () + 1;
      if (! ml_set || nl > max_level) {
        max_level = nl;
        ml_set = true;
      }
    }
  }

  //  and set the levels
  if (ml_set) {
    hier_levels (std::make_pair (std::min (m_from_level, max_level), max_level));
  }
}

void 
LayoutView::background_color (QColor c)
{
  if (c == background_color ()) {
    return;
  }

  //  replace by "real" background color if required
  if (! c.isValid ()) {
    c = palette ().color (QPalette::Normal, QColorGroup::Base);
  }

  QColor contrast;
  if (c.green () > 128) {
    contrast = QColor (0, 0, 0);
  } else {
    contrast = QColor (255, 255, 255);
  }

  mp_control_panel->set_background_color (c);
  mp_control_panel->set_text_color (contrast);

  mp_hierarchy_panel->set_background_color (c);
  mp_hierarchy_panel->set_text_color (contrast);

  mp_zoom_service->set_color (contrast); // TODO: should have own color?
  mp_selection_service->set_color (contrast); // TODO: should have own color?
  mp_canvas->set_colors (c, contrast, mp_canvas->active_color ());

  set_view_ops ();
}

QColor 
LayoutView::background_color () const
{
  return mp_canvas->background_color ();
}

QColor 
LayoutView::foreground_color () const
{
  return mp_canvas->foreground_color ();
}

void 
LayoutView::select_cellviews_fit (const std::vector <CellView> &cvs)
{
  if (m_cellviews != cvs) {
    clear_selection (); 
    m_cellviews = cvs;
    redraw ();
    zoom_fit ();
    update_content ();
  }
}

void 
LayoutView::select_cell_dispatch (const cell_path_type &path, int cellview_index)
{
  if (m_fit_new_cell) {
    select_cell_fit (path, cellview_index);
  } else {
    select_cell (path, cellview_index);
  }
}

void 
LayoutView::select_cell_fit (const cell_path_type &path, int index)
{ 
  if (index >= 0 && int (m_cellviews.size ()) > index && m_cellviews [index].path () != path) {  
    clear_selection (); 
    m_cellviews [index].set_path (path);
    mp_hierarchy_panel->select_active (index);
    redraw ();
    zoom_fit ();
    update_content ();
  }
}

void 
LayoutView::select_cell_fit (cell_index_type cell_index, int index)
{ 
  if (index >= 0 && int (m_cellviews.size ()) > index && m_cellviews [index].cell_index () != cell_index) {  
    clear_selection (); 
    m_cellviews [index].set_cell (cell_index);
    mp_hierarchy_panel->select_active (index);
    redraw ();
    zoom_fit ();
    update_content ();
  }
}

void 
LayoutView::select_cellviews (const std::vector <CellView> &cvs)
{
  if (m_cellviews != cvs) {
    clear_selection (); 
    m_cellviews = cvs;
    redraw ();
    update_content ();
  }
}

void 
LayoutView::select_cell (const cell_path_type &path, int index)
{ 
  if (index >= 0 && int (m_cellviews.size ()) > index && m_cellviews [index].path () != path) {  
    clear_selection (); 
    m_cellviews [index].set_path (path);
    mp_hierarchy_panel->select_active (index);
    redraw ();
    update_content ();
  }
}

void 
LayoutView::select_cell (cell_index_type cell_index, int index)
{ 
  if (index >= 0 && int (m_cellviews.size ()) > index && m_cellviews [index].cell_index () != cell_index) {  
    clear_selection (); 
    m_cellviews [index].set_cell (cell_index);
    mp_hierarchy_panel->select_active (index);
    redraw ();
    update_content ();
  }
}

bool
LayoutView::is_cell_hidden (cell_index_type ci, int cellview_index) const
{
  if (int (m_hidden_cells.size ()) > cellview_index) {
    return m_hidden_cells [cellview_index].find (ci) != m_hidden_cells [cellview_index].end ();
  } else {
    return false;
  }
}

void 
LayoutView::hide_cell (cell_index_type ci, int cellview_index)
{
  while (int (m_hidden_cells.size ()) <= cellview_index) {
    m_hidden_cells.push_back (std::set <cell_index_type> ());
  }
  if (m_hidden_cells [cellview_index].insert (ci).second) {
    if (transacting ()) {
      manager ()->queue (this, new OpHideShowCell (ci, cellview_index, false /*=hide*/));
    }
    m_cell_visibility_observed.signal_observers ();
    redraw (true);  //  needs redraw
  }
}

void 
LayoutView::show_cell (cell_index_type ci, int cellview_index)
{
  if (int (m_hidden_cells.size ()) > cellview_index) {
    if (m_hidden_cells [cellview_index].erase (ci) > 0) {
      if (transacting ()) {
        manager ()->queue (this, new OpHideShowCell (ci, cellview_index, true /*=show*/));
      }
      m_cell_visibility_observed.signal_observers ();
      redraw (true);  //  needs redraw
    }
  }
}

void 
LayoutView::show_all_cells ()
{
  for (unsigned int i = 0; i < m_hidden_cells.size (); ++i) {
    if (! m_hidden_cells [i].empty ()) {
      if (transacting ()) {
        for (std::set<cell_index_type>::const_iterator ci = m_hidden_cells [i].begin (); ci != m_hidden_cells [i].end (); ++ci) {
          manager ()->queue (this, new OpHideShowCell (*ci, i, true /*=show*/));
        }
      }
      m_hidden_cells [i].clear ();
      m_cell_visibility_observed.signal_observers ();
      redraw (true);  //  needs redraw
      return;
    }
  }
}

void 
LayoutView::min_inst_label_size (int px)
{
  if (m_min_size_for_label != px) {
    m_min_size_for_label = px;
    redraw ();
  }
}

void 
LayoutView::text_visible (bool vis)
{
  if (m_text_visible != vis) {
    m_text_visible = vis;
    set_view_ops ();
    redraw (); //  required because we do some optimizations is text is not visible ..
  }
}

void 
LayoutView::text_lazy_rendering (bool l)
{
  if (m_text_lazy_rendering != l) {
    m_text_lazy_rendering = l;
    redraw (); //  required because we do some optimizations is text is not visible ..
  }
}

void 
LayoutView::cell_box_visible (bool vis)
{
  if (m_cell_box_visible != vis) {
    m_cell_box_visible = vis;
    set_view_ops ();
  }
}

void 
LayoutView::text_font (unsigned int f)
{
  if (m_text_font != f) {
    m_text_font = f;
    redraw ();
  }
}

void 
LayoutView::default_text_size (double fs)
{
  if (m_default_text_size != fs) {
    m_default_text_size = fs;
    redraw ();
  }
}

void 
LayoutView::fit_new_cell (bool f)
{
  m_fit_new_cell = f;
}
  
void 
LayoutView::apply_text_trans (bool f)
{
  if (m_apply_text_trans != f) {
    m_apply_text_trans = f;
    redraw ();
  }
}
  
void 
LayoutView::text_color (QColor c)
{
  if (m_text_color != c) {
    m_text_color = c;
    set_view_ops ();
  }
}

void 
LayoutView::last_display_state ()
{
  if (m_display_state_ptr > 0) {
    m_display_state_ptr--;
    goto_view (m_display_states [m_display_state_ptr]);
  }
}

bool 
LayoutView::has_last_display_state ()
{
  return m_display_state_ptr > 0;
}

void 
LayoutView::next_display_state ()
{
  if (m_display_state_ptr + 1 < m_display_states.size ()) {
    m_display_state_ptr++;
    goto_view (m_display_states [m_display_state_ptr]);
  }
}

bool 
LayoutView::has_next_display_state ()
{
  return m_display_state_ptr + 1 < m_display_states.size ();
}

int 
LayoutView::active_cellview_index () const
{
  return mp_hierarchy_panel->active ();
}

void
LayoutView::current_cell_path (cell_path_type &path) const
{
  mp_hierarchy_panel->current_cell (path);
}

void
LayoutView::activate ()
{
  if (! m_activated) {
    for (std::vector<lay::Plugin *>::iterator p = mp_plugins.begin (); p != mp_plugins.end (); ++p) {
      if ((*p) && (*p)->browser_interface () && (*p)->browser_interface ()->active ()) {
        (*p)->browser_interface ()->show ();
      }
    }
    mp_timer->start (timer_interval);
    m_activated = true;
    update_content ();
  }
  mp_canvas->connect_status_bar (mp_status_bar);  
}

void
LayoutView::deactivate ()
{
  for (std::vector<lay::Plugin *>::iterator p = mp_plugins.begin (); p != mp_plugins.end (); ++p) {
    if ((*p) && (*p)->browser_interface ()) {
      (*p)->browser_interface ()->hide ();
    }
  }

  mp_canvas->connect_status_bar (0);
  canvas ()->free_resources ();
  mp_timer->stop ();
  m_activated = false;
}

void
LayoutView::deactivate_all_browsers ()
{
  for (std::vector<lay::Plugin *>::iterator p = mp_plugins.begin (); p != mp_plugins.end (); ++p) {
    if ((*p) && (*p)->browser_interface ()) {
      (*p)->browser_interface ()->deactivate ();
    }
  }
}

void
LayoutView::stop_redraw ()
{
  canvas ()->stop_redraw ();
}

void
LayoutView::stop ()
{
  stop_redraw ();
  deactivate_all_browsers ();
}

void 
LayoutView::mode (unsigned int m)
{
  if (m != m_mode) {
    m_mode = m;
    if (m >= intrinsic_mouse_modes (0)) {
      mp_canvas->activate (mp_plugins [m - intrinsic_mouse_modes (0)]->view_service_interface ());
    } else if (m == 0) {
      mp_canvas->activate (mp_selection_service);
    } else if (m == 1) {
      mp_canvas->activate (mp_move_service);
    }
  }
}
  
unsigned int 
LayoutView::intrinsic_mouse_modes (std::vector<std::string> *descriptions)
{
  if (descriptions) {
    descriptions->push_back (tl::translate ("Select") + "<:select.png>");
    descriptions->push_back (tl::translate ("Move") + "<:move.png>");
  }
  return 2;
}

unsigned int 
LayoutView::default_mode ()
{
  return 0; // @@@ any generic scheme? is select, should be ruler..
}

void 
LayoutView::enable_editable (unsigned int m, bool en)
{
  enable (mp_plugins [m]->editable_interface (), en);
}
  
void 
LayoutView::menu_activated (const std::string &symbol)
{
  //  distribute the menu on the plugins - one should take it.
  for (std::vector<lay::Plugin *>::iterator p = mp_plugins.begin (); p != mp_plugins.end (); ++p) {
    if (*p) {
      (*p)->menu_activated (symbol);
    }
  }
}

} // namespace lay

