
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef HDR_layLayoutView
#define HDR_layLayoutView

#include <vector>
#include <map>
#include <set>
#include <string>

#include <QtGui/QFrame>

#include "layLayoutCanvas.h"
#include "layCellView.h"
#include "layViewOp.h"
#include "layEditable.h"
#include "layPlugin.h"
#include "tlException.h"

class QSpinBox;

namespace db {
  class Layout;
  class Manager;
}

namespace lay {

class LayerControlPanel;
class HierarchyControlPanel;
class MouseTracker;
class ZoomService;
class SelectionService;
class MoveService;
class MainWindow;
class Browser;
class ColorButton;

/**
 *  @brief A layer display properties structure
 *
 *  The layer properties encapsulate the settings relevant for
 *  the display of a layer.
 *
 *  "brightness" is a index that indicates how much to make the
 *  color brighter to darker rendering the effective color 
 *  (eff_frame_color (), eff_fill_color ()). It's value is roughly between
 *  -255 and 255.
 */
struct LayerDisplayProperties
{
  LayerDisplayProperties ();

  bool operator== (const LayerDisplayProperties &d);
  bool operator!= (const LayerDisplayProperties &d);

  /**
   *  @brief render the effective frame color
   *  
   *  The effective frame color is computed from the frame color brightness and the
   *  frame color.
   */
  color_t eff_frame_color () const;

  /**
   *  @brief render the effective frame color
   *  
   *  The effective frame color is computed from the frame color brightness and the
   *  frame color.
   */
  color_t eff_fill_color () const;

  //  display styles
  color_t frame_color;
  color_t fill_color;
  int frame_brightness;
  int fill_brightness;
  unsigned int dither_pattern;
  bool visible;
  bool transparent;
  int width;
  bool marked;
  int animation;
  std::string name;
};

/**
 *  @brief A layer properties structure
 *
 *  The layer properties encapsulate the settings relevant for
 *  the definition of a layer. That includes source spec and
 *  display.
 */
struct LayerProperties
{
  /**
   *  @brief Constructor
   */
  LayerProperties ();

  /**
   *  @brief Compute the bbox of this layer
   *
   *  This takes the layout and path definition (supported by the
   *  given default layout or path, if no specific is given).
   *  
   *  @return A bbox in micron
   */
  db::DBox bbox (const lay::LayoutView *view) const;

  /**
   *  @brief Obtain the display string for this layer
   */
  std::string display_string (const lay::LayoutView *view) const;

  /**
   *  @brief Obtain the source expression for this layer
   */
  std::string source_expr (const lay::LayoutView *view) const;

  bool operator== (const LayerProperties &d);
  bool operator!= (const LayerProperties &d);

  //  display properties
  LayerDisplayProperties display;

  //  the "source specification"
  unsigned int layer_index;
  int cellview_index;
};


/**
 *  @brief This class encapsulates a display state
 *
 *  A display state is comprised of a box and a cell path for all views.
 *  It can be used to transfer a display state from one layout to another.
 */
struct DisplayState
{
  /**
   *  @brief Default ctor
   */
  DisplayState ();

  /**
   *  @brief Create a display state from a given set of cell views and a box
   */
  DisplayState (const db::DBox &b, const std::vector <lay::CellView> &cvs);

  /**
   *  @brief Create a display state from a given set of cell name paths 
   */
  DisplayState (const db::DBox &b, const std::vector< std::vector<std::string> > &cns);

  /**
   *  @brief Transform a partial display state back into a cellview 
   *
   *  This method provides some safety: if the display state cannot be
   *  transferred into the layout, some reasonable assumption is made 
   *  and an artifical state is created.
   *
   *  @param index The index of the cellview to obtain
   *  @param layout The layout to refer to
   *  @return A valid cellview for the given layout
   */
  lay::CellView cellview (unsigned int index, lay::LayoutHandle *layout_h) const;

  /**
   *  @brief Obtain the box
   */
  const db::DBox &box () const
  {
    return m_box;
  }

  /**
   *  @brief Obtain the raw cell name list
   */
  const std::vector< std::vector<std::string> > cellnames () const
  {
    return m_cellnames;
  }

private:
  db::DBox m_box;
  std::vector< std::vector<std::string> > m_cellnames;
};

/**
 *  @brief The list of bookmarks
 */
class BookmarkList
{
public:
  typedef std::pair<std::string, DisplayState> bookmark_type;
  typedef std::vector<bookmark_type> bookmark_list_type;
  typedef bookmark_list_type::const_iterator const_iterator;

  /**
   *  @brief Add a bookmark 
   */
  void add (const std::string &name, const DisplayState &state)
  {
    m_list.push_back (std::make_pair (name, state));
  }

  /**
   *  @brief Reserve a certain number of elements
   */
  void reserve (size_t n)
  {
    m_list.reserve (n);
  }

  /**
   *  @brief Clear the bookmark list
   */
  void clear ()
  {
    m_list.clear ();
  }
  
  /**
   *  @brief Obtain the number of bookmarks
   */
  size_t size () const
  {
    return m_list.size ();
  }

  /**
   *  @brief Rename the element with the given index
   */
  void rename (size_t index, const std::string &name)
  {
    m_list [index].first = name;
  }
  
  /**
   *  @brief Obtain the name of the element with the given index
   */
  const std::string &name (size_t index) const
  {
    return m_list [index].first;
  }
  
  /**
   *  @brief Set the state of the element with the given index
   */
  void state (size_t index, const DisplayState &state)
  {
    m_list [index].second = state;
  }
  
  /**
   *  @brief Obtain the name of the element with the given index
   */
  const DisplayState &state (size_t index) const
  {
    return m_list [index].second;
  }

  /**
   *  @brief Save the list 
   */
  void save (const std::string &fn) const throw (tl::Exception);

  /**
   *  @brief Load the list
   */
  void load (const std::string &fn) throw (tl::Exception);

private:
  bookmark_list_type m_list;
};

/**
 *  @brief The layout view object
 *
 *  The layout view is responsible for displaying one or a set of layouts.
 *  It is composed of a canvas and controls to control the appearance.
 *  It manages the layer display list, bookmark list etc.
 */
class LayoutView
  : public QFrame,
    public lay::Editables,
    public lay::Plugin
{
Q_OBJECT

public:
  typedef std::vector<lay::LayerProperties> layer_list_type;
  typedef layer_list_type::iterator layer_iterator;
  typedef layer_list_type::const_iterator const_layer_iterator;
  typedef lay::CellView::cell_path_type cell_path_type;
  typedef lay::CellView::cell_index_type cell_index_type;
  typedef std::pair<std::string, DisplayState> bookmark_type;

  /**
   *  @brief Constructor
   */
  LayoutView (lay::MainWindow *main_widget, QWidget *parent = 0, const char *name = "main_window");

  /** 
   *  @brief Destructor
   */
  ~LayoutView ();

  /**
   *  @brief Retrieve the layer properties of a layer with the given index
   */
  LayerProperties properties (unsigned int n);

  /**
   *  @brief Set the layer properties of a layer with the given index
   */
  void set_properties (unsigned int n, const LayerProperties &props);

  /**
   *  @brief Set the dither pattern 
   */
  void set_dither_pattern (const lay::DitherPattern &pattern);

  /**
   *  @brief Obtain the dither pattern 
   */
  const lay::DitherPattern &dither_pattern () const;

  /**
   *  @brief Insert a layer at a given position
   *  @param before The index before which to insert the layer 
   *  @param props The properties of the layer to insert
   *  @return The internal layer index for db::Cell assigned to the
   *          layer.
   */
  unsigned int insert_layer (unsigned int before, const LayerProperties &props);

  /**
   *  @brief Reserve space for n layers
   */
  void reserve_layers (unsigned int n);

  /**
   *  @brief Delete a layer
   */
  void delete_layer (unsigned int n);

  /**
   *  @brief Add a layer list observer
   */
  void add_layer_list_observer (tl::Observer &observer); 

  /**
   *  @brief Remove a layer list observer
   */
  void remove_layer_list_observer (tl::Observer &observer); 

  /**
   *  @brief Add a cell visibility observer
   */
  void add_cell_visibility_observer (tl::Observer &observer); 

  /**
   *  @brief Remove a cell visibility observer
   */
  void remove_cell_visibility_observer (tl::Observer &observer); 

  /**
   *  @brief Begin of layer properties list (const)
   */
  const_layer_iterator begin_layers () const
  {
    return m_layers.begin ();
  }

  /**
   *  @brief End of layer properties list (const)
   */
  const_layer_iterator end_layers () const
  {
    return m_layers.end ();
  }

  /**
   *  @brief Retrieve the number of layer in the layer list
   */
  unsigned int layers ();

  /**
   *  @brief Retrieve the index of the currently active layer
   *
   *  The active layer is the one that is active in the layer
   *  browser panel.
   *  This method returns a negative value, if no layer is active.
   */
  int current_layer () const;

  /**
   *  @brief Return the layers that are selected in the layer browser
   *
   *  Returns an empty list if no layer is selected.
   */
  std::vector <unsigned int> selected_layers () const;

  /**
   *  @brief Implementation of the undo operations
   */
  virtual void undo (db::Op *op) throw ();

  /**
   *  @brief Implementation of the redo operations
   */
  virtual void redo (db::Op *op) throw ();

  /** 
   *  @brief Set the cellview at the given index
   *
   *  If cvindex is used as the cellview index to associate the 
   *  layout with.
   */
  void set_layout (const lay::CellView &cv, unsigned int cvindex);

  /** 
   *  @brief Create a decoration layer for the cellview at the given position
   *
   *  If there already is a decoration layer, just leave it.
   */
  void create_deco_layer (unsigned int cvindex);

  /**
   *  @brief clear the cellviews
   */
  void clear_cellviews ();

  /**
   *  @brief erase one cellview with the given index
   */
  void erase_cellview (unsigned int index);

  /**
   *  @brief Save the layer properties 
   */
  void save_layer_props (const std::string &fn) throw (tl::Exception);

  /**
   *  @brief Load the layer properties 
   */
  void load_layer_props (const std::string &fn) throw (tl::Exception);

  /**
   *  @brief Save the screen content to a file
   */
  void save_screenshot (const std::string &fn) throw (tl::Exception);

  /**
   *  @brief Hierarchy level selection setter
   */
  void hier_levels (std::pair<unsigned int, unsigned int> l);

  /**
   *  @brief Hierarchy level selection setter
   */
  void min_hier_levels (unsigned int l)
  {
    hier_levels (std::pair<unsigned int, unsigned int> (l, max_hier_levels ()));
  }

  /**
   *  @brief Hierarchy level selection setter
   */
  void max_hier_levels (unsigned int l)
  {
    hier_levels (std::pair<unsigned int, unsigned int> (min_hier_levels (), l));
  }

  /**
   *  @brief Hierarchy level selection getter
   */
  std::pair<unsigned int, unsigned int> hier_levels () const;

  /**
   *  @brief Hierarchy level selection getter
   */
  unsigned int min_hier_levels () const
  {
    return hier_levels ().first;
  }

  /**
   *  @brief Hierarchy level selection getter
   */
  unsigned int max_hier_levels () const
  {
    return hier_levels ().second;
  }

  /**
   *  @brief Return true, if there is a "last" display state
   */
  bool has_last_display_state ();

  /**
   *  @brief Return true, if there is a "next" display state
   */
  bool has_next_display_state ();

  /**
   *  @brief Cell box/label color setter
   */
  void cell_box_color (QColor c);

  /**
   *  @brief Cell box/label getter
   */
  QColor cell_box_color () const
  {
    return m_box_color;
  }

  /**
   *  @brief Transform label flag setter
   */
  void cell_box_text_transform (bool xform);

  /**
   *  @brief Transform label flag getter
   */
  bool cell_box_text_transform () const
  {
    return m_box_text_transform;
  }

  /**
   *  @brief Label font setter
   */
  void cell_box_text_font (unsigned int f);

  /**
   *  @brief Label font setter
   */
  unsigned int cell_box_text_font () const
  {
    return m_box_font;
  }

  /** 
   *  @brief Visibility of cell boxes
   */
  void cell_box_visible (bool vis);

  /** 
   *  @brief Visibility of cell boxes
   */
  bool cell_box_visible () const
  {
    return m_cell_box_visible;
  }

  /**
   *  @brief Min instance label size setter
   */
  void min_inst_label_size (int px);

  /**
   *  @brief Min instance label size getter
   */
  int min_inst_label_size () const
  {
    return m_min_size_for_label;
  }

  /** 
   *  @brief Visibility of text objects
   */
  void text_visible (bool vis);

  /** 
   *  @brief Visibility of text objects
   */
  bool text_visible () const
  {
    return m_text_visible;
  }

  /** 
   *  @brief Lazy rendering of text objects
   */
  void text_lazy_rendering (bool lzy);

  /** 
   *  @brief Lazy rendering of text objects
   */
  bool text_lazy_rendering () 
  {
    return m_text_lazy_rendering;
  }

  /**
   *  @brief Text object font setter
   */
  void text_font (unsigned int f);

  /**
   *  @brief Text object font getter
   */
  unsigned int text_font () const
  {
    return m_text_font;
  }

  /**
   *  @brief The default text size property
   */
  void default_text_size (double fs);

  /**
   *  @brief The default text size property
   */
  double default_text_size () const
  {
    return m_default_text_size;
  }

  /**
   *  @brief Apply text transformation property
   */
  void apply_text_trans (bool f);
  
  /**
   *  @brief Apply text transformation property
   */
  bool apply_text_trans () const
  {
    return m_apply_text_trans;
  }
  
  /**
   *  @brief Text object color
   */
  void text_color (QColor c);

  /**
   *  @brief Text object color
   */
  QColor text_color () const
  {
    return m_text_color;
  }

  /**
   *  @brief Fit new cell property
   *
   *  If this property is set to false, the coordinate window is not
   *  changed if the cell is changed.
   */
  void fit_new_cell (bool f);
  
  /**
   *  @brief Fit new cell
   *
   *  If this property is set to false, the coordinate window is not
   *  changed if the cell is changed.
   */
  bool fit_new_cell () const
  {
    return m_fit_new_cell;
  }
  
  /**
   *  @brief Get the canvas object 
   */
  lay::LayoutCanvas *canvas () 
  {
    return mp_canvas;
  }

  /** 
   *  @brief Reload the given cellview into the current view
   *
   *  The cellview is given by index in the current view's cellview list.
   */
  void reload_layout (unsigned int cv_index) throw (tl::Exception);

  /** 
   *  @brief Load a (new) file into the layout
   *
   *  The add_cellview param controls whether to create a new cellview
   *  or clear all cellviews before.
   *
   *  @return The index of the cellview loaded.
   */
  unsigned int load_layout (const std::string &filename, bool add_cellview) throw (tl::Exception);

  /**
   *  @brief Deactivate the view
   *
   *  A view gets "deactivated" if it is pushed into the background.
   *  This is the opposite operation than activation. 
   */
  void deactivate ();

  /**
   *  @brief Activate the view
   *
   *  A view gets "activated" soon before it is shown. Activation 
   *  may involve allocation of resources only required in visible state.
   *  The active window is made owner of the status bar - i.e. only 
   *  this window may write to it.
   */
  void activate ();

  /**
   *  @brief Pass the current position from the mouse tracker to the status bar
   */ 
  void current_pos (double x, double y);

  /**
   *  @brief Object the index of the active cellview (shown in hierarchy browser)
   */
  int active_cellview_index () const;
  
  /**
   *  @brief Cell path of the current cell
   *
   *  The current cell is the one highlighted in the browser. The 
   *  cellview index corresponding to the current cell can be obtained with
   *  "active_cellview_index".
   */
  void current_cell_path (cell_path_type &path) const;
  
  /**
   *  @brief Obtain the number of cellviews
   */
  unsigned int cellviews () const
  {
    return m_cellviews.size ();
  }
  
  /**
   *  @brief Obtain the cell view reference for an index
   *
   *  If the index is invalid, an empty cell view reference is returned
   */
  const CellView &cellview (unsigned int index) const
  {
    static const CellView empty;
    if (index >= m_cellviews.size ()) {
      return empty;
    } else {
      return m_cellviews [index];
    }
  }

  /** 
   *  @brief Select the list of cellviews for this window and fit cell
   *
   *  Warning: use with care!
   */
  void select_cellviews_fit (const std::vector <CellView> &cvs);

  /** 
   *  @brief Select the list of cellviews for this window
   *
   *  Warning: use with care!
   */
  void select_cellviews (const std::vector <CellView> &cvs);

  /** 
   *  @brief Shift and scale the window
   */
  void shift_window (double f, double dx, double dy);

  /** 
   *  @brief Goto window
   *
   *  Position the window to the new position x and y with a size of s (approximately).
   *  If s is negative or zero, the windows is just shifted.
   */
  void goto_window (const db::DPoint &p, double s = -1.0);

  /** 
   *  @brief Return the displayed window
   */
  db::DBox box () const;
  
  /**
   *  @brief Bookmark the current view under the given name
   */
  void bookmark_view (const std::string &name);

  /**
   *  @brief Show the bookmark management form
   */
  void manage_bookmarks ();

  /**
   *  @brief Obtain the bookmarks list
   */
  const BookmarkList &bookmarks () const
  {
    return m_bookmarks;
  }

  /**
   *  @brief Set the bookmarks list
   */
  void bookmarks (const BookmarkList &b);

  std::pair <bool, std::string> redo_available ();
  std::pair <bool, std::string> undo_available ();

  /**
   *  @brief Select a certain mode (by index)
   */
  void mode (unsigned int m);
  
  /**
   *  @brief Query the intrinsic mouse modes available
   */
  static unsigned int intrinsic_mouse_modes (std::vector<std::string> *descriptions);

  /**
   *  @brief Query the default mode
   */
  static unsigned int default_mode ();

  /**
   *  @brief Enable an editable service by index
   */
  void enable_editable (unsigned int m, bool en);
  
public slots:
  /** 
   *  @brief Select last display state
   */
  void last_display_state ();

  /** 
   *  @brief Select next display state
   */
  void next_display_state ();

  void zoom_fit ();
  void zoom_to (const db::DMagTrans &t);
  void zoom_box (const db::DBox &b);
  void zoom_in ();
  void zoom_out ();
  void pan_left ();
  void pan_up ();
  void pan_right ();
  void pan_down ();
  void pan_center (const db::DPoint &p);

  /**
   *  @brief Goto a position/cell view that was saved with save_view 
   */
  void goto_view (const DisplayState &state);

  /**
   *  @brief Save a position/cell view that can be used for "goto_view"
   */
  void save_view (DisplayState &state);

  /**
   *  @brief Store the current state on the "last states" stack
   */
  void store_state ();

  /**
   *  @brief Cancel all edit operations
   */
  void cancel ();

  /**
   *  @brief Select all levels of hierarchy available
   */
  void max_hier ();

  /**
   *  @brief Stop activities like shape browser and redrawing thread
   */
  void stop ();

  /**
   *  @brief Stop redrawing thread
   */
  void stop_redraw ();

  /**
   *  @brief Undo operation
   */
  void undo () 
  {
    m_manager.undo ();
  }

  /**
   *  @brief Redo operation
   */
  void redo () 
  {
    m_manager.redo ();
  }

  /**
   *  @brief Enable or disable the actions for edit functions 
   * 
   *  This method is used by non-modal dialogs that want to suppress any editing
   *  activities (like browsers) while they are open.
   */
  void enable_edits (bool enable);

  /** 
   *  @brief Select a cell by path for a certain cell view and fit cell
   */
  void select_cell_fit (const cell_path_type &path, int cellview_index);

  /** 
   *  @brief Select a cell by index for a certain cell view and fit cell
   */
  void select_cell_fit (cell_index_type index, int cellview_index);

  /** 
   *  @brief Select a cell by path for a certain cell view
   */
  void select_cell (const cell_path_type &path, int cellview_index);

  /** 
   *  @brief Select a cell by index for a certain cell view
   */
  void select_cell (cell_index_type index, int cellview_index);

  /** 
   *  @brief Select a cell by index for a certain cell view
   *
   *  This will be forwarded to select_cell or select_cell_fit depending
   *  on the m_fit_new_cell flag.
   */
  void select_cell_dispatch (const cell_path_type &path, int cellview_index);

  /**
   *  @brief Check if a cell is hidden
   */
  bool is_cell_hidden (cell_index_type ci, int cellview_index) const;

  /**
   *  @brief Add a cell to the hidden cell list
   */
  void hide_cell (cell_index_type ci, int cellview_index);

  /**
   *  @brief Remove a cell from the hidden cell list
   */
  void show_cell (cell_index_type ci, int cellview_index);

  /**
   *  @brief Clear the hidden cell list 
   */
  void show_all_cells ();

  /**
   *  @brief Access to the hidden cell list
   */
  const std::vector <std::set <cell_index_type> > &hidden_cells () const
  { 
    return m_hidden_cells;
  }

  // called by children and owner
  void redraw (bool clear = true);
  void redraw_layer (unsigned int index);
  void redraw_deco_layer ();
  void redraw_cell_boxes ();
  void prop_changed ();
  void visibility_changed ();
  void timer ();
  void update_content ();
  void menu_activated (const std::string &symbol);
  void deactivate_all_browsers ();
  void min_hier_changed (int i);
  void max_hier_changed (int i);

protected:
  void set_view_ops ();

private:
  db::Manager m_manager;
  QTimer *mp_timer;
  lay::StatusBar *mp_status_bar;
  lay::LayoutCanvas *mp_canvas;
  lay::LayerControlPanel *mp_control_panel;
  lay::HierarchyControlPanel *mp_hierarchy_panel;
  QSpinBox *mp_min_hier_spbx;
  QSpinBox *mp_max_hier_spbx;
  std::vector <CellView> m_cellviews;
  std::vector <std::set <cell_index_type> > m_hidden_cells;

  unsigned int m_from_level, m_to_level;

  QColor m_box_color;
  bool m_box_text_transform;
  unsigned int m_box_font;
  int m_min_size_for_label;
  bool m_cell_box_visible;

  bool m_text_visible;
  bool m_text_lazy_rendering;
  QColor m_text_color;
  bool m_apply_text_trans;
  double m_default_text_size;
  unsigned int m_text_font;

  bool m_fit_new_cell;

  bool m_activated;
  bool m_animated;
  unsigned int m_phase;

  std::vector <DisplayState> m_display_states;
  unsigned int m_display_state_ptr;

  layer_list_type m_layers;
  tl::Observed m_layer_list_observed;
  tl::Observed m_cell_visibility_observed;

  BookmarkList m_bookmarks;

  //  service and editable management
  unsigned int m_mode;
  
  //  services & editables
  lay::MouseTracker *mp_tracker;
  lay::ZoomService *mp_zoom_service;
  lay::SelectionService *mp_selection_service;
  lay::MoveService *mp_move_service;

  std::vector<lay::Plugin *> mp_plugins;

  void do_redraw (int layer);
  void do_redraw ();

  void background_color (QColor c);
  QColor background_color () const;
  QColor foreground_color () const;

  bool configure (const std::string &name, const std::string &value);
};

}

#endif

