
/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include <stdlib.h>
#include <ctype.h>
#include <limits>

#include <tlString.h>
#include <tlInternational.h>

std::string 
tl::to_string (double d, const char *f)
{
  char b[32];
  ::sprintf (b, f, d); 
  return b;
}

std::string 
tl::to_string (int d)
{
  char b[32];
  ::sprintf (b, "%d", d); 
  return b;
}

std::string 
tl::to_string (unsigned int d)
{
  char b[32];
  ::sprintf (b, "%u", d); 
  return b;
}

std::string 
tl::to_string (long d)
{
  char b[32];
  ::sprintf (b, "%ld", d); 
  return b;
}

std::string 
tl::to_string (unsigned long d)
{
  char b[32];
  ::sprintf (b, "%lu", d); 
  return b;
}

std::string
tl::to_string (const char *cp)
{
  return std::string (cp);
}

std::string
tl::to_string (const unsigned char *cp)
{
  return std::string ((const char *) cp);
}

std::string
tl::to_string (const char *cp, int length)
{
  return std::string (cp, 0, length);
}

std::string
tl::to_string (const unsigned char *cp, int length)
{
  return std::string ((const char *) cp, 0, length);
}

std::string
tl::to_string (bool b)
{
  return b ? "true" : "false";
}

void
tl::from_string (const std::string &s, const char * &result)
{
  result = s.c_str ();
}

void
tl::from_string (const std::string &s, const unsigned char * &result)
{
  result = (unsigned char *) s.c_str ();
}

void 
tl::from_string (const std::string &s, double &v) throw (tl::Exception)
{
  const char *cp = s.c_str ();
  while (*cp && isspace (*cp)) {
    ++cp;
  }
  if (! *cp) {
    throw tl::Exception (tl::translate ("Got empty string where a real number was expected"));
  }
  char *cp_end = const_cast <char *> (cp);
  v = strtod (cp, &cp_end);
  while (*cp_end && isspace (*cp_end)) {
    ++cp_end;
  }
  if (*cp_end) {
    throw tl::Exception (tl::translate ("Not a valid real number: ") + s);
  }
}

template <class T>
void 
convert_string_to_int (const std::string &s, T &v) throw (tl::Exception)
{
  double x;
  tl::from_string (s, x);
  if (x < std::numeric_limits <T>::min ()) {
    throw tl::Exception (tl::translate ("Range underflow: ") + s);
  }
  if (x > std::numeric_limits <T>::max ()) {
    throw tl::Exception (tl::translate ("Range overflow: ") + s);
  }
  v = T (x);
  if (x != v) {
    throw tl::Exception (tl::translate ("Number cannot be represented precisely: ") + s);
  }
}

void 
tl::from_string (const std::string &s, int &v) throw (tl::Exception)
{
  convert_string_to_int (s, v);
}

void 
tl::from_string (const std::string &s, long &v) throw (tl::Exception)
{
  convert_string_to_int (s, v);
}

void 
tl::from_string (const std::string &s, unsigned int &v) throw (tl::Exception)
{
  convert_string_to_int (s, v);
}

void 
tl::from_string (const std::string &s, unsigned long &v) throw (tl::Exception)
{
  convert_string_to_int (s, v);
}

void
tl::from_string (const std::string &s, bool &b) throw (tl::Exception)
{
  std::string t (trim (s));
  if (t == "true") {
    b = true;
  } else if (t == "false") {
    b = false;
  } else if (t == "1") {
    b = true;
  } else if (t == "0") {
    b = false;
  } else {
    throw tl::Exception (tl::translate ("Invalid boolean value: ") + s);
  }
}


std::string 
tl::sprintf (const char *fmt, va_list args)
{
  char b[256];
  int n = vsnprintf (b, sizeof (b), fmt, args);
  if (n >= int (sizeof (b))) {

    //  the fixed buffer size above was not sufficient - 
    //  dynamically allocate enough memory
    char *bp = new char [n + 1];
    vsnprintf (bp, n + 1, fmt, args);
    std::string s (bp);
    delete [] bp;
    return s;

  } else {
    return std::string (b);
  }
}

std::string 
tl::sprintf (const char *fmt, ...)
{
  va_list args;
  va_start (args, fmt);

  char b[256];
  int n = vsnprintf (b, sizeof (b), fmt, args);
  if (n >= int (sizeof (b))) {

    //  the fixed buffer size above was not sufficient - 
    //  dynamically allocate enough memory
    char *bp = new char [n + 1];
    vsnprintf (bp, n + 1, fmt, args);
    std::string s (bp);
    delete [] bp;
    return s;

  } else {
    return std::string (b);
  }
}

std::string 
tl::trim (const std::string &s)
{
  const char *cp = s.c_str ();
  while (isspace (*cp) && *cp) {
    ++cp;
  }

  const char *cq = s.c_str () + s.size ();
  while (cq > cp && isspace (cq [-1])) {
    --cq;
  }

  return std::string (cp, 0, cq - cp);
}

// -------------------------------------------------------------------
//  tl::string implementation

tl::string::string (const char *c)
{
  if (c && *c) {
    m_capacity = m_size = strlen (c);
    allocator_t alloc;
    mp_rep = alloc.allocate (m_capacity + 1);
    strcpy (mp_rep, c);
  } else {
    mp_rep = 0;
    m_capacity = m_size = 0;
  }
}

tl::string::string (const char *c, size_t from, size_t to)
{
  m_capacity = m_size = to - from;
  if (m_size > 0) {
    allocator_t alloc;
    mp_rep = alloc.allocate (m_capacity + 1);
    strncpy (mp_rep, c + from, m_size);
    mp_rep [m_size] = 0;
  } else {
    mp_rep = 0;
  }
}

tl::string::string (const tl::string &s)
{
  m_capacity = m_size = s.size ();
  if (m_size > 0) {
    allocator_t alloc;
    mp_rep = alloc.allocate (m_capacity + 1);
    strncpy (mp_rep, s.c_str (), m_size);
    mp_rep [m_size] = 0;
  } else {
    mp_rep = 0;
  }
}

tl::string::string (const tl::string &s, size_t from, size_t to)
{
  m_capacity = m_size = to - from;
  if (m_size > 0) {
    allocator_t alloc;
    mp_rep = alloc.allocate (m_capacity + 1);
    strncpy (mp_rep, s.c_str () + from, m_size);
    mp_rep [m_size] = 0;
  } else {
    mp_rep = 0;
  }
}

tl::string::string (const std::string &s)
{
  m_capacity = m_size = s.size ();
  if (m_size > 0) {
    allocator_t alloc;
    mp_rep = alloc.allocate (m_capacity + 1);
    strncpy (mp_rep, s.c_str (), m_size);
    mp_rep [m_size] = 0;
  } else {
    mp_rep = 0;
  }
}

tl::string::string (const std::string &s, size_t from, size_t to)
{
  m_capacity = m_size = to - from;
  if (m_size > 0) {
    allocator_t alloc;
    mp_rep = alloc.allocate (m_capacity + 1);
    strncpy (mp_rep, s.c_str () + from, m_size);
    mp_rep [m_size] = 0;
  } else {
    mp_rep = 0;
  }
}

tl::string::~string ()
{
  if (mp_rep) {
    allocator_t alloc;
    alloc.deallocate (mp_rep, m_capacity + 1);
  }
  mp_rep = 0;
}

tl::string &
tl::string::operator= (const char *c)
{
  if (c && *c) {
    assign (c, 0, strlen (c));
  } else {
    m_size = 0;
    if (mp_rep) {
      mp_rep [0] = 0;
    }
  }
  return *this;
}

void 
tl::string::assign (const char *c, size_t from, size_t to)
{
  m_size = to - from;
  if (m_size > 0) {
    if (m_capacity < m_size) {
      allocator_t alloc;
      if (mp_rep) {
        alloc.deallocate (mp_rep, m_capacity + 1);
      }
      mp_rep = alloc.allocate (m_size + 1);
      m_capacity = m_size;
    }
    strncpy (mp_rep, c + from, m_size);
    mp_rep [m_size] = 0;
  } else {
    if (mp_rep) {
      mp_rep [0] = 0;
    }
  }
}

tl::string &
tl::string::operator= (const tl::string &s)
{
  if (&s != this) {
    m_size = s.size ();
    if (m_size > 0) {
      if (m_capacity < m_size) {
        allocator_t alloc;
        if (mp_rep) {
          alloc.deallocate (mp_rep, m_capacity + 1);
        }
        mp_rep = alloc.allocate (m_size + 1);
        m_capacity = m_size;
      }
      strncpy (mp_rep, s.mp_rep, m_size);
      mp_rep [m_size] = 0;
    } else {
      if (mp_rep) {
        mp_rep [0] = 0;
      }
    }
  }
  return *this;
}

void 
tl::string::assign (const tl::string &s, size_t from, size_t to)
{
  if (&s != this) {
    assign (s.c_str (), from, to);
  } else if (from != 0 || to != m_size) {
    tl::string substr (s, from, to);
    swap (substr);
  }
}

tl::string &
tl::string::operator= (const std::string &s)
{
  assign (s.c_str (), 0, s.size ());
  return *this;
}

void 
tl::string::assign (const std::string &s, size_t from, size_t to)
{
  assign (s.c_str (), from, to);
}

void
tl::string::clear ()
{
  if (mp_rep) {
    allocator_t alloc;
    alloc.deallocate (mp_rep, m_capacity + 1);
    mp_rep = 0;
  }
  m_size = 0;
  m_capacity = 0;
}

void 
tl::string::reserve (size_t n) 
{
  if (m_capacity < n) {
    allocator_t alloc;
    char *nrep = alloc.allocate (n + 1);
    strncpy (nrep, mp_rep, m_size);
    if (mp_rep) {
      alloc.deallocate (mp_rep, m_capacity + 1);
    }
    mp_rep = nrep;
    m_capacity = n; 
  }
}

bool 
tl::string::operator== (const char *c) const
{
  return (c[0] == c_str()[0] && strcmp (c, c_str()) == 0);
}

bool 
tl::string::operator== (const tl::string &s) const
{
  return (c_str()[0] == s.c_str()[0] && strcmp (c_str(), s.c_str()) == 0);
}

bool 
tl::string::operator!= (const char *c) const
{
  return (c[0] != c_str()[0] || strcmp (c, c_str()) != 0);
}

bool 
tl::string::operator!= (const tl::string &s) const
{
  return (c_str()[0] != s.c_str()[0] || strcmp (c_str(), s.c_str()) != 0);
}

bool 
tl::string::operator< (const char *c) const
{
  return strcmp (c_str(), c) < 0;
}
 
bool 
tl::string::operator< (const tl::string &s) const
{
  return strcmp (c_str(), s.c_str()) < 0;
}

bool 
tl::string::operator<= (const char *c) const
{
  return strcmp (c_str(), c) <= 0;
}
 
bool 
tl::string::operator<= (const tl::string &s) const
{
  return strcmp (c_str(), s.c_str()) <= 0;
}

bool 
tl::string::operator> (const char *c) const
{
  return strcmp (c_str(), c) > 0;
}
 
bool 
tl::string::operator> (const tl::string &s) const
{
  return strcmp (c_str(), s.c_str()) > 0;
}

bool 
tl::string::operator>= (const char *c) const
{
  return strcmp (c_str(), c) >= 0;
}
 
bool 
tl::string::operator>= (const tl::string &s) const
{
  return strcmp (c_str(), s.c_str()) >= 0;
}

